@@ -33,6 +33,7 @@
 #include <crm/cluster/internal.h>
 
 #include <crm/common/xml.h>
+#include <crm/common/remote_internal.h>
 
 #include <cibio.h>
 #include <callbacks.h>
@@ -40,6 +41,8 @@
 #include <notify.h>
 #include "common.h"
 
+#define EXIT_ESCALATION_MS 10000
+
 static unsigned long cib_local_bcast_num = 0;
 
 typedef struct cib_local_notify_s {
@@ -641,8 +644,8 @@ parse_peer_options_v1(int call_type, xmlNode * request,
         return TRUE;
     }
 
-    crm_trace("Processing %s request sent by %s", op, originator);
     op = crm_element_value(request, F_CIB_OPERATION);
+    crm_trace("Processing %s request sent by %s", op, originator);
     if (safe_str_eq(op, "cib_shutdown_req")) {
         /* Always process these */
         *local_notify = FALSE;
@@ -697,11 +700,11 @@ parse_peer_options_v1(int call_type, xmlNode * request,
 
     } else if (safe_str_eq(op, "cib_shutdown_req")) {
         if (reply_to != NULL) {
-            crm_debug("Processing %s from %s", op, host);
+            crm_debug("Processing %s from %s", op, originator);
             *needs_reply = FALSE;
 
         } else {
-            crm_debug("Processing %s reply from %s", op, host);
+            crm_debug("Processing %s reply from %s", op, originator);
         }
         return TRUE;
 
@@ -752,10 +755,18 @@ parse_peer_options_v2(int call_type, xmlNode * request,
          * limit on how far newer nodes will go
          */
         const char *max = crm_element_value(request, F_CIB_SCHEMA_MAX);
+        const char *upgrade_rc = crm_element_value(request, F_CIB_UPGRADE_RC);
 
-        crm_trace("Parsing %s operation%s for %s with max=%s",
-                  op, is_reply?" reply":"", cib_is_master?"master":"slave", max);
-        if(max == NULL && cib_is_master) {
+        crm_trace("Parsing %s operation%s for %s with max=%s and upgrade_rc=%s",
+                  op, (is_reply? " reply" : ""),
+                  (cib_is_master? "master" : "slave"),
+                  (max? max : "none"), (upgrade_rc? upgrade_rc : "none"));
+
+        if (upgrade_rc != NULL) {
+            // Our upgrade request was rejected by DC, notify clients of result
+            crm_xml_add(request, F_CIB_RC, upgrade_rc);
+
+        } else if ((max == NULL) && cib_is_master) {
             /* We are the DC, check if this upgrade is allowed */
             goto skip_is_reply;
 
@@ -764,7 +775,8 @@ parse_peer_options_v2(int call_type, xmlNode * request,
             goto skip_is_reply;
 
         } else {
-            return FALSE; /* Ignore */
+            // Ignore broadcast client requests when we're not DC
+            return FALSE;
         }
 
     } else if (crm_is_true(update)) {
@@ -1706,7 +1718,7 @@ initiate_exit(void)
     send_cluster_message(NULL, crm_msg_cib, leaving, TRUE);
     free_xml(leaving);
 
-    g_timeout_add(crm_get_msec("5s"), cib_force_exit, NULL);
+    g_timeout_add(EXIT_ESCALATION_MS, cib_force_exit, NULL);
 }
 
 extern int remote_fd;
