@@ -1,5 +1,7 @@
 /* 
- * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
+ * Copyright 2004-2019 the Pacemaker project contributors
+ *
+ * The version control history for this file may have further details.
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
@@ -95,6 +97,9 @@ pe_ipc_destroy(gpointer user_data)
         crm_info("Connection to the Policy Engine released");
     }
 
+    // If we aren't connected to the scheduler, we can't expect a reply
+    controld_expect_sched_reply(NULL);
+
     clear_bit(fsa_input_register, pe_subsystem->flag_connected);
     pe_subsystem->pid = -1;
     pe_subsystem->source = NULL;
@@ -135,6 +140,9 @@ do_pe_control(long long action,
     };
 
     if (action & stop_actions) {
+        // If we aren't connected to the scheduler, we can't expect a reply
+        controld_expect_sched_reply(NULL);
+
         clear_bit(fsa_input_register, pe_subsystem->flag_required);
 
         mainloop_del_ipc_client(pe_subsystem->source);
@@ -171,6 +179,87 @@ do_pe_control(long long action,
 
 int fsa_pe_query = 0;
 char *fsa_pe_ref = NULL;
+static mainloop_timer_t *controld_sched_timer = NULL;
+
+// @TODO Make this a configurable cluster option if there's demand for it
+#define SCHED_TIMEOUT_MS (120000)
+
+/*!
+ * \internal
+ * \brief Handle a timeout waiting for scheduler reply
+ *
+ * \param[in] user_data  Ignored
+ *
+ * \return FALSE (indicating that timer should not be restarted)
+ */
+static gboolean
+controld_sched_timeout(gpointer user_data)
+{
+    if (AM_I_DC) {
+        /* If this node is the DC but can't communicate with the scheduler, just
+         * exit (and likely get fenced) so this node doesn't interfere with any
+         * further DC elections.
+         *
+         * @TODO We could try something less drastic first, like disconnecting
+         * and reconnecting to the scheduler, but something is likely going
+         * seriously wrong, so perhaps it's better to just fail as quickly as
+         * possible.
+         */
+        crmd_exit(DAEMON_RESPAWN_STOP);
+    }
+    return FALSE;
+}
+
+void
+controld_stop_sched_timer()
+{
+    if (controld_sched_timer && fsa_pe_ref) {
+        crm_trace("Stopping timer for scheduler reply %s", fsa_pe_ref);
+    }
+    mainloop_timer_stop(controld_sched_timer);
+}
+
+/*!
+ * \internal
+ * \brief Set the scheduler request currently being waited on
+ *
+ * \param[in] msg  Request to expect reply to (or NULL for none)
+ */
+void
+controld_expect_sched_reply(xmlNode *msg)
+{
+    char *ref = NULL;
+
+    if (msg) {
+        ref = crm_element_value_copy(msg, XML_ATTR_REFERENCE);
+        CRM_ASSERT(ref != NULL);
+
+        if (controld_sched_timer == NULL) {
+            controld_sched_timer = mainloop_timer_add("scheduler_reply_timer",
+                                                      SCHED_TIMEOUT_MS, FALSE,
+                                                      controld_sched_timeout,
+                                                      NULL);
+        }
+        mainloop_timer_start(controld_sched_timer);
+    } else {
+        controld_stop_sched_timer();
+    }
+    free(fsa_pe_ref);
+    fsa_pe_ref = ref;
+}
+
+/*!
+ * \internal
+ * \brief Free the scheduler reply timer
+ */
+void
+controld_free_sched_timer()
+{
+    if (controld_sched_timer != NULL) {
+        mainloop_timer_del(controld_sched_timer);
+        controld_sched_timer = NULL;
+    }
+}
 
 /*	 A_PE_INVOKE	*/
 void
@@ -216,10 +305,7 @@ do_pe_invoke(long long action,
     crm_debug("Query %d: Requesting the current CIB: %s", fsa_pe_query,
               fsa_state2string(fsa_state));
 
-    /* Make sure any queued calculations are discarded */
-    free(fsa_pe_ref);
-    fsa_pe_ref = NULL;
-
+    controld_expect_sched_reply(NULL);
     fsa_register_cib_callback(fsa_pe_query, FALSE, NULL, do_pe_invoke_callback);
 }
 
@@ -320,8 +406,9 @@ do_pe_invoke_callback(xmlNode * msg, int call_id, int rc, xmlNode * output, void
 
     CRM_LOG_ASSERT(output != NULL);
 
-    /* refresh our remote-node cache when the pengine is invoked */
-    crm_remote_peer_cache_refresh(output);
+    /* Refresh the remote node cache and the known node cache when the
+     * scheduler is invoked */
+    crm_peer_caches_refresh(output);
 
     crm_xml_add(output, XML_ATTR_DC_UUID, fsa_our_uuid);
     crm_xml_add_int(output, XML_ATTR_HAVE_QUORUM, fsa_has_quorum);
@@ -334,16 +421,14 @@ do_pe_invoke_callback(xmlNode * msg, int call_id, int rc, xmlNode * output, void
 
     cmd = create_request(CRM_OP_PECALC, output, NULL, CRM_SYSTEM_PENGINE, CRM_SYSTEM_DC, NULL);
 
-    free(fsa_pe_ref);
-    fsa_pe_ref = crm_element_value_copy(cmd, XML_ATTR_REFERENCE);
-
     sent = crm_ipc_send(mainloop_get_ipc_client(pe_subsystem->source), cmd, 0, 0, NULL);
     if (sent <= 0) {
         crm_err("Could not contact the pengine: %d", sent);
         register_fsa_error_adv(C_FSA_INTERNAL, I_ERROR, NULL, NULL, __FUNCTION__);
+    } else {
+        controld_expect_sched_reply(cmd);
+        crm_debug("Invoking the PE: query=%d, ref=%s, seq=%llu, quorate=%d",
+                  fsa_pe_query, fsa_pe_ref, crm_peer_seq, fsa_has_quorum);
     }
-
-    crm_debug("Invoking the PE: query=%d, ref=%s, seq=%llu, quorate=%d",
-              fsa_pe_query, fsa_pe_ref, crm_peer_seq, fsa_has_quorum);
     free_xml(cmd);
 }
