@@ -121,7 +121,7 @@ class CtsLab:
 
     This class is derived from a UserDict because we hold many
     different parameters of different kinds, and this provides
-    provide a uniform and extensible interface useful for any kind of
+    a uniform and extensible interface useful for any kind of
     communication between the user/administrator/tester and CTS.
 
     At this point in time, it is the intent of this class to model static
@@ -246,7 +246,7 @@ class NodeStatus:
         else:
             try:
                 answer = raw_input('Continue? [nY]')
-            except EOFError, e:
+            except EOFError as e:
                 answer = "n"
         if answer and answer == "n":
             raise ValueError("%s did not come up within %d tries" % (node, Timeout))
@@ -346,7 +346,8 @@ class ClusterManager(UserDict):
             else:
                 self.ShouldBeStatus[node] = "down"
 
-            self.unisolate_node(node)
+            if self.Env["experimental-tests"]:
+                self.unisolate_node(node)
 
     def upcount(self):
         '''How many nodes are up?'''
@@ -372,6 +373,25 @@ class ClusterManager(UserDict):
             self.rsh.cp(file_with_path, "root@%s:%s/%s" % (node, destdir, filename))
         return file_with_path
 
+    def uninstall_helper(self, filename, dir=None, nodes=None):
+        if not nodes:
+            nodes = self.Env["nodes"]
+
+        if not dir:
+            dir = CTSvars.CTS_home
+
+        for node in nodes:
+            self.debug("Uninstalling %s from %s on %s" % (filename, dir, repr(node)))
+            self.rsh(node, "rm -rf %s/%s" % (dir, filename))
+
+    def systemctl_daemon_reload(self, nodes=None):
+        if not nodes:
+            nodes = self.Env["nodes"]
+
+        for node in nodes:
+            self.debug("Reloading the system manager configuration on %s" % (repr(node)))
+            self.rsh(node, "systemctl daemon-reload")
+
     def install_config(self, node):
         return None
 
@@ -615,7 +635,7 @@ class ClusterManager(UserDict):
         if self.rsh(node, self.templates["StopCmd"]) == 0:
             # Make sure we can continue even if corosync leaks
             # fdata-* is the old name
-            #self.rsh(node, "rm -f /dev/shm/qb-* /dev/shm/fdata-*")
+            #self.rsh(node, "rm -rf /dev/shm/qb-* /dev/shm/fdata-*")
             self.ShouldBeStatus[node] = "down"
             self.cluster_stable(self.Env["DeadTime"])
             return 1
