@@ -45,54 +45,155 @@
 
 /* *INDENT-OFF* */
 static struct crm_option long_options[] = {
-    {"help",        0, 0, '?', "\tThis text"},
-    {"version",     0, 0, '$', "\tVersion information"  },
-    {"verbose",     0, 0, 'V', "\tIncrease debug output"},
-    {"quiet",       0, 0, 'q', "\tPrint only essential output"},
+    {   "help", no_argument, NULL, '?',
+        "\tDisplay this text and exit."
+    },
+    {   "version", no_argument, NULL, '$',
+        "\tDisplay version information and exit."
+    },
+    {   "verbose", no_argument, NULL, 'V',
+        "\tIncrease debug output (may be specified multiple times)."
+    },
+    {   "quiet", no_argument, NULL, 'q',
+        "\tBe less descriptive in output."
+    },
+    {   "cleanup", no_argument, NULL, 'c',
+        "\tCleanup wherever appropriate. Requires: --history."
+    },
+    {   "broadcast", no_argument, NULL, 'b',
+        "Broadcast wherever appropriate."
+    },
+    {   "-spacer-", no_argument, NULL, '-', "\nDevice definition commands:" },
 
-    {"-spacer-",    0, 0, '-', "Commands:"},
-    {"list",            1, 0, 'l', "List devices that can terminate the specified host"},
-    {"list-registered", 0, 0, 'L', "List all registered devices"},
-    {"list-installed",  0, 0, 'I', "List all installed devices"},
+    {   "register", required_argument, NULL, 'R',
+        "Register the named stonith device. Requires: --agent.\n"
+        "\t\t\tOptional: --option, --env-option."
+    },
+    {   "deregister", required_argument, NULL, 'D',
+        "De-register the named stonith device."
+    },
+    {   "register-level", required_argument, NULL, 'r',
+        "Register a stonith level for the named target,\n"
+        "\t\t\tspecified as one of NAME, @PATTERN, or ATTR=VALUE.\n"
+        "\t\t\tRequires: --index and one or more --device entries."
+    },
+    {   "deregister-level", required_argument, NULL, 'd',
+        "Unregister a stonith level for the named target,\n"
+        "\t\t\tspecified as for --register-level. Requires: --index."
+    },
 
-    {"-spacer-",    0, 0, '-', ""},
-    {"metadata",    0, 0, 'M', "Check the device's metadata"},
-    {"query",       1, 0, 'Q', "Check the device's status"},
-    {"fence",       1, 0, 'F', "Fence the named host"},
-    {"unfence",     1, 0, 'U', "Unfence the named host"},
-    {"reboot",      1, 0, 'B', "Reboot the named host"},
-    {"confirm",     1, 0, 'C', "Confirm the named host is now safely down"},
-    {"history",     1, 0, 'H', "Retrieve last fencing operation for specified node (or '*' for all)"},
-    {"last",        1, 0, 'h', "Indicate when the named node was last fenced. Optional: --as-node-id"},
+    {   "-spacer-", no_argument, NULL, '-', "\nQueries:" },
 
-    {"-spacer-",    0, 0, '-', ""},
-    {"register",    1, 0, 'R', "Register the named stonith device. Requires: --agent, optional: --option"},
-    {"deregister",  1, 0, 'D', "De-register the named stonith device"},
+    {   "list", required_argument, NULL, 'l',
+        "List devices that can terminate the specified host.\n"
+        "\t\t\tOptional: --timeout."
+    },
+    {   "list-registered", no_argument, NULL, 'L',
+        "List all registered devices. Optional: --timeout."
+    },
+    {   "list-installed", no_argument, NULL, 'I',
+        "List all installed devices. Optional: --timeout."
+    },
+    {   "list-targets", required_argument, NULL, 's',
+        "List the targets that can be fenced by the\n"
+        "\t\t\tnamed device. Optional: --timeout."
+    },
+    {   "metadata", no_argument, NULL, 'M',
+        "\tShow agent metadata. Requires: --agent.\n"
+        "\t\t\tOptional: --timeout."
+    },
+    {   "query", required_argument, NULL, 'Q',
+        "Check the named device's status. Optional: --timeout."
+    },
+    {   "history", required_argument, NULL, 'H',
+        "Show last successful fencing operation for named node\n"
+        "\t\t\t(or '*' for all nodes). Optional: --timeout, --cleanup,\n"
+        "\t\t\t--quiet (show only the operation's epoch timestamp),\n"
+        "\t\t\t--verbose (show all recorded and pending operations),\n"
+        "\t\t\t--broadcast (update history from all nodes available)."
+    },
+    {   "last", required_argument, NULL, 'h',
+        "Indicate when the named node was last fenced.\n"
+        "\t\t\tOptional: --as-node-id."
+    },
+    {   "validate", no_argument, NULL, 'K',
+        "\tValidate a fence device configuration.\n"
+        "\t\t\tRequires: --agent. Optional: --option, --env-option,\n"
+        "\t\t\t--quiet (print no output, only return status).\n"
+    },
 
-    {"register-level",    1, 0, 'r',
-     "Register a stonith level for the named target, specified as one of:\n\t"
-     "NAME, @PATTERN, or ATTR=VALUE. Requires: --index, one or more --device entries"},
-    {"deregister-level",  1, 0, 'd', "De-register a stonith level for the named target\n\t"
-     "Target is specified as for --register-level. Requires: --index"},
+    {   "-spacer-", no_argument, NULL, '-', "\nFencing Commands:" },
 
-    {"-spacer-",    0, 0, '-', ""},
-    {"-spacer-",    0, 0, '-', "Options and modifiers:"},
-    {"agent",       1, 0, 'a', "The agent (eg. fence_xvm) to instantiate when calling with --register"},
-    {"env-option",  1, 0, 'e'},
-    {"option",      1, 0, 'o'},
-    {"tag",         1, 0, 'T', "Identify fencing operations with the specified tag.\n\t"
-     "Useful when there are multiple entities that might be invoking stonith_admin(8)"},
+    {   "fence", required_argument, NULL, 'F',
+        "Fence named host. Optional: --timeout, --tolerance, --delay."
+    },
+    {   "unfence", required_argument, NULL, 'U',
+        "Unfence named host. Optional: --timeout, --tolerance, --delay."
+    },
+    {   "reboot", required_argument, NULL, 'B',
+        "Reboot named host. Optional: --timeout, --tolerance, --delay."
+    },
+    {   "confirm", required_argument, NULL, 'C',
+        "Tell cluster that named host is now safely down."
+    },
 
-    {"device",      1, 0, 'v', "A device to associate with a given host and stonith level"},
-    {"index",       1, 0, 'i', "The stonith level (1-9)"},
+    {   "-spacer-", no_argument, NULL, '-', "\nAdditional Options:" },
 
-    {"timeout",     1, 0, 't', "Operation timeout in seconds"},
-    {"as-node-id",  0, 0, 'n', "(Advanced) The supplied node is the corosync nodeid (Only for use with --last)" },
-    {"tolerance",   1, 0,   0, "(Advanced) Do nothing if an equivalent --fence request succeeded less than N seconds earlier" },
+    {   "agent", required_argument, NULL, 'a',
+        "The agent to use (for example, fence_xvm;\n"
+        "\t\t\twith --register, --metadata, --validate)."
+    },
+    {   "option", required_argument, NULL, 'o',
+        "Specify a device configuration parameter as NAME=VALUE\n"
+        "\t\t\t(may be specified multiple times; with --register,\n"
+        "\t\t\t--validate)."
+    },
+    {   "env-option", required_argument, NULL, 'e',
+        "Specify a device configuration parameter with the\n"
+        "\t\t\tspecified name, using the value of the\n"
+        "\t\t\tenvironment variable of the same name prefixed with\n"
+        "\t\t\tOCF_RESKEY_ (may be specified multiple times;\n"
+        "\t\t\twith --register, --validate)."
+    },
+    {   "tag", required_argument, NULL, 'T',
+        "Identify fencing operations in logs with the specified\n"
+        "\t\t\ttag; useful when multiple entities might invoke\n"
+        "\t\t\tstonith_admin (used with most commands)."
+    },
+    {   "device", required_argument, NULL, 'v',
+        "Device ID (with --register-level, device to associate with\n"
+        "\t\t\ta given host and level; may be specified multiple times)"
+#if SUPPORT_CIBSECRETS
+        "\n\t\t\t(with --validate, name to use to load CIB secrets)"
+#endif
+        "."
+    },
+    {   "index", required_argument, NULL, 'i',
+        "The stonith level (1-9) (with --register-level,\n"
+        "\t\t\t--deregister-level)."
+    },
+    {   "timeout", required_argument, NULL, 't',
+        "Operation timeout in seconds (default 120;\n"
+        "\t\t\tused with most commands)."
+    },
+    {   "delay", required_argument, NULL, 'y',
+        "Apply a fencing delay in seconds. Any static/random delays from\n"
+        "\t\t\tpcmk_delay_base/max will be added, otherwise all\n"
+        "\t\t\tdisabled with the value -1\n"
+        "\t\t\t(default 0; with --fence, --reboot, --unfence)."
+    },
+    {   "as-node-id", no_argument, NULL, 'n',
+        "(Advanced) The supplied node is the corosync node ID\n"
+        "\t\t\t(with --last)."
+    },
+    {   "tolerance", required_argument, NULL,   0,
+        "(Advanced) Do nothing if an equivalent --fence request\n"
+        "\t\t\tsucceeded less than this many seconds earlier\n"
+        "\t\t\t(with --fence, --unfence, --reboot)."
+    },
 
-    {"list-all",    0, 0, 'L', "legacy alias for --list-registered"},
-
-    {0, 0, 0, 0}
+    {   "list-all", no_argument, NULL, 'L', NULL, pcmk_option_hidden },
+    { 0, 0, 0, 0 }
 };
 /* *INDENT-ON* */
 
@@ -106,6 +207,7 @@ struct {
     char *name;
     int timeout;
     int tolerance;
+    int delay;
     int rc;
 } async_fence_data;
 
@@ -170,11 +272,13 @@ async_fence_helper(gpointer user_data)
 
     st->cmds->register_notification(st, T_STONITH_NOTIFY_FENCE, notify_callback);
 
-    call_id = st->cmds->fence(st,
-                              st_opt_allow_suicide,
-                              async_fence_data.target,
-                              async_fence_data.action,
-                              async_fence_data.timeout, async_fence_data.tolerance);
+    call_id = st->cmds->fence_with_delay(st,
+                                         st_opt_allow_suicide,
+                                         async_fence_data.target,
+                                         async_fence_data.action,
+                                         async_fence_data.timeout,
+                                         async_fence_data.tolerance,
+                                         async_fence_data.delay);
 
     if (call_id < 0) {
         g_main_loop_quit(mainloop);
@@ -190,7 +294,8 @@ async_fence_helper(gpointer user_data)
 }
 
 static int
-mainloop_fencing(stonith_t * st, const char *target, const char *action, int timeout, int tolerance)
+mainloop_fencing(stonith_t * st, const char *target, const char *action,
+                 int timeout, int tolerance, int delay)
 {
     crm_trigger_t *trig;
 
@@ -199,6 +304,7 @@ mainloop_fencing(stonith_t * st, const char *target, const char *action, int tim
     async_fence_data.action = action;
     async_fence_data.timeout = timeout;
     async_fence_data.tolerance = tolerance;
+    async_fence_data.delay = delay;
     async_fence_data.rc = -1;
 
     trig = mainloop_add_trigger(G_PRIORITY_HIGH, async_fence_helper, NULL);
@@ -217,9 +323,15 @@ handle_level(stonith_t *st, char *target, int fence_level,
     char *node = NULL;
     char *pattern = NULL;
     char *name = NULL;
-    char *value = strchr(target, '=');
+    char *value = NULL;
+
+    if (target == NULL) {
+        // Not really possible, but makes static analysis happy
+        return -EINVAL;
+    }
 
     /* Determine if targeting by attribute, node name pattern or node name */
+    value = strchr(target, '=');
     if (value != NULL)  {
         name = target;
         *value++ = '\0';
@@ -239,14 +351,52 @@ handle_level(stonith_t *st, char *target, int fence_level,
                                        name, value, fence_level);
 }
 
-static int
-show_history(stonith_t *st, const char *target, int timeout, int quiet,
-             int verbose)
+static char *
+fence_action_str(const char *action)
 {
-    stonith_history_t *history, *hp, *latest = NULL;
+    char *str = NULL;
+
+    if (action == NULL) {
+        str = strdup("unknown");
+    } else if (action[0] == 'o') { // on, off
+        str = crm_concat("turn", action, ' ');
+    } else {
+        str = strdup(action);
+    }
+    return str;
+}
+
+static void
+print_fence_event(stonith_history_t *event)
+{
+    char *action_s = fence_action_str(event->action);
+    time_t complete = event->completed;
+
+    printf("%s was able to %s node %s on behalf of %s from %s at %s\n",
+           (event->delegate? event->delegate : "This node"), action_s,
+           event->target, event->client, event->origin, ctime(&complete));
+    free(action_s);
+}
+
+static int
+handle_history(stonith_t *st, const char *target, int timeout, int quiet,
+             int verbose, int cleanup, int broadcast)
+{
+    stonith_history_t *history = NULL, *hp, *latest = NULL;
     int rc = 0;
 
-    rc = st->cmds->history(st, st_opts,
+
+    if (!quiet) {
+        if (cleanup) {
+            printf("cleaning up fencing-history%s%s\n",
+                   target?" for node ":"", target?target:"");
+        }
+        if (broadcast) {
+            printf("gather fencing-history from all nodes\n");
+        }
+    }
+    rc = st->cmds->history(st, st_opts | (cleanup?st_opt_cleanup:0) |
+                           (broadcast?st_opt_broadcast:0),
                            (safe_str_eq(target, "*")? NULL : target),
                            &history, timeout);
     for (hp = history; hp; hp = hp->next) {
@@ -259,32 +409,23 @@ show_history(stonith_t *st, const char *target, int timeout, int quiet,
 
         if (quiet || !verbose) {
             continue;
-        } else if (hp->action == NULL) {
-            action_s = strdup("unknown");
-        } else if (hp->action[0] != 'r') {
-            action_s = crm_concat("turn", hp->action, ' ');
-        } else {
-            action_s = strdup(hp->action);
         }
 
         if (hp->state == st_failed) {
+            action_s = fence_action_str(hp->action);
             printf("%s failed to %s node %s on behalf of %s from %s at %s\n",
                    hp->delegate ? hp->delegate : "We", action_s, hp->target,
                    hp->client, hp->origin, ctime(&complete));
 
-        } else if (hp->state == st_done && hp->delegate) {
-            printf("%s was able to %s node %s on behalf of %s from %s at %s\n",
-                   hp->delegate, action_s, hp->target,
-                   hp->client, hp->origin, ctime(&complete));
-
         } else if (hp->state == st_done) {
-            printf("We were able to %s node %s on behalf of %s from %s at %s\n",
-                   action_s, hp->target, hp->client, hp->origin, ctime(&complete));
+            print_fence_event(latest);
+
         } else {
             /* ocf:pacemaker:controld depends on "wishes to" being
              * in this output, when used with older versions of DLM
              * that don't report stateful_merge_wait
              */
+            action_s = fence_action_str(hp->action);
             printf("%s at %s wishes to %s node %s - %d %d\n",
                    hp->client, hp->origin, action_s, hp->target, hp->state, hp->completed);
         }
@@ -295,37 +436,74 @@ show_history(stonith_t *st, const char *target, int timeout, int quiet,
     if (latest) {
         if (quiet) {
             printf("%d\n", latest->completed);
-        } else {
-            char *action_s = NULL;
-            time_t complete = latest->completed;
+        } else if (!verbose) { // already printed if verbose
+            print_fence_event(latest);
+        }
+    }
 
-            if (latest->action == NULL) {
-                action_s = strdup("unknown");
-            } else if (latest->action[0] != 'r') {
-                action_s = crm_concat("turn", latest->action, ' ');
-            } else {
-                action_s = strdup(latest->action);
-            }
+    stonith_history_free(history);
+    return rc;
+}
 
-            printf("%s was able to %s node %s on behalf of %s from %s at %s\n",
-                   latest->delegate ? latest->delegate : "We", action_s, latest->target,
-                   latest->client, latest->origin, ctime(&complete));
+static int
+validate(stonith_t *st, const char *agent, const char *id,
+         stonith_key_value_t *params, int timeout, int quiet)
+{
+    int rc = 1;
+    char *output = NULL;
+    char *error_output = NULL;
 
-            free(action_s);
+    rc = st->cmds->validate(st, st_opt_sync_call, id, NULL, agent, params,
+                            timeout, &output, &error_output);
+
+    if (!quiet) {
+        printf("Validation of %s %s\n", agent, (rc? "failed" : "succeeded"));
+        if (output && *output) {
+            puts(output);
+            free(output);
+        }
+        if (error_output && *error_output) {
+            puts(error_output);
+            free(error_output);
         }
     }
     return rc;
 }
 
+static void
+show_last_fenced(int as_nodeid, const char *target)
+{
+    time_t when = 0;
+
+    if (target == NULL) {
+        // Not really possible, but makes static analysis happy
+        return;
+    }
+    if (as_nodeid) {
+        uint32_t nodeid = atol(target);
+        when = stonith_api_time(nodeid, NULL, FALSE);
+    } else {
+        when = stonith_api_time(0, target, FALSE);
+    }
+    if(when) {
+        printf("Node %s last kicked at: %s\n", target, ctime(&when));
+    } else {
+        printf("Node %s has never been kicked\n", target);
+    }
+}
+
 int
 main(int argc, char **argv)
 {
     int flag;
     int rc = 0;
     int quiet = 0;
+    int cleanup = 0;
+    int broadcast = 0;
     int verbose = 0;
     int argerr = 0;
     int timeout = 120;
+    int delay = 0;
     int option_index = 0;
     int fence_level = 0;
     int no_connect = 0;
@@ -335,20 +513,21 @@ main(int argc, char **argv)
     char *name = NULL;
     char *value = NULL;
     char *target = NULL;
+    char *lists = NULL;
     const char *agent = NULL;
     const char *device = NULL;
     const char *longname = NULL;
 
     char action = 0;
+    int exit_code = 0;
     stonith_t *st = NULL;
     stonith_key_value_t *params = NULL;
     stonith_key_value_t *devices = NULL;
     stonith_key_value_t *dIter = NULL;
 
     crm_log_cli_init("stonith_admin");
-    crm_set_options(NULL, "mode [options]", long_options,
-                    "Provides access to the stonith-ng API.\n"
-                    "\nAllows the administrator to add/remove/list devices, check device and host status and fence hosts\n");
+    crm_set_options(NULL, "<command> [<options>]", long_options,
+                    "access the Pacemaker fencing API");
 
     async_fence_data.name = strdup(crm_system_name);
 
@@ -367,6 +546,7 @@ main(int argc, char **argv)
                 crm_help(flag, EX_OK);
                 break;
             case 'I':
+            case 'K':
                 no_connect = 1;
                 /* fall through */
             case 'L':
@@ -375,9 +555,16 @@ main(int argc, char **argv)
             case 'q':
                 quiet = 1;
                 break;
+            case 'c':
+                cleanup = 1;
+                break;
+            case 'b':
+                broadcast = 1;
+                break;
             case 'Q':
             case 'R':
             case 'D':
+            case 's':
                 action = flag;
                 device = optarg;
                 break;
@@ -399,6 +586,9 @@ main(int argc, char **argv)
             case 't':
                 timeout = crm_atoi(optarg, NULL);
                 break;
+            case 'y':
+                delay = crm_atoi(optarg, NULL);
+                break;
             case 'B':
             case 'F':
             case 'U':
@@ -452,6 +642,7 @@ main(int argc, char **argv)
                         crm_info("Got: '%s'='%s'", optarg, env);
                         params = stonith_key_value_add(params, optarg, env);
                     }
+                    free(key);
                 }
                 break;
             case 0:
@@ -469,21 +660,22 @@ main(int argc, char **argv)
         ++argerr;
     }
 
+    if (action == 0) {
+        ++argerr;
+    }
+
     if (argerr) {
         crm_help('?', EX_USAGE);
     }
 
-    crm_debug("Create");
     st = stonith_api_new();
-    crm_debug("Created");
 
     if (!no_connect) {
-        crm_debug("Connecting as %s", async_fence_data.name);
         rc = st->cmds->connect(st, async_fence_data.name, NULL);
-
-        crm_debug("Connect: %d", rc);
-
         if (rc < 0) {
+            fprintf(stderr, "Could not connect to fencer: %s\n",
+                    pcmk_strerror(rc));
+            exit_code = rc;
             goto done;
         }
     }
@@ -522,8 +714,40 @@ main(int argc, char **argv)
                 rc = st->cmds->list(st, st_opts, device, NULL, timeout);
             }
             break;
+        case 's':
+            rc = st->cmds->list(st, st_opts, device, &lists, timeout);
+            if (rc == 0) {
+                if (lists) {
+                    char *source = lists, *dest = lists; 
+
+                    while (*dest) {
+                        if ((*dest == '\\') && (*(dest+1) == 'n')) {
+                            *source = '\n';
+                            dest++;
+                            dest++;
+                            source++;
+                        } else if ((*dest == ',') || (*dest == ';')) {
+                            dest++;
+                        } else {
+                            *source = *dest;
+                            dest++;
+                            source++;
+                        }
+
+                        if (!(*dest)) {
+                            *source = 0;
+                        }
+                    }
+                    fprintf(stdout, "%s", lists);
+                    free(lists);
+                }
+            } else {
+                fprintf(stderr, "List command returned error. rc : %d\n", rc);
+            }
+            break;
         case 'R':
-            rc = st->cmds->register_device(st, st_opts, device, "stonith-ng", agent, params);
+            rc = st->cmds->register_device(st, st_opts, device, NULL, agent,
+                                           params);
             break;
         case 'D':
             rc = st->cmds->remove_device(st, st_opts, device);
@@ -535,7 +759,8 @@ main(int argc, char **argv)
         case 'M':
             if (agent == NULL) {
                 printf("Please specify an agent to query using -a,--agent [value]\n");
-                return -1;
+                exit_code = -1;
+                goto done;
             } else {
                 char *buffer = NULL;
 
@@ -550,49 +775,39 @@ main(int argc, char **argv)
             rc = st->cmds->confirm(st, st_opts, target);
             break;
         case 'B':
-            rc = mainloop_fencing(st, target, "reboot", timeout, tolerance);
+            rc = mainloop_fencing(st, target, "reboot", timeout, tolerance, delay);
             break;
         case 'F':
-            rc = mainloop_fencing(st, target, "off", timeout, tolerance);
+            rc = mainloop_fencing(st, target, "off", timeout, tolerance, delay);
             break;
         case 'U':
-            rc = mainloop_fencing(st, target, "on", timeout, tolerance);
+            rc = mainloop_fencing(st, target, "on", timeout, tolerance, delay);
             break;
         case 'h':
-            {
-                time_t when = 0;
-
-                if(as_nodeid) {
-                    uint32_t nodeid = atol(target);
-                    when = stonith_api_time(nodeid, NULL, FALSE);
-                } else {
-                    when = stonith_api_time(0, target, FALSE);
-                }
-                if(when) {
-                    printf("Node %s last kicked at: %s\n", target, ctime(&when));
-                } else {
-                    printf("Node %s has never been kicked\n", target);
-                }
-            }
+            show_last_fenced(as_nodeid, target);
             break;
         case 'H':
-            rc = show_history(st, target, timeout, quiet, verbose);
+            rc = handle_history(st, target, timeout, quiet,
+                                verbose, cleanup, broadcast);
+            break;
+        case 'K':
+            if (agent == NULL) {
+                printf("Please specify an agent to validate with --agent\n");
+                exit_code = -1;
+                goto done;
+            }
+            device = (devices? devices->key : NULL);
+            rc = validate(st, agent, device, params, timeout, quiet);
             break;
     }
 
+    crm_info("Command returned: %s (%d)", pcmk_strerror(rc), rc);
+    exit_code = rc;
+
   done:
     free(async_fence_data.name);
-    crm_info("Command returned: %s (%d)", pcmk_strerror(rc), rc);
-    if (rc < 0) {
-        printf("Command failed: %s\n", pcmk_strerror(rc));
-    }
-
     stonith_key_value_freeall(params, 1, 1);
     st->cmds->disconnect(st);
-    crm_debug("Disconnect: %d", rc);
-
-    crm_debug("Destroy");
     stonith_api_delete(st);
-
-    return rc;
+    return exit_code;
 }
