@@ -82,7 +82,7 @@ typedef struct st_query_result_s {
     GHashTable *devices;
 } st_query_result_t;
 
-GHashTable *remote_op_list = NULL;
+GHashTable *stonith_remote_op_list = NULL;
 void call_remote_stonith(remote_fencing_op_t * op, st_query_result_t * peer);
 static void remote_op_done(remote_fencing_op_t * op, xmlNode * data, int rc, int dup);
 extern xmlNode *stonith_create_op(int call_id, const char *token, const char *op, xmlNode * data,
@@ -111,6 +111,74 @@ free_remote_query(gpointer data)
     }
 }
 
+void
+free_stonith_remote_op_list()
+{
+    if (stonith_remote_op_list != NULL) {
+        g_hash_table_destroy(stonith_remote_op_list);
+        stonith_remote_op_list = NULL;
+    }
+}
+
+static void
+clear_remote_op_timers(remote_fencing_op_t * op)
+{
+    if (op->query_timer) {
+        g_source_remove(op->query_timer);
+        op->query_timer = 0;
+    }
+    if (op->op_timer_total) {
+        g_source_remove(op->op_timer_total);
+        op->op_timer_total = 0;
+    }
+    if (op->op_timer_one) {
+        g_source_remove(op->op_timer_one);
+        op->op_timer_one = 0;
+    }
+}
+
+static void
+free_remote_op(gpointer data)
+{
+    remote_fencing_op_t *op = data;
+
+    crm_trace("Free'ing op %s for %s", op->id, op->target);
+    crm_log_xml_debug(op->request, "Destroying");
+
+    clear_remote_op_timers(op);
+
+    free(op->id);
+    free(op->action);
+    free(op->delegate);
+    free(op->target);
+    free(op->client_id);
+    free(op->client_name);
+    free(op->originator);
+
+    if (op->query_results) {
+        g_list_free_full(op->query_results, free_remote_query);
+    }
+    if (op->request) {
+        free_xml(op->request);
+        op->request = NULL;
+    }
+    if (op->devices_list) {
+        g_list_free_full(op->devices_list, free);
+        op->devices_list = NULL;
+    }
+    g_list_free_full(op->automatic_list, free);
+    g_list_free(op->duplicates);
+    free(op);
+}
+
+void
+init_stonith_remote_op_hash_table(GHashTable **table)
+{
+    if (*table == NULL) {
+        *table = g_hash_table_new_full(crm_str_hash, g_str_equal, NULL, free_remote_op);
+    }
+}
+
 struct peer_count_data {
     const remote_fencing_op_t *op;
     gboolean verified_only;
@@ -209,55 +277,6 @@ grab_peer_device(const remote_fencing_op_t *op, st_query_result_t *peer,
     return TRUE;
 }
 
-static void
-clear_remote_op_timers(remote_fencing_op_t * op)
-{
-    if (op->query_timer) {
-        g_source_remove(op->query_timer);
-        op->query_timer = 0;
-    }
-    if (op->op_timer_total) {
-        g_source_remove(op->op_timer_total);
-        op->op_timer_total = 0;
-    }
-    if (op->op_timer_one) {
-        g_source_remove(op->op_timer_one);
-        op->op_timer_one = 0;
-    }
-}
-
-static void
-free_remote_op(gpointer data)
-{
-    remote_fencing_op_t *op = data;
-
-    crm_trace("Free'ing op %s for %s", op->id, op->target);
-    crm_log_xml_debug(op->request, "Destroying");
-
-    clear_remote_op_timers(op);
-
-    free(op->id);
-    free(op->action);
-    free(op->target);
-    free(op->client_id);
-    free(op->client_name);
-    free(op->originator);
-
-    if (op->query_results) {
-        g_list_free_full(op->query_results, free_remote_query);
-    }
-    if (op->request) {
-        free_xml(op->request);
-        op->request = NULL;
-    }
-    if (op->devices_list) {
-        g_list_free_full(op->devices_list, free);
-        op->devices_list = NULL;
-    }
-    g_list_free_full(op->automatic_list, free);
-    free(op);
-}
-
 /*!
  * \internal
  * \brief Return an operation's originally requested action (before any remap)
@@ -281,7 +300,7 @@ op_requested_action(const remote_fencing_op_t *op)
 static void
 op_phase_off(remote_fencing_op_t *op)
 {
-    crm_info("Remapping multiple-device reboot of %s (%s) to off",
+    crm_info("Remapping multiple-device reboot targeting %s (%s) to 'off'",
              op->target, op->id);
     op->phase = st_phase_off;
 
@@ -302,7 +321,8 @@ op_phase_on(remote_fencing_op_t *op)
 {
     GListPtr iter = NULL;
 
-    crm_info("Remapped off of %s complete, remapping to on for %s.%.8s",
+    crm_info("Remapped 'off' targeting %s complete, "
+             "remapping to 'on' for %s.%.8s",
              op->target, op->client_name, op->id);
     op->phase = st_phase_on;
     strcpy(op->action, "on");
@@ -335,7 +355,7 @@ static void
 undo_op_remap(remote_fencing_op_t *op)
 {
     if (op->phase > 0) {
-        crm_info("Undoing remap of reboot of %s for %s.%.8s",
+        crm_info("Undoing remap of reboot targeting %s for %s.%.8s",
                  op->target, op->client_name, op->id);
         op->phase = st_phase_requested;
         strcpy(op->action, "reboot");
@@ -360,8 +380,8 @@ create_op_done_notify(remote_fencing_op_t * op, int rc)
     return notify_data;
 }
 
-static void
-bcast_result_to_peers(remote_fencing_op_t * op, int rc)
+void
+stonith_bcast_result_to_peers(remote_fencing_op_t * op, int rc)
 {
     static int count = 0;
     xmlNode *bcast = create_xml_node(NULL, T_STONITH_REPLY);
@@ -406,6 +426,7 @@ handle_local_reply_and_notify(remote_fencing_op_t * op, xmlNode * data, int rc)
 
     /* bcast to all local clients that the fencing operation happend */
     do_stonith_notify(0, T_STONITH_NOTIFY_FENCE, rc, notify_data);
+    do_stonith_notify(0, T_STONITH_NOTIFY_HISTORY, 0, NULL);
 
     /* mark this op as having notify's already sent */
     op->notify_sent = TRUE;
@@ -472,9 +493,12 @@ remote_op_done(remote_fencing_op_t * op, xmlNode * data, int rc, int dup)
     undo_op_remap(op);
 
     if (op->notify_sent == TRUE) {
-        crm_err("Already sent notifications for '%s of %s by %s' (for=%s@%s.%.8s, state=%d): %s",
-                op->action, op->target, op->delegate ? op->delegate : "<no-one>",
-                op->client_name, op->originator, op->id, op->state, pcmk_strerror(rc));
+        crm_err("Already sent notifications for '%s' targeting %s on %s for "
+                "client %s@%s.%.8s: %s " CRM_XS " rc=%d state=%d",
+                op->action, op->target,
+                (op->delegate? op->delegate : "unknown node"),
+                op->client_name, op->originator, op->id, pcmk_strerror(rc),
+                rc, op->state);
         goto remote_op_done_cleanup;
     }
 
@@ -498,7 +522,7 @@ remote_op_done(remote_fencing_op_t * op, xmlNode * data, int rc, int dup)
     subt = crm_element_value(data, F_SUBTYPE);
     if (dup == FALSE && safe_str_neq(subt, "broadcast")) {
         /* Defer notification until the bcast message arrives */
-        bcast_result_to_peers(op, rc);
+        stonith_bcast_result_to_peers(op, rc);
         goto remote_op_done_cleanup;
     }
 
@@ -508,9 +532,10 @@ remote_op_done(remote_fencing_op_t * op, xmlNode * data, int rc, int dup)
         level = LOG_NOTICE;
     }
 
-    do_crm_log(level,
-               "Operation %s of %s by %s for %s@%s.%.8s: %s",
-               op->action, op->target, op->delegate ? op->delegate : "<no-one>",
+    do_crm_log(level, "Operation '%s'%s%s on %s for %s@%s.%.8s: %s",
+               op->action, (op->target? " targeting " : ""),
+               (op->target? op->target : ""),
+               (op->delegate? op->delegate : "<no-one>"),
                op->client_name, op->originator, op->id, pcmk_strerror(rc));
 
     handle_local_reply_and_notify(op, data, rc);
@@ -557,8 +582,8 @@ remote_op_timeout_one(gpointer userdata)
 
     op->op_timer_one = 0;
 
-    crm_notice("Peer's fencing (%s) of %s for %s timed out" CRM_XS "id=%s",
-               op->action, op->target, op->client_name, op->id);
+    crm_notice("Peer's '%s' action targeting %s for client %s timed out " CRM_XS
+               " id=%s", op->action, op->target, op->client_name, op->id);
     call_remote_stonith(op, NULL);
     return FALSE;
 }
@@ -571,13 +596,15 @@ remote_op_timeout(gpointer userdata)
     op->op_timer_total = 0;
 
     if (op->state == st_done) {
-        crm_debug("Action %s (%s) for %s (%s) already completed",
-                  op->action, op->id, op->target, op->client_name);
+        crm_debug("Action '%s' targeting %s for client %s already completed "
+                  CRM_XS " id=%s",
+                  op->action, op->target, op->client_name, op->id);
         return FALSE;
     }
 
-    crm_debug("Action %s (%s) for %s (%s) timed out",
-              op->action, op->id, op->target, op->client_name);
+    crm_debug("Action '%s' targeting %s for client %s timed out "
+              CRM_XS " id=%s",
+              op->action, op->target, op->client_name, op->id);
 
     if (op->phase == st_phase_on) {
         /* A remapped reboot operation timed out in the "on" phase, but the
@@ -602,14 +629,18 @@ remote_op_query_timeout(gpointer data)
 
     op->query_timer = 0;
     if (op->state == st_done) {
-        crm_debug("Operation %s for %s already completed", op->id, op->target);
+        crm_debug("Operation %s targeting %s already completed",
+                  op->id, op->target);
     } else if (op->state == st_exec) {
-        crm_debug("Operation %s for %s already in progress", op->id, op->target);
+        crm_debug("Operation %s targeting %s already in progress",
+                  op->id, op->target);
     } else if (op->query_results) {
-        crm_debug("Query %s for %s complete: %d", op->id, op->target, op->state);
+        crm_debug("Query %s targeting %s complete (state=%d)",
+                  op->id, op->target, op->state);
         call_remote_stonith(op, NULL);
     } else {
-        crm_debug("Query %s for %s timed out: %d", op->id, op->target, op->state);
+        crm_debug("Query %s targeting %s timed out (state=%d)",
+                  op->id, op->target, op->state);
         if (op->op_timer_total) {
             g_source_remove(op->op_timer_total);
             op->op_timer_total = 0;
@@ -798,11 +829,17 @@ stonith_topology_next(remote_fencing_op_t * op)
     } while (op->level < ST_LEVEL_MAX && tp->levels[op->level] == NULL);
 
     if (op->level < ST_LEVEL_MAX) {
-        crm_trace("Attempting fencing level %d for %s (%d devices) - %s@%s.%.8s",
+        crm_trace("Attempting fencing level %d targeting %s (%d devices) "
+                  "for client %s@%s.%.8s",
                   op->level, op->target, g_list_length(tp->levels[op->level]),
                   op->client_name, op->originator, op->id);
         set_op_device_list(op, tp->levels[op->level]);
 
+        // The requested delay has been applied for the first fencing level
+        if (op->level > 1 && op->delay > 0) {
+            op->delay = 0;
+        }
+
         if (g_list_next(op->devices_list) && safe_str_eq(op->action, "reboot")) {
             /* A reboot has been requested for a topology level with multiple
              * devices. Instead of rebooting the devices sequentially, we will
@@ -814,7 +851,7 @@ stonith_topology_next(remote_fencing_op_t * op)
         return pcmk_ok;
     }
 
-    crm_notice("All fencing options to fence %s for %s@%s.%.8s failed",
+    crm_notice("All fencing options targeting %s for client %s@%s.%.8s failed",
                op->target, op->client_name, op->originator, op->id);
     return -EINVAL;
 }
@@ -832,7 +869,7 @@ merge_duplicates(remote_fencing_op_t * op)
 
     time_t now = time(NULL);
 
-    g_hash_table_iter_init(&iter, remote_op_list);
+    g_hash_table_iter_init(&iter, stonith_remote_op_list);
     while (g_hash_table_iter_next(&iter, NULL, (void **)&other)) {
         crm_node_t *peer = NULL;
         const char *other_action = op_requested_action(other);
@@ -857,13 +894,15 @@ merge_duplicates(remote_fencing_op_t * op)
 
         peer = crm_get_peer(0, other->originator);
         if(fencing_peer_active(peer) == FALSE) {
-            crm_notice("Failing stonith action %s for node %s originating from %s@%s.%.8s: Originator is dead",
+            crm_notice("Failing action '%s' targeting %s originating from "
+                       "client %s@%s.%.8s: Originator is dead",
                        other->action, other->target, other->client_name, other->originator, other->id);
             other->state = st_failed;
             continue;
 
         } else if(other->total_timeout > 0 && now > (other->total_timeout + other->created)) {
-            crm_info("Stonith action %s for node %s originating from %s@%s.%.8s is too old: %d vs. %d + %d",
+            crm_info("Action '%s' targeting %s originating from client "
+                     "%s@%s.%.8s is too old: %ld vs. %ld + %d",
                      other->action, other->target, other->client_name, other->originator, other->id,
                      now, other->created, other->total_timeout);
             continue;
@@ -878,10 +917,11 @@ merge_duplicates(remote_fencing_op_t * op)
             other->total_timeout = op->total_timeout =
                 TIMEOUT_MULTIPLY_FACTOR * get_op_total_timeout(op, NULL);
         }
-        crm_notice
-            ("Merging stonith action %s for node %s originating from client %s.%.8s with identical request from %s@%s.%.8s (%ds)",
-             op->action, op->target, op->client_name, op->id, other->client_name, other->originator,
-             other->id, other->total_timeout);
+        crm_notice("Merging stonith action '%s' targeting %s originating from "
+                   "client %s.%.8s with identical request from %s@%s.%.8s (%ds)",
+                   op->action, op->target, op->client_name, op->id,
+                   other->client_name, other->originator, other->id,
+                   other->total_timeout);
         report_timeout_period(op, other->total_timeout);
         op->state = st_duplicate;
     }
@@ -920,30 +960,6 @@ stonith_manual_ack(xmlNode * msg, remote_fencing_op_t * op)
     return -EINPROGRESS;
 }
 
-char *
-stonith_get_peer_name(unsigned int nodeid)
-{
-    crm_node_t *node = crm_find_peer(nodeid, NULL);
-    char *nodename = NULL;
-
-    if (node && node->uname) {
-        return strdup(node->uname);
-
-    } else if ((nodename = get_node_name(nodeid))) {
-        return nodename;
-
-    } else {
-        const char *last_known_name = g_hash_table_lookup(known_peer_names, GUINT_TO_POINTER(nodeid));
-
-        if (last_known_name) {
-            crm_debug("Use the last known name %s for nodeid %u", last_known_name, nodeid);
-            return strdup(last_known_name);
-        }
-    }
-
-    return NULL;
-}
-
 /*!
  * \internal
  * \brief Create a new remote stonith operation
@@ -961,10 +977,9 @@ create_remote_stonith_op(const char *client, xmlNode * request, gboolean peer)
     remote_fencing_op_t *op = NULL;
     xmlNode *dev = get_xpath_object("//@" F_STONITH_TARGET, request, LOG_TRACE);
     int call_options = 0;
+    const char *operation = NULL;
 
-    if (remote_op_list == NULL) {
-        remote_op_list = g_hash_table_new_full(crm_str_hash, g_str_equal, NULL, free_remote_op);
-    }
+    init_stonith_remote_op_hash_table(&stonith_remote_op_list);
 
     /* If this operation is owned by another node, check to make
      * sure we haven't already created this operation. */
@@ -973,7 +988,7 @@ create_remote_stonith_op(const char *client, xmlNode * request, gboolean peer)
 
         CRM_CHECK(op_id != NULL, return NULL);
 
-        op = g_hash_table_lookup(remote_op_list, op_id);
+        op = g_hash_table_lookup(stonith_remote_op_list, op_id);
         if (op) {
             crm_debug("%s already exists", op_id);
             return op;
@@ -983,6 +998,8 @@ create_remote_stonith_op(const char *client, xmlNode * request, gboolean peer)
     op = calloc(1, sizeof(remote_fencing_op_t));
 
     crm_element_value_int(request, F_STONITH_TIMEOUT, &(op->base_timeout));
+    // Value -1 means disable any static/random fencing delays
+    crm_element_value_int(request, F_STONITH_DELAY, &(op->delay));
 
     if (peer && dev) {
         op->id = crm_element_value_copy(dev, F_STONITH_REMOTE_OP_ID);
@@ -990,8 +1007,8 @@ create_remote_stonith_op(const char *client, xmlNode * request, gboolean peer)
         op->id = crm_generate_uuid();
     }
 
-    g_hash_table_replace(remote_op_list, op->id, op);
-    CRM_LOG_ASSERT(g_hash_table_lookup(remote_op_list, op->id) != NULL);
+    g_hash_table_replace(stonith_remote_op_list, op->id, op);
+    CRM_LOG_ASSERT(g_hash_table_lookup(stonith_remote_op_list, op->id) != NULL);
     crm_trace("Created %s", op->id);
 
     op->state = st_query;
@@ -1011,7 +1028,15 @@ create_remote_stonith_op(const char *client, xmlNode * request, gboolean peer)
         op->client_id = strdup(client);
     }
 
-    op->client_name = crm_element_value_copy(request, F_STONITH_CLIENTNAME);
+    /* For a RELAY operation, set fenced on the client. */
+    operation = crm_element_value(request, F_STONITH_OPERATION);
+
+    if (crm_str_eq(operation, STONITH_OP_RELAY, TRUE)) {
+        op->client_name = crm_strdup_printf("%s.%lu", crm_system_name,
+                                         (unsigned long) getpid());
+    } else {
+        op->client_name = crm_element_value_copy(request, F_STONITH_CLIENTNAME);
+    }
 
     op->target = crm_element_value_copy(dev, F_STONITH_TARGET);
     op->request = copy_xml(request);    /* TODO: Figure out how to avoid this */
@@ -1020,20 +1045,20 @@ create_remote_stonith_op(const char *client, xmlNode * request, gboolean peer)
 
     crm_element_value_int(request, F_STONITH_CALLID, &(op->client_callid));
 
-    crm_trace("%s new stonith op: %s - %s of %s for %s",
-              (peer
-               && dev) ? "Recorded" : "Generated", op->id, op->action, op->target, op->client_name);
+    crm_trace("%s new stonith op %s ('%s' targeting %s for client %s)",
+              (peer && dev)? "Recorded" : "Generated", op->id, op->action,
+              op->target, op->client_name);
 
     if (op->call_options & st_opt_cs_nodeid) {
         int nodeid = crm_atoi(op->target, NULL);
-        char *nodename = stonith_get_peer_name(nodeid);
+        crm_node_t *node = crm_find_known_peer_full(nodeid, NULL, CRM_GET_PEER_ANY);
 
         /* Ensure the conversion only happens once */
         op->call_options &= ~st_opt_cs_nodeid;
 
-        if (nodename) {
+        if (node && node->uname) {
             free(op->target);
-            op->target = nodename;
+            op->target = strdup(node->uname);
 
         } else {
             crm_warn("Could not expand nodeid '%s' into a host name", op->target);
@@ -1043,6 +1068,14 @@ create_remote_stonith_op(const char *client, xmlNode * request, gboolean peer)
     /* check to see if this is a duplicate operation of another in-flight operation */
     merge_duplicates(op);
 
+    if (op->state != st_duplicate) {
+        /* kick history readers */
+        do_stonith_notify(0, T_STONITH_NOTIFY_HISTORY, 0, NULL);
+    }
+
+    /* safe to trim as long as that doesn't touch pending ops */
+    stonith_fence_history_trim();
+
     return op;
 }
 
@@ -1053,6 +1086,8 @@ initiate_remote_stonith_op(crm_client_t * client, xmlNode * request, gboolean ma
     xmlNode *query = NULL;
     const char *client_id = NULL;
     remote_fencing_op_t *op = NULL;
+    const char *relay_op_id = NULL;
+    const char *operation = NULL;
 
     if (client) {
         client_id = client->id;
@@ -1077,18 +1112,18 @@ initiate_remote_stonith_op(crm_client_t * client, xmlNode * request, gboolean ma
 
     switch (op->state) {
         case st_failed:
-            crm_warn("Could not request peer fencing (%s) of %s "
+            crm_warn("Could not request peer fencing (%s) targeting %s "
                      CRM_XS " id=%s", op->action, op->target, op->id);
             remote_op_done(op, NULL, -EINVAL, FALSE);
             return op;
 
         case st_duplicate:
-            crm_info("Requesting peer fencing (%s) of %s (duplicate) "
+            crm_info("Requesting peer fencing (%s) targeting %s (duplicate) "
                      CRM_XS " id=%s", op->action, op->target, op->id);
             return op;
 
         default:
-            crm_notice("Requesting peer fencing (%s) of %s "
+            crm_notice("Requesting peer fencing (%s) targeting %s "
                        CRM_XS " id=%s state=%d",
                        op->action, op->target, op->id, op->state);
     }
@@ -1104,6 +1139,15 @@ initiate_remote_stonith_op(crm_client_t * client, xmlNode * request, gboolean ma
     crm_xml_add(query, F_STONITH_CLIENTNAME, op->client_name);
     crm_xml_add_int(query, F_STONITH_TIMEOUT, op->base_timeout);
 
+    /* In case of RELAY operation, RELAY information is added to the query to delete the original operation of RELAY. */
+    operation = crm_element_value(request, F_STONITH_OPERATION);
+    if (crm_str_eq(operation, STONITH_OP_RELAY, TRUE)) {
+        relay_op_id = crm_element_value(request, F_STONITH_REMOTE_OP_ID);
+        if (relay_op_id) {
+            crm_xml_add(query, F_STONITH_REMOTE_OP_ID_RELAY, relay_op_id);
+        }
+    }
+
     send_cluster_message(NULL, crm_msg_stonith_ng, query, FALSE);
     free_xml(query);
 
@@ -1135,7 +1179,7 @@ find_best_peer(const char *device, remote_fencing_op_t * op, enum find_best_peer
     for (iter = op->query_results; iter != NULL; iter = iter->next) {
         st_query_result_t *peer = iter->data;
 
-        crm_trace("Testing result from %s for %s with %d devices: %d %x",
+        crm_trace("Testing result from %s targeting %s with %d devices: %d %x",
                   peer->host, op->target, peer->ndevices, peer->tried, options);
         if ((options & FIND_PEER_SKIP_TARGET) && safe_str_eq(peer->host, op->target)) {
             continue;
@@ -1421,12 +1465,19 @@ advance_op_topology(remote_fencing_op_t *op, const char *device, xmlNode *msg,
 
     if (op->devices) {
         /* Necessary devices remain, so execute the next one */
-        crm_trace("Next for %s on behalf of %s@%s (rc was %d)",
+        crm_trace("Next targeting %s on behalf of %s@%s (rc was %d)",
                   op->target, op->originator, op->client_name, rc);
+
+        // The requested delay has been applied for the first device
+        if (op->delay > 0) {
+            op->delay = 0;
+        }
+
         call_remote_stonith(op, NULL);
     } else {
         /* We're done with all devices and phases, so finalize operation */
-        crm_trace("Marking complex fencing op for %s as complete", op->target);
+        crm_trace("Marking complex fencing op targeting %s as complete",
+                  op->target);
         op->state = st_done;
         remote_op_done(op, msg, rc, FALSE);
     }
@@ -1449,7 +1500,7 @@ call_remote_stonith(remote_fencing_op_t * op, st_query_result_t * peer)
         op->total_timeout = TIMEOUT_MULTIPLY_FACTOR * total_timeout;
         op->op_timer_total = g_timeout_add(1000 * op->total_timeout, remote_op_timeout, op);
         report_timeout_period(op, op->total_timeout);
-        crm_info("Total timeout set to %d for peer's fencing of %s for %s"
+        crm_info("Total timeout set to %d for peer's fencing targeting %s for %s"
                  CRM_XS "id=%s",
                  total_timeout, op->target, op->client_name, op->id);
     }
@@ -1476,21 +1527,25 @@ call_remote_stonith(remote_fencing_op_t * op, st_query_result_t * peer)
         crm_xml_add(remote_op, F_STONITH_CLIENTNAME, op->client_name);
         crm_xml_add_int(remote_op, F_STONITH_TIMEOUT, timeout);
         crm_xml_add_int(remote_op, F_STONITH_CALLOPTS, op->call_options);
+        crm_xml_add_int(remote_op, F_STONITH_DELAY, op->delay);
 
         if (device) {
             timeout_one = TIMEOUT_MULTIPLY_FACTOR *
                           get_device_timeout(op, peer, device);
-            crm_info("Requesting that '%s' perform op '%s %s' with '%s' for %s (%ds)", peer->host,
-                     op->target, op->action, device, op->client_name, timeout_one);
+            crm_notice("Requesting that %s perform '%s' action targeting %s "
+                       "using '%s' " CRM_XS " for client %s (%ds)",
+                       peer->host, op->action, op->target, device,
+                       op->client_name, timeout_one);
             crm_xml_add(remote_op, F_STONITH_DEVICE, device);
             crm_xml_add(remote_op, F_STONITH_MODE, "slave");
 
         } else {
             timeout_one = TIMEOUT_MULTIPLY_FACTOR * get_peer_timeout(op, peer);
-            crm_info("Requesting that '%s' perform op '%s %s' for %s (%ds, %ds)",
-                     peer->host, op->target, op->action, op->client_name, timeout_one, stonith_watchdog_timeout_ms);
+            crm_notice("Requesting that %s perform '%s' action targeting %s "
+                       CRM_XS " for client %s (%ds, %lds)",
+                       peer->host, op->action, op->target, op->client_name,
+                       timeout_one, stonith_watchdog_timeout_ms);
             crm_xml_add(remote_op, F_STONITH_MODE, "smart");
-
         }
 
         op->state = st_exec;
@@ -1499,18 +1554,18 @@ call_remote_stonith(remote_fencing_op_t * op, st_query_result_t * peer)
         }
 
         if(stonith_watchdog_timeout_ms > 0 && device && safe_str_eq(device, "watchdog")) {
-            crm_notice("Waiting %ds for %s to self-fence (%s) for %s.%.8s (%p)",
-                       stonith_watchdog_timeout_ms/1000, op->target,
-                       op->action, op->client_name, op->id, device);
+            crm_notice("Waiting %lds for %s to self-fence (%s) for client %s.%.8s",
+                       stonith_watchdog_timeout_ms/1000, op->target, op->action,
+                       op->client_name, op->id);
             op->op_timer_one = g_timeout_add(stonith_watchdog_timeout_ms, remote_op_watchdog_done, op);
 
             /* TODO check devices to verify watchdog will be in use */
         } else if(stonith_watchdog_timeout_ms > 0
                   && safe_str_eq(peer->host, op->target)
                   && safe_str_neq(op->action, "on")) {
-            crm_notice("Waiting %ds for %s to self-fence (%s) for %s.%.8s (%p)",
-                       stonith_watchdog_timeout_ms/1000, op->target,
-                       op->action, op->client_name, op->id, device);
+            crm_notice("Waiting %lds for %s to self-fence (%s) for client %s.%.8s",
+                       stonith_watchdog_timeout_ms/1000, op->target, op->action,
+                       op->client_name, op->id);
             op->op_timer_one = g_timeout_add(stonith_watchdog_timeout_ms, remote_op_watchdog_done, op);
 
         } else {
@@ -1527,19 +1582,20 @@ call_remote_stonith(remote_fencing_op_t * op, st_query_result_t * peer)
         /* A remapped "on" cannot be executed, but the node was already
          * turned off successfully, so ignore the error and continue.
          */
-        crm_warn("Ignoring %s 'on' failure (no capable peers) for %s after successful 'off'",
-                 device, op->target);
+        crm_warn("Ignoring %s 'on' failure (no capable peers) targeting %s "
+                 "after successful 'off'", device, op->target);
         advance_op_topology(op, device, NULL, pcmk_ok);
         return;
 
     } else if (op->owner == FALSE) {
-        crm_err("Fencing (%s) of %s for %s is not ours to control",
+        crm_err("Fencing (%s) targeting %s for client %s is not ours to control",
                 op->action, op->target, op->client_name);
 
     } else if (op->query_timer == 0) {
         /* We've exhausted all available peers */
-        crm_info("No remaining peers capable of fencing (%s) %s for %s (%d)",
-                 op->target, op->action, op->client_name, op->state);
+        crm_info("No remaining peers capable of fencing (%s) %s for client %s "
+                 CRM_XS " state=%d",
+                 op->action, op->target, op->client_name, op->state);
         CRM_LOG_ASSERT(op->state < st_done);
         remote_op_timeout(op);
 
@@ -1551,35 +1607,36 @@ call_remote_stonith(remote_fencing_op_t * op, st_query_result_t * peer)
          * are available to execute the fencing operation. */
 
         if(stonith_watchdog_timeout_ms && (device == NULL || safe_str_eq(device, "watchdog"))) {
-            crm_notice("Waiting %ds for %s to self-fence (%s) for %s.%.8s (%p)",
+            crm_notice("Waiting %lds for %s to self-fence (%s) for client %s.%.8s",
                      stonith_watchdog_timeout_ms/1000, op->target,
-                     op->action, op->client_name, op->id, device);
+                     op->action, op->client_name, op->id);
 
             op->op_timer_one = g_timeout_add(stonith_watchdog_timeout_ms, remote_op_watchdog_done, op);
             return;
         }
 
         if (op->state == st_query) {
-           crm_info("No peers (out of %d) have devices capable of fencing (%s) %s for %s (%d)",
-                   op->replies, op->action, op->target, op->client_name,
-                   op->state);
+           crm_info("No peers (out of %d) have devices capable of fencing "
+                    "(%s) %s for client %s " CRM_XS " state=%d",
+                    op->replies, op->action, op->target, op->client_name,
+                    op->state);
 
             rc = -ENODEV;
         } else {
-           crm_info("No peers (out of %d) are capable of fencing (%s) %s for %s (%d)",
-                   op->replies, op->action, op->target, op->client_name,
-                   op->state);
+           crm_info("No peers (out of %d) are capable of fencing (%s) %s "
+                    "for client %s " CRM_XS " state=%d",
+                    op->replies, op->action, op->target, op->client_name,
+                    op->state);
         }
 
         op->state = st_failed;
         remote_op_done(op, NULL, rc, FALSE);
 
-    } else if (device) {
-        crm_info("Waiting for additional peers capable of fencing (%s) %s with %s for %s.%.8s",
-                 op->action, op->target, device, op->client_name, op->id);
     } else {
-        crm_info("Waiting for additional peers capable of fencing (%s) %s for %s%.8s",
-                 op->action, op->target, op->client_name, op->id);
+        crm_info("Waiting for additional peers capable of fencing (%s) %s%s%s "
+                 "for client %s%.8s",
+                 op->action, op->target, (device? " with " : ""),
+                 (device? device : ""), op->client_name, op->id);
     }
 }
 
@@ -1829,7 +1886,7 @@ process_remote_stonith_query(xmlNode * msg)
     CRM_CHECK(dev != NULL, return -EPROTO);
     crm_element_value_int(dev, F_STONITH_AVAILABLE_DEVICES, &ndevices);
 
-    op = g_hash_table_lookup(remote_op_list, id);
+    op = g_hash_table_lookup(stonith_remote_op_list, id);
     if (op == NULL) {
         crm_debug("Received query reply for unknown or expired operation %s",
                   id);
@@ -1924,8 +1981,8 @@ process_remote_stonith_exec(xmlNode * msg)
 
     device = crm_element_value(dev, F_STONITH_DEVICE);
 
-    if (remote_op_list) {
-        op = g_hash_table_lookup(remote_op_list, id);
+    if (stonith_remote_op_list) {
+        op = g_hash_table_lookup(stonith_remote_op_list, id);
     }
 
     if (op == NULL && rc == pcmk_ok) {
@@ -1972,9 +2029,10 @@ process_remote_stonith_exec(xmlNode * msg)
     if (is_set(op->call_options, st_opt_topology)) {
         const char *device = crm_element_value(msg, F_STONITH_DEVICE);
 
-        crm_notice("Call to %s for '%s %s' on behalf of %s@%s: %s (%d)",
-                   device, op->target, op->action, op->client_name, op->originator,
-                   pcmk_strerror(rc), rc);
+        crm_notice("Action '%s' targeting %s using %s on behalf of %s@%s: %s "
+                   CRM_XS " rc=%d",
+                   op->action, op->target, device, op->client_name,
+                   op->originator, pcmk_strerror(rc), rc);
 
         /* We own the op, and it is complete. broadcast the result to all nodes
          * and notify our local clients. */
@@ -1987,8 +2045,8 @@ process_remote_stonith_exec(xmlNode * msg)
             /* A remapped "on" failed, but the node was already turned off
              * successfully, so ignore the error and continue.
              */
-            crm_warn("Ignoring %s 'on' failure (exit code %d) for %s after successful 'off'",
-                     device, rc, op->target);
+            crm_warn("Ignoring %s 'on' failure (exit code %d) targeting %s "
+                     "after successful 'off'", device, rc, op->target);
             rc = pcmk_ok;
         }
 
@@ -2028,61 +2086,6 @@ process_remote_stonith_exec(xmlNode * msg)
     return rc;
 }
 
-int
-stonith_fence_history(xmlNode * msg, xmlNode ** output)
-{
-    int rc = 0;
-    const char *target = NULL;
-    xmlNode *dev = get_xpath_object("//@" F_STONITH_TARGET, msg, LOG_TRACE);
-    char *nodename = NULL;
-
-    if (dev) {
-        int options = 0;
-
-        target = crm_element_value(dev, F_STONITH_TARGET);
-        crm_element_value_int(msg, F_STONITH_CALLOPTS, &options);
-        if (target && (options & st_opt_cs_nodeid)) {
-            int nodeid = crm_atoi(target, NULL);
-
-            nodename = stonith_get_peer_name(nodeid);
-            if (nodename) {
-                target = nodename;
-            }
-        }
-    }
-
-    crm_trace("Looking for operations on %s in %p", target, remote_op_list);
-
-    *output = create_xml_node(NULL, F_STONITH_HISTORY_LIST);
-    if (remote_op_list) {
-        GHashTableIter iter;
-        remote_fencing_op_t *op = NULL;
-
-        g_hash_table_iter_init(&iter, remote_op_list);
-        while (g_hash_table_iter_next(&iter, NULL, (void **)&op)) {
-            xmlNode *entry = NULL;
-
-            if (target && strcmp(op->target, target) != 0) {
-                continue;
-            }
-
-            rc = 0;
-            crm_trace("Attaching op %s", op->id);
-            entry = create_xml_node(*output, STONITH_OP_EXEC);
-            crm_xml_add(entry, F_STONITH_TARGET, op->target);
-            crm_xml_add(entry, F_STONITH_ACTION, op->action);
-            crm_xml_add(entry, F_STONITH_ORIGIN, op->originator);
-            crm_xml_add(entry, F_STONITH_DELEGATE, op->delegate);
-            crm_xml_add(entry, F_STONITH_CLIENTNAME, op->client_name);
-            crm_xml_add_int(entry, F_STONITH_DATE, op->completed);
-            crm_xml_add_int(entry, F_STONITH_STATE, op->state);
-        }
-    }
-
-    free(nodename);
-    return rc;
-}
-
 gboolean
 stonith_check_fence_tolerance(int tolerance, const char *target, const char *action)
 {
@@ -2090,13 +2093,15 @@ stonith_check_fence_tolerance(int tolerance, const char *target, const char *act
     time_t now = time(NULL);
     remote_fencing_op_t *rop = NULL;
 
-    crm_trace("tolerance=%d, remote_op_list=%p", tolerance, remote_op_list);
+    crm_trace("tolerance=%d, remote_op_list=%p", tolerance,
+              stonith_remote_op_list);
 
-    if (tolerance <= 0 || !remote_op_list || target == NULL || action == NULL) {
+    if (tolerance <= 0 || !stonith_remote_op_list || target == NULL ||
+        action == NULL) {
         return FALSE;
     }
 
-    g_hash_table_iter_init(&iter, remote_op_list);
+    g_hash_table_iter_init(&iter, stonith_remote_op_list);
     while (g_hash_table_iter_next(&iter, NULL, (void **)&rop)) {
         if (strcmp(rop->target, target) != 0) {
             continue;
