@@ -1,20 +1,10 @@
 /*
- * Copyright (C) 2009 Andrew Beekhof <andrew@beekhof.net>
+ * Copyright 2009-2018 Andrew Beekhof <andrew@beekhof.net>
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This software is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * This source code is licensed under the GNU Lesser General Public License
+ * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
  */
+
 #ifndef CRM_COMMON_MAINLOOP__H
 #  define CRM_COMMON_MAINLOOP__H
 
@@ -77,6 +67,30 @@ struct ipc_client_callbacks {
 qb_ipcs_service_t *mainloop_add_ipc_server(const char *name, enum qb_ipc_type type,
                                            struct qb_ipcs_service_handlers *callbacks);
 
+/*!
+ * \brief Start server-side API end-point, hooked into the internal event loop
+ *
+ * \param[in] name    name of the IPC end-point ("address" for the client)
+ * \param[in] type    selects libqb's IPC back-end (or use #QB_IPC_NATIVE)
+ * \param[in] callbacks  defines libqb's IPC service-level handlers
+ * \param[in] priority  priority relative to other events handled in the
+ *                      abstract handling loop, use #QB_LOOP_MED when unsure
+ *
+ * \return libqb's opaque handle to the created service abstraction
+ *
+ * \note For portability concerns, do not use this function if you keep
+ *       \p priority as #QB_LOOP_MED, stick with #mainloop_add_ipc_server
+ *       (with exactly such semantics) instead (once you link with this new
+ *       symbol employed, you can't downgrade the library freely anymore).
+ *
+ * \note The intended effect will only get fully reflected when run-time
+ *       linked to patched libqb: https://github.com/ClusterLabs/libqb/pull/352
+ */
+qb_ipcs_service_t *mainloop_add_ipc_server_with_prio(const char *name,
+                                                    enum qb_ipc_type type,
+                                                    struct qb_ipcs_service_handlers *callbacks,
+                                                    enum qb_loop_priority prio);
+
 void mainloop_del_ipc_server(qb_ipcs_service_t * server);
 
 mainloop_io_t *mainloop_add_ipc_client(const char *name, int priority, size_t max_size,
@@ -121,6 +135,9 @@ pid_t mainloop_child_pid(mainloop_child_t * child);
 void mainloop_clear_child_userdata(mainloop_child_t * child);
 gboolean mainloop_child_kill(pid_t pid);
 
+void pcmk_drain_main_loop(GMainLoop *mloop, guint timer_ms,
+                          bool (*check)(guint));
+
 #  define G_PRIORITY_MEDIUM (G_PRIORITY_HIGH/2)
 
 #endif
