@@ -1,26 +1,25 @@
-/* 
- * Copyright (C) 2011 Andrew Beekhof <andrew@beekhof.net>
- * 
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- * 
- * This software is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- * 
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+/*
+ * Copyright 2011-2018 Andrew Beekhof <andrew@beekhof.net>
+ *
+ * This source code is licensed under the GNU Lesser General Public License
+ * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
  */
+
 #ifndef STONITH_NG_INTERNAL__H
 #  define STONITH_NG_INTERNAL__H
 
 #  include <crm/common/ipc.h>
 #  include <crm/common/xml.h>
 
+enum st_device_flags
+{
+    st_device_supports_list   = 0x0001,
+    st_device_supports_status = 0x0002,
+    st_device_supports_reboot = 0x0004,
+    st_device_supports_parameter_plug = 0x0008,
+    st_device_supports_parameter_port = 0x0010,
+};
+
 struct stonith_action_s;
 typedef struct stonith_action_s stonith_action_t;
 
@@ -29,26 +28,34 @@ stonith_action_t *stonith_action_create(const char *agent,
                                         const char *victim,
                                         uint32_t victim_nodeid,
                                         int timeout,
-                                        GHashTable * device_args, GHashTable * port_map);
+                                        GHashTable * device_args,
+                                        GHashTable * port_map,
+                                        const char * host_arg);
+void stonith__destroy_action(stonith_action_t *action);
+void stonith__action_result(stonith_action_t *action, int *rc, char **output,
+                            char **error_output);
 
-GPid
+int
 stonith_action_execute_async(stonith_action_t * action,
                              void *userdata,
                              void (*done) (GPid pid, int rc, const char *output,
-                                           gpointer user_data));
+                                           gpointer user_data),
+                             void (*fork_cb) (GPid pid, gpointer user_data));
 
-int
- stonith_action_execute(stonith_action_t * action, int *agent_result, char **output);
-
-gboolean is_redhat_agent(const char *agent);
+int stonith__execute(stonith_action_t *action);
 
 xmlNode *create_level_registration_xml(const char *node, const char *pattern,
                                        const char *attr, const char *value,
                                        int level,
                                        stonith_key_value_t *device_list);
 
-xmlNode *create_device_registration_xml(const char *id, const char *namespace, const char *agent,
-                                        stonith_key_value_t * params, const char *rsc_provides);
+xmlNode *create_device_registration_xml(const char *id,
+                                        enum stonith_namespace namespace,
+                                        const char *agent,
+                                        stonith_key_value_t *params,
+                                        const char *rsc_provides);
+
+long long stonith__device_parameter_flags(xmlNode *metadata);
 
 #  define ST_LEVEL_MAX 10
 
@@ -59,10 +66,12 @@ xmlNode *create_device_registration_xml(const char *id, const char *namespace, c
 #  define F_STONITH_OPERATION     "st_op"
 #  define F_STONITH_TARGET        "st_target"
 #  define F_STONITH_REMOTE_OP_ID  "st_remote_op"
+#  define F_STONITH_REMOTE_OP_ID_RELAY  "st_remote_op_relay"
 #  define F_STONITH_RC            "st_rc"
 /*! Timeout period per a device execution */
 #  define F_STONITH_TIMEOUT       "st_timeout"
 #  define F_STONITH_TOLERANCE     "st_tolerance"
+#  define F_STONITH_DELAY         "st_delay"
 /*! Action specific timeout period returned in query of fencing devices. */
 #  define F_STONITH_ACTION_TIMEOUT       "st_action_timeout"
 /*! Host in query result is not allowed to run this action */
@@ -94,6 +103,7 @@ xmlNode *create_device_registration_xml(const char *id, const char *namespace, c
 #  define F_STONITH_DATE          "st_date"
 #  define F_STONITH_STATE         "st_state"
 #  define F_STONITH_ACTIVE        "st_active"
+#  define F_STONITH_DIFFERENTIAL  "st_differential"
 
 #  define F_STONITH_DEVICE        "st_device_id"
 #  define F_STONITH_ACTION        "st_device_action"
@@ -136,4 +146,22 @@ xmlNode *create_device_registration_xml(const char *id, const char *namespace, c
 
 #  define STONITH_WATCHDOG_AGENT  "#watchdog"
 
+#  ifdef HAVE_STONITH_STONITH_H
+// utilities from st_lha.c
+int stonith__list_lha_agents(stonith_key_value_t **devices);
+int stonith__lha_metadata(const char *agent, int timeout, char **output);
+bool stonith__agent_is_lha(const char *agent);
+int stonith__lha_validate(stonith_t *st, int call_options, const char *target,
+                          const char *agent, GHashTable *params,
+                          int timeout, char **output, char **error_output);
+#  endif
+
+// utilities from st_rhcs.c
+int stonith__list_rhcs_agents(stonith_key_value_t **devices);
+int stonith__rhcs_metadata(const char *agent, int timeout, char **output);
+bool stonith__agent_is_rhcs(const char *agent);
+int stonith__rhcs_validate(stonith_t *st, int call_options, const char *target,
+                           const char *agent, GHashTable *params, const char *host_arg,
+                           int timeout, char **output, char **error_output);
+
 #endif
