@@ -254,6 +254,8 @@ typedef struct lrmd_event_data_s {
     const char *exit_reason;
 } lrmd_event_data_t;
 
+lrmd_event_data_t *lrmd_new_event(const char *rsc_id, const char *task,
+                                  int interval_ms);
 lrmd_event_data_t *lrmd_copy_event(lrmd_event_data_t * event);
 void lrmd_free_event(lrmd_event_data_t * event);
 
@@ -264,6 +266,8 @@ typedef struct lrmd_rsc_info_s {
     char *provider;
 } lrmd_rsc_info_t;
 
+lrmd_rsc_info_t *lrmd_new_rsc_info(const char *rsc_id, const char *standard,
+                                   const char *provider, const char *type);
 lrmd_rsc_info_t *lrmd_copy_rsc_info(lrmd_rsc_info_t * rsc_info);
 void lrmd_free_rsc_info(lrmd_rsc_info_t * rsc_info);
 
@@ -493,6 +497,28 @@ typedef struct lrmd_api_operations_s {
                        const char *alert_path, int timeout, /* ms */
                        lrmd_key_value_t *params); /* ownership of params is given up to api here */
 
+    /*!
+     * \brief Get resource metadata for a resource agent, passing parameters
+     *
+     * \param[in]  lrmd      Executor connection (unused)
+     * \param[in]  standard  Resource agent class
+     * \param[in]  provider  Resource agent provider
+     * \param[in]  agent     Resource agent type
+     * \param[out] output    Metadata will be stored here (must not be NULL)
+     * \param[in]  options   Options to use with any executor API calls (unused)
+     * \param[in]  params    Parameters to pass to agent via environment
+     *
+     * \note This is identical to the get_metadata() API call, except parameters
+     *       will be passed to the resource agent via environment variables.
+     * \note The API will handle freeing params.
+     *
+     * \return lrmd_ok on success, negative error code on failure
+     */
+    int (*get_metadata_params) (lrmd_t *lrmd, const char *standard,
+                                const char *provider, const char *agent,
+                                char **output, enum lrmd_call_options options,
+                                lrmd_key_value_t *params);
+
 } lrmd_api_operations_t;
 
 struct lrmd_s {
