@@ -1,20 +1,10 @@
 /*
- * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
+ * Copyright 2004-2019 Andrew Beekhof <andrew@beekhof.net>
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This software is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * This source code is licensed under the GNU Lesser General Public License
+ * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
  */
+
 #ifndef PE_INTERNAL__H
 #  define PE_INTERNAL__H
 #  include <string.h>
@@ -32,8 +22,31 @@
 #  define pe_set_action_bit(action, bit) action->flags = crm_set_bit(__FUNCTION__, __LINE__, action->uuid, action->flags, bit)
 #  define pe_clear_action_bit(action, bit) action->flags = crm_clear_bit(__FUNCTION__, __LINE__, action->uuid, action->flags, bit)
 
+typedef struct pe__location_constraint_s {
+    char *id;                           // Constraint XML ID
+    pe_resource_t *rsc_lh;              // Resource being located
+    enum rsc_role_e role_filter;        // Role to locate
+    enum pe_discover_e discover_mode;   // Resource discovery
+    GListPtr node_list_rh;              // List of pe_node_t*
+} pe__location_t;
+
+typedef struct pe__order_constraint_s {
+    int id;
+    enum pe_ordering type;
+
+    void *lh_opaque;
+    resource_t *lh_rsc;
+    action_t *lh_action;
+    char *lh_action_task;
+
+    void *rh_opaque;
+    resource_t *rh_rsc;
+    action_t *rh_action;
+    char *rh_action_task;
+} pe__ordering_t;
+
 typedef struct notify_data_s {
-    GHashTable *keys;
+    GSList *keys;               // Environment variable name/value pairs
 
     const char *action;
 
@@ -62,14 +75,13 @@ void append_hashtable(gpointer key, gpointer value, gpointer user_data);
 
 char *native_parameter(resource_t * rsc, node_t * node, gboolean create, const char *name,
                        pe_working_set_t * data_set);
-node_t *native_location(resource_t * rsc, GListPtr * list, gboolean current);
+pe_node_t *native_location(const pe_resource_t *rsc, GList **list, int current);
 
 void pe_metadata(void);
 void verify_pe_options(GHashTable * options);
 
 void common_update_score(resource_t * rsc, const char *id, int score);
 void native_add_running(resource_t * rsc, node_t * node, pe_working_set_t * data_set);
-node_t *rsc_known_on(resource_t * rsc, GListPtr * list);
 
 gboolean native_unpack(resource_t * rsc, pe_working_set_t * data_set);
 gboolean group_unpack(resource_t * rsc, pe_working_set_t * data_set);
@@ -103,6 +115,9 @@ enum rsc_role_e clone_resource_state(const resource_t * rsc, gboolean current);
 enum rsc_role_e master_resource_state(const resource_t * rsc, gboolean current);
 enum rsc_role_e container_resource_state(const resource_t * rsc, gboolean current);
 
+void pe__count_common(pe_resource_t *rsc);
+void pe__count_bundle(pe_resource_t *rsc);
+
 gboolean common_unpack(xmlNode * xml_obj, resource_t ** rsc, resource_t * parent,
                        pe_working_set_t * data_set);
 void common_free(resource_t * rsc);
@@ -125,6 +140,25 @@ int pe_get_failcount(node_t *node, resource_t *rsc, time_t *last_failure,
                      uint32_t flags, xmlNode *xml_op,
                      pe_working_set_t *data_set);
 
+pe_action_t *pe__clear_failcount(pe_resource_t *rsc, pe_node_t *node,
+                                 const char *reason,
+                                 pe_working_set_t *data_set);
+
+/* Functions for finding/counting a resource's active nodes */
+
+pe_node_t *pe__find_active_on(const resource_t *rsc,
+                              unsigned int *count_all,
+                              unsigned int *count_clean);
+pe_node_t *pe__find_active_requires(const resource_t *rsc,
+                                    unsigned int *count);
+
+static inline pe_node_t *
+pe__current_node(const resource_t *rsc)
+{
+    return pe__find_active_on(rsc, NULL, NULL);
+}
+
+
 /* Binary like operators for lists of nodes */
 extern void node_list_exclude(GHashTable * list, GListPtr list2, gboolean merge_scores);
 extern GListPtr node_list_dup(GListPtr list, gboolean reset, gboolean filter);
@@ -156,17 +190,16 @@ extern void pe_free_shallow_adv(GListPtr alist, gboolean with_data);
 /* Printing functions for debug */
 extern void print_node(const char *pre_text, node_t * node, gboolean details);
 
-extern void print_resource(int log_level, const char *pre_text, resource_t * rsc, gboolean details);
-
-extern void dump_node_scores_worker(int level, const char *file, const char *function, int line,
-                                    resource_t * rsc, const char *comment, GHashTable * nodes);
-
 extern void dump_node_capacity(int level, const char *comment, node_t * node);
 extern void dump_rsc_utilization(int level, const char *comment, resource_t * rsc, node_t * node);
 
-#  define dump_node_scores(level, rsc, text, nodes) do {		\
-        dump_node_scores_worker(level, __FILE__, __FUNCTION__, __LINE__, rsc, text, nodes); \
-    } while(0)
+void pe__show_node_weights_as(const char *file, const char *function,
+                              int line, bool to_log, pe_resource_t *rsc,
+                              const char *comment, GHashTable *nodes);
+
+#define pe__show_node_weights(level, rsc, text, nodes)              \
+        pe__show_node_weights_as(__FILE__, __FUNCTION__, __LINE__,  \
+                                 (level), (rsc), (text), (nodes))
 
 /* Sorting functions */
 extern gint sort_rsc_priority(gconstpointer a, gconstpointer b);
@@ -296,7 +329,7 @@ typedef struct op_digest_cache_s {
 op_digest_cache_t *rsc_action_digest_cmp(resource_t * rsc, xmlNode * xml_op, node_t * node,
                                          pe_working_set_t * data_set);
 
-action_t *pe_fence_op(node_t * node, const char *op, bool optional, const char *reason, pe_working_set_t * data_set);
+pe_action_t *pe_fence_op(pe_node_t * node, const char *op, bool optional, const char *reason, bool priority_delay, pe_working_set_t * data_set);
 void trigger_unfencing(
     resource_t * rsc, node_t *node, const char *reason, action_t *dependency, pe_working_set_t * data_set);
 
@@ -313,16 +346,26 @@ gboolean add_tag_ref(GHashTable * tags, const char * tag_name,  const char * obj
 
 void print_rscs_brief(GListPtr rsc_list, const char * pre_text, long options,
                       void * print_data, gboolean print_all);
-void pe_fence_node(pe_working_set_t * data_set, node_t * node, const char *reason);
+void pe_fence_node(pe_working_set_t * data_set, pe_node_t * node, const char *reason, bool priority_delay);
 
 node_t *pe_create_node(const char *id, const char *uname, const char *type,
                        const char *score, pe_working_set_t * data_set);
 bool remote_id_conflict(const char *remote_name, pe_working_set_t *data);
 void common_print(resource_t * rsc, const char *pre_text, const char *name, node_t *node, long options, void *print_data);
-resource_t *find_container_child(const char *stem, resource_t * rsc, node_t *node);
+resource_t *find_container_child(const resource_t *bundle, const node_t *node);
 bool container_fix_remote_addr(resource_t *rsc);
 const char *container_fix_remote_addr_in(resource_t *rsc, xmlNode *xml, const char *field);
 const char *pe_node_attribute_calculated(pe_node_t *node, const char *name, resource_t *rsc);
 const char *pe_node_attribute_raw(pe_node_t *node, const char *name);
+void pe__add_param_check(xmlNode *rsc_op, pe_resource_t *rsc, pe_node_t *node,
+                         enum pe_check_parameters, pe_working_set_t *data_set);
+void pe__foreach_param_check(pe_working_set_t *data_set,
+                             void (*cb)(pe_resource_t*, pe_node_t*, xmlNode*,
+                                        enum pe_check_parameters,
+                                        pe_working_set_t*));
+void pe__free_param_checks(pe_working_set_t *data_set);
+
+bool pe__shutdown_requested(pe_node_t *node);
+bool pe__resource_is_disabled(pe_resource_t *rsc);
 
 #endif
