@@ -34,6 +34,8 @@
 
 #  define T_STONITH_NOTIFY_DISCONNECT     "st_notify_disconnect"
 #  define T_STONITH_NOTIFY_FENCE          "st_notify_fence"
+#  define T_STONITH_NOTIFY_HISTORY        "st_notify_history"
+#  define T_STONITH_NOTIFY_HISTORY_SYNCED "st_notify_history_synced"
 
 /* *INDENT-OFF* */
 enum stonith_state {
@@ -59,6 +61,10 @@ enum stonith_call_options {
     st_opt_timeout_updates = 0x00002000,
     /*! Only report back if operation is a success in callback */
     st_opt_report_only_success = 0x00004000,
+    /* used where ever apropriate - e.g. cleanup of history */
+    st_opt_cleanup         = 0x000080000,
+    /* used where ever apropriate - e.g. send out a history query to all nodes */
+    st_opt_broadcast       = 0x000100000,
 };
 
 /*! Order matters here, do not change values */
@@ -71,6 +77,24 @@ enum op_state
     st_failed,
 };
 
+// Supported fence agent interface standards
+enum stonith_namespace {
+    st_namespace_invalid,
+    st_namespace_any,
+    st_namespace_internal,  // Implemented internally by Pacemaker
+
+    /* Neither of these projects are active any longer, but the fence agent
+     * interfaces they created are still in use and supported by Pacemaker.
+     */
+    st_namespace_rhcs,      // Red Hat Cluster Suite compatible
+    st_namespace_lha,       // Linux-HA compatible
+};
+
+enum stonith_namespace stonith_text2namespace(const char *namespace_s);
+const char *stonith_namespace2text(enum stonith_namespace st_namespace);
+enum stonith_namespace stonith_get_namespace(const char *agent,
+                                             const char *namespace_s);
+
 typedef struct stonith_key_value_s {
     char *key;
     char *value;
@@ -357,6 +381,49 @@ typedef struct stonith_api_operations_s
                                const char *attr, const char *value,
                                int level, stonith_key_value_t *device_list);
 
+    /*!
+     * \brief Validate an arbitrary stonith device configuration
+     *
+     * \param[in]  st            Stonithd connection to use
+     * \param[in]  call_options  Bitmask of stonith_call_options to use with fencer
+     * \param[in]  rsc_id        ID used to replace CIB secrets in params
+     * \param[in]  namespace_s   Namespace of fence agent to validate (optional)
+     * \param[in]  agent         Fence agent to validate
+     * \param[in]  params        Configuration parameters to pass to fence agent
+     * \param[in]  timeout       Fail if no response within this many seconds
+     * \param[out] output        If non-NULL, where to store any agent output
+     * \param[out] error_output  If non-NULL, where to store agent error output
+     *
+     * \return pcmk_ok if validation succeeds, -errno otherwise
+     *
+     * \note If pcmk_ok is returned, the caller is responsible for freeing
+     *       the output (if requested).
+     */
+    int (*validate)(stonith_t *st, int call_options, const char *rsc_id,
+                    const char *namespace_s, const char *agent,
+                    stonith_key_value_t *params, int timeout, char **output,
+                    char **error_output);
+
+    /*!
+     * \brief Issue a fencing action against a node with requested fencing delay.
+     *
+     * \note Possible actions are, 'on', 'off', and 'reboot'.
+     *
+     * \param st, stonith connection
+     * \param options, call options
+     * \param node, The target node to fence
+     * \param action, The fencing action to take
+     * \param timeout, The default per device timeout to use with each device
+     *                 capable of fencing the target.
+     * \param delay, Apply a fencing delay. Value -1 means disable also any
+     *               static/random fencing delays from pcmk_delay_base/max
+     *
+     * \retval 0 success
+     * \retval negative error code on failure.
+     */
+    int (*fence_with_delay)(stonith_t *st, int options, const char *node, const char *action,
+                            int timeout, int tolerance, int delay);
+
 } stonith_api_operations_t;
 
 struct stonith_s
@@ -377,6 +444,7 @@ void stonith_api_delete(stonith_t * st);
 
 void stonith_dump_pending_callbacks(stonith_t * st);
 
+// deprecated (use stonith_get_namespace() instead)
 const char *get_stonith_provider(const char *agent, const char *provider);
 
 bool stonith_dispatch(stonith_t * st);
@@ -385,6 +453,8 @@ stonith_key_value_t *stonith_key_value_add(stonith_key_value_t * kvp, const char
                                            const char *value);
 void stonith_key_value_freeall(stonith_key_value_t * kvp, int keys, int values);
 
+void stonith_history_free(stonith_history_t *history);
+
 /* Basic helpers that allows nodes to be fenced and the history to be
  * queried without mainloop or the caller understanding the full API
  *
