@@ -36,7 +36,6 @@
 
 /* Dynamic loading of libraries */
 void *find_library_function(void **handle, const char *lib, const char *fn, int fatal);
-void *convert_const_pointer(const void *ptr);
 
 /* For ACLs */
 char *uid2username(uid_t uid);
@@ -140,6 +139,27 @@ extern int node_score_yellow;
 extern int node_score_infinity;
 
 /* Assorted convenience functions */
+
+/*!
+ * \internal
+ * \brief Detect if process per PID and optionally exe path (component) exists
+ *
+ * \param[in] pid     PID of process assumed alive, disproving of which to try
+ * \param[in] daemon  exe path (component) to possibly match with procfs entry
+ *
+ * \return -1 on invalid PID specification, -2 when the calling process has no
+ *         (is refused an) ability to (dis)prove the predicate,
+ *         0 if the negation of the predicate is confirmed (check-through-kill
+ *         indicates so, or the subsequent check-through-procfs-match on
+ *         \p daemon when provided and procfs available at the standard path),
+ *         1 if it cannot be disproved (reliably [modulo race conditions]
+ *         when \p daemon provided, procfs available at the standard path
+ *         and the calling process has permissions to access the respective
+ *         procfs location, less so otherwise, since mere check-through-kill
+ *         is exercised without powers to exclude PID recycled in the interim).
+ *
+ * \note This function cannot be used to verify \e authenticity of the process.
+ */
 int crm_pid_active(long pid, const char *daemon);
 void crm_make_daemon(const char *name, gboolean daemonize, const char *pidfile);
 
@@ -192,57 +212,6 @@ crm_set_bit(const char *function, int line, const char *target, long long word,
 
 char *generate_hash_key(const char *crm_msg_reference, const char *sys);
 
-/*! remote tcp/tls helper functions */
-typedef struct crm_remote_s crm_remote_t;
-
-int crm_remote_send(crm_remote_t * remote, xmlNode * msg);
-int crm_remote_ready(crm_remote_t * remote, int total_timeout /*ms */ );
-gboolean crm_remote_recv(crm_remote_t * remote, int total_timeout /*ms */ , int *disconnected);
-xmlNode *crm_remote_parse_buffer(crm_remote_t * remote);
-int crm_remote_tcp_connect(const char *host, int port);
-int crm_remote_tcp_connect_async(const char *host, int port, int timeout,       /*ms */
-                                 int *timer_id, void *userdata, void (*callback) (void *userdata, int sock));
-int crm_remote_accept(int ssock);
-void crm_sockaddr2str(void *sa, char *s);
-
-#  ifdef HAVE_GNUTLS_GNUTLS_H
-/*!
- * \internal
- * \brief Initiate the client handshake after establishing the tcp socket.
- * \note This is a blocking function, it will block until the entire handshake
- *       is complete or until the timeout period is reached.
- * \retval 0 success
- * \retval negative, failure
- */
-int crm_initiate_client_tls_handshake(crm_remote_t * remote, int timeout_ms);
-
-/*!
- * \internal
- * \brief Create client or server session for anon DH encryption credentials
- * \param sock, the socket the session will use for transport
- * \param type, GNUTLS_SERVER or GNUTLS_CLIENT
- * \param credentials, gnutls_anon_server_credentials_t or gnutls_anon_client_credentials_t
- *
- * \retval gnutls_session_t * on success
- * \retval NULL on failure
- */
-void *crm_create_anon_tls_session(int sock, int type, void *credentials);
-
-/*!
- * \internal
- * \brief Create client or server session for PSK credentials
- * \param sock, the socket the session will use for transport
- * \param type, GNUTLS_SERVER or GNUTLS_CLIENT
- * \param credentials, gnutls_psk_server_credentials_t or gnutls_osk_client_credentials_t
- *
- * \retval gnutls_session_t * on success
- * \retval NULL on failure
- */
-void *create_psk_tls_session(int csock, int type, void *credentials);
-#  endif
-
-#  define REMOTE_MSG_TERMINATOR "\r\n\r\n"
-
 const char *daemon_option(const char *option);
 void set_daemon_option(const char *option, const char *value);
 gboolean daemon_option_enabled(const char *daemon, const char *option);
@@ -276,6 +245,7 @@ long crm_read_pidfile(const char *filename);
 #  define F_ATTRD_RESOURCE          "attr_resource"
 #  define F_ATTRD_OPERATION         "attr_clear_operation"
 #  define F_ATTRD_INTERVAL          "attr_clear_interval"
+#  define F_ATTRD_IS_FORCE_WRITE "attrd_is_force_write"
 
 /* attrd operations */
 #  define ATTRD_OP_PEER_REMOVE   "peer-remove"
