@@ -1,21 +1,10 @@
 /*
- * Copyright (c) 2004 Andrew Beekhof <andrew@beekhof.net>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- * 
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- * 
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * Copyright 2004-2018 Andrew Beekhof <andrew@beekhof.net>
  *
+ * This source code is licensed under the GNU Lesser General Public License
+ * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
  */
+
 #include <crm_internal.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -231,8 +220,10 @@ get_shadow_file(const char *suffix)
         if (pwent) {
             user = pwent->pw_name;
         } else {
-            crm_perror(LOG_ERR, "Cannot get password entry for uid: %d", uid);
             user = getenv("USER");
+            crm_perror(LOG_ERR,
+                       "Assuming %s because cannot get user details for user ID %d",
+                       (user? user : "unprivileged user"), uid);
         }
 
         if (safe_str_eq(user, "root") || safe_str_eq(user, CRM_DAEMON_USER)) {
