@@ -34,6 +34,7 @@
 #include <crm/msg_xml.h>
 #include <crm/common/ipcs.h>
 #include <crm/common/mainloop.h>
+#include <crm/common/remote_internal.h>
 
 #ifdef HAVE_GNUTLS_GNUTLS_H
 #  undef KEYFILE
@@ -238,7 +239,13 @@ cib_tls_signon(cib_t * cib, crm_remote_t * connection, gboolean event_channel)
         }
 
         /* bind the socket to GnuTls lib */
-        connection->tls_session = crm_create_anon_tls_session(sock, GNUTLS_CLIENT, anon_cred_c);
+        connection->tls_session = pcmk__new_tls_session(sock, GNUTLS_CLIENT,
+                                                        GNUTLS_CRD_ANON,
+                                                        anon_cred_c);
+        if (connection->tls_session == NULL) {
+            cib_tls_close(cib);
+            return -1;
+        }
 
         if (crm_initiate_client_tls_handshake(connection, DEFAULT_CLIENT_HANDSHAKE_TIMEOUT) != 0) {
             crm_err("Session creation for %s:%d failed", private->server, private->port);
@@ -428,7 +435,7 @@ cib_remote_signon(cib_t * cib, const char *name, enum cib_conn_type type)
     }
 
     if (rc == pcmk_ok) {
-        crm_notice("%s: Opened connection to %s:%d\n", name, private->server, private->port);
+        crm_notice("%s: Opened connection to %s:%d", name, private->server, private->port);
         cib->state = cib_connected_command;
         cib->type = cib_command;
 
@@ -540,7 +547,7 @@ cib_remote_perform_op(cib_t * cib, const char *op, const char *host, const char
         return cib->call_id;
     }
 
-    crm_trace("Waiting for a syncronous reply");
+    crm_trace("Waiting for a synchronous reply");
 
     start_time = time(NULL);
     remaining_time = cib->call_timeout ? cib->call_timeout : 60;
@@ -594,7 +601,7 @@ cib_remote_perform_op(cib_t * cib, const char *op, const char *host, const char
         return -ENOMSG;
     }
 
-    crm_trace("Syncronous reply received");
+    crm_trace("Synchronous reply received");
 
     /* Start processing the reply... */
     if (crm_element_value_int(op_reply, F_CIB_RC, &rc) != 0) {
