@@ -33,7 +33,6 @@
 
 #include <stdlib.h>
 #include <errno.h>
-#include <fcntl.h>
 #include <glib.h>
 
 #include <bzlib.h>
@@ -41,6 +40,7 @@
 #include <crm/common/ipcs.h>
 #include <crm/common/xml.h>
 #include <crm/common/mainloop.h>
+#include <crm/common/remote_internal.h>
 
 #ifdef HAVE_GNUTLS_GNUTLS_H
 #  undef KEYFILE
@@ -166,61 +166,247 @@ crm_initiate_client_tls_handshake(crm_remote_t * remote, int timeout_ms)
     return rc;
 }
 
-void *
-crm_create_anon_tls_session(int csock, int type /* GNUTLS_SERVER, GNUTLS_CLIENT */ ,
-                            void *credentials)
+/*!
+ * \internal
+ * \brief Set minimum prime size required by TLS client
+ *
+ * \param[in] session  TLS session to affect
+ */
+static void
+pcmk__set_minimum_dh_bits(gnutls_session_t *session)
 {
-    gnutls_session_t *session = gnutls_malloc(sizeof(gnutls_session_t));
+    const char *dh_min_bits_s = getenv("PCMK_dh_min_bits");
 
-    gnutls_init(session, type);
-#  ifdef HAVE_GNUTLS_PRIORITY_SET_DIRECT
-/*      http://www.manpagez.com/info/gnutls/gnutls-2.10.4/gnutls_81.php#Echo-Server-with-anonymous-authentication */
-    gnutls_priority_set_direct(*session, "NORMAL:+ANON-DH", NULL);
-/*	gnutls_priority_set_direct (*session, "NONE:+VERS-TLS-ALL:+CIPHER-ALL:+MAC-ALL:+SIGN-ALL:+COMP-ALL:+ANON-DH", NULL); */
-#  else
-    gnutls_set_default_priority(*session);
-    gnutls_kx_set_priority(*session, anon_tls_kx_order);
-#  endif
-    gnutls_transport_set_ptr(*session, (gnutls_transport_ptr_t) GINT_TO_POINTER(csock));
-    switch (type) {
-        case GNUTLS_SERVER:
-            gnutls_credentials_set(*session, GNUTLS_CRD_ANON,
-                                   (gnutls_anon_server_credentials_t) credentials);
-            break;
-        case GNUTLS_CLIENT:
-            gnutls_credentials_set(*session, GNUTLS_CRD_ANON,
-                                   (gnutls_anon_client_credentials_t) credentials);
-            break;
+    if (dh_min_bits_s) {
+        int dh_min_bits = crm_parse_int(dh_min_bits_s, "0");
+
+        /* This function is deprecated since GnuTLS 3.1.7, in favor of letting
+         * the priority string imply the DH requirements, but this is the only
+         * way to give the user control over compatibility with older servers.
+         */
+        if (dh_min_bits > 0) {
+            crm_info("Requiring server use a Diffie-Hellman prime of at least %d bits",
+                     dh_min_bits);
+            gnutls_dh_set_prime_bits(*session, dh_min_bits);
+        }
     }
-
-    return session;
 }
 
-void *
-create_psk_tls_session(int csock, int type /* GNUTLS_SERVER, GNUTLS_CLIENT */ , void *credentials)
+static unsigned int
+pcmk__bound_dh_bits(unsigned int dh_bits)
 {
-    gnutls_session_t *session = gnutls_malloc(sizeof(gnutls_session_t));
+    const char *dh_min_bits_s = getenv("PCMK_dh_min_bits");
+    const char *dh_max_bits_s = getenv("PCMK_dh_max_bits");
+    int dh_min_bits = 0;
+    int dh_max_bits = 0;
 
-    gnutls_init(session, type);
+    if (dh_min_bits_s) {
+        dh_min_bits = crm_parse_int(dh_min_bits_s, "0");
+    }
+    if (dh_max_bits_s) {
+        dh_max_bits = crm_parse_int(dh_max_bits_s, "0");
+        if ((dh_min_bits > 0) && (dh_max_bits > 0)
+            && (dh_max_bits < dh_min_bits)) {
+            crm_warn("Ignoring PCMK_dh_max_bits because it is less than PCMK_dh_min_bits");
+            dh_max_bits = 0;
+        }
+    }
+    if ((dh_min_bits > 0) && (dh_bits < dh_min_bits)) {
+        return dh_min_bits;
+    }
+    if ((dh_max_bits > 0) && (dh_bits > dh_max_bits)) {
+        return dh_max_bits;
+    }
+    return dh_bits;
+}
+
+/*!
+ * \internal
+ * \brief Initialize a new TLS session
+ *
+ * \param[in] csock       Connected socket for TLS session
+ * \param[in] conn_type   GNUTLS_SERVER or GNUTLS_CLIENT
+ * \param[in] cred_type   GNUTLS_CRD_ANON or GNUTLS_CRD_PSK
+ * \param[in] credentials TLS session credentials
+ *
+ * \return Pointer to newly created session object, or NULL on error
+ */
+gnutls_session_t *
+pcmk__new_tls_session(int csock, unsigned int conn_type,
+                      gnutls_credentials_type_t cred_type, void *credentials)
+{
+    int rc = GNUTLS_E_SUCCESS;
 #  ifdef HAVE_GNUTLS_PRIORITY_SET_DIRECT
-    gnutls_priority_set_direct(*session, "NORMAL:+DHE-PSK:+PSK", NULL);
-#  else
-    gnutls_set_default_priority(*session);
-    gnutls_kx_set_priority(*session, psk_tls_kx_order);
+    const char *prio_base = NULL;
+    char *prio = NULL;
 #  endif
-    gnutls_transport_set_ptr(*session, (gnutls_transport_ptr_t) GINT_TO_POINTER(csock));
-    switch (type) {
-        case GNUTLS_SERVER:
-            gnutls_credentials_set(*session, GNUTLS_CRD_PSK,
-                                   (gnutls_psk_server_credentials_t) credentials);
-            break;
-        case GNUTLS_CLIENT:
-            gnutls_credentials_set(*session, GNUTLS_CRD_PSK,
-                                   (gnutls_psk_client_credentials_t) credentials);
-            break;
+    gnutls_session_t *session = NULL;
+
+#  ifdef HAVE_GNUTLS_PRIORITY_SET_DIRECT
+    /* Determine list of acceptable ciphers, etc. Pacemaker always adds the
+     * values required for its functionality.
+     *
+     * For an example of anonymous authentication, see:
+     * http://www.manpagez.com/info/gnutls/gnutls-2.10.4/gnutls_81.php#Echo-Server-with-anonymous-authentication
+     */
+
+    prio_base = getenv("PCMK_tls_priorities");
+    if (prio_base == NULL) {
+        prio_base = PCMK_GNUTLS_PRIORITIES;
+    }
+    prio = crm_strdup_printf("%s:%s", prio_base,
+                             (cred_type == GNUTLS_CRD_ANON)? "+ANON-DH" : "+DHE-PSK:+PSK");
+#  endif
+
+    session = gnutls_malloc(sizeof(gnutls_session_t));
+    if (session == NULL) {
+        rc = GNUTLS_E_MEMORY_ERROR;
+        goto error;
     }
 
+    rc = gnutls_init(session, conn_type);
+    if (rc != GNUTLS_E_SUCCESS) {
+        goto error;
+    }
+
+#  ifdef HAVE_GNUTLS_PRIORITY_SET_DIRECT
+    /* @TODO On the server side, it would be more efficient to cache the
+     * priority with gnutls_priority_init2() and set it with
+     * gnutls_priority_set() for all sessions.
+     */
+    rc = gnutls_priority_set_direct(*session, prio, NULL);
+    if (rc != GNUTLS_E_SUCCESS) {
+        goto error;
+    }
+    if (conn_type == GNUTLS_CLIENT) {
+        pcmk__set_minimum_dh_bits(session);
+    }
+#  else
+    gnutls_set_default_priority(*session);
+    gnutls_kx_set_priority(*session, (cred_type == GNUTLS_CRD_ANON)? anon_tls_kx_order : psk_tls_kx_order);
+#  endif
+
+    gnutls_transport_set_ptr(*session,
+                             (gnutls_transport_ptr_t) GINT_TO_POINTER(csock));
+
+    rc = gnutls_credentials_set(*session, cred_type, credentials);
+    if (rc != GNUTLS_E_SUCCESS) {
+        goto error;
+    }
+#  ifdef HAVE_GNUTLS_PRIORITY_SET_DIRECT
+    free(prio);
+#  endif
     return session;
+
+error:
+    {
+#  ifdef HAVE_GNUTLS_PRIORITY_SET_DIRECT
+        const char *prio_s = prio;
+#  else
+        const char *prio_s = "default";
+#  endif
+
+        crm_err("Could not initialize %s TLS %s session: %s "
+                CRM_XS " rc=%d priority='%s'",
+                (cred_type == GNUTLS_CRD_ANON)? "anonymous" : "PSK",
+                (conn_type == GNUTLS_SERVER)? "server" : "client",
+                gnutls_strerror(rc), rc, prio_s);
+    }
+#  ifdef HAVE_GNUTLS_PRIORITY_SET_DIRECT
+    free(prio);
+#  endif
+    if (session != NULL) {
+        gnutls_free(session);
+    }
+    return NULL;
+}
+
+/*!
+ * \internal
+ * \brief Initialize Diffie-Hellman parameters for a TLS server
+ *
+ * \param[out] dh_params  Parameter object to initialize
+ *
+ * \return GNUTLS_E_SUCCESS on success, GnuTLS error code on error
+ * \todo The current best practice is to allow the client and server to
+ *       negotiate the Diffie-Hellman parameters via a TLS extension (RFC 7919).
+ *       However, we have to support both older versions of GnuTLS (<3.6) that
+ *       don't support the extension on our side, and older Pacemaker versions
+ *       that don't support the extension on the other side. The next best
+ *       practice would be to use a known good prime (see RFC 5114 section 2.2),
+ *       possibly stored in a file distributed with Pacemaker.
+ */
+int
+pcmk__init_tls_dh(gnutls_dh_params_t *dh_params)
+{
+    int rc = GNUTLS_E_SUCCESS;
+    unsigned int dh_bits = 0;
+
+    rc = gnutls_dh_params_init(dh_params);
+    if (rc != GNUTLS_E_SUCCESS) {
+        goto error;
+    }
+
+#ifdef HAVE_GNUTLS_SEC_PARAM_TO_PK_BITS
+    dh_bits = gnutls_sec_param_to_pk_bits(GNUTLS_PK_DH,
+                                          GNUTLS_SEC_PARAM_NORMAL);
+    if (dh_bits == 0) {
+        rc = GNUTLS_E_DH_PRIME_UNACCEPTABLE;
+        goto error;
+    }
+#else
+    dh_bits = 1024;
+#endif
+    dh_bits = pcmk__bound_dh_bits(dh_bits);
+
+    crm_info("Generating Diffie-Hellman parameters with %u-bit prime for TLS",
+             dh_bits);
+    rc = gnutls_dh_params_generate2(*dh_params, dh_bits);
+    if (rc != GNUTLS_E_SUCCESS) {
+        goto error;
+    }
+
+    return rc;
+
+error:
+    crm_err("Could not initialize Diffie-Hellman parameters for TLS: %s "
+            CRM_XS " rc=%d", gnutls_strerror(rc), rc);
+    CRM_ASSERT(rc == GNUTLS_E_SUCCESS);
+    return rc;
+}
+
+/*!
+ * \internal
+ * \brief Process handshake data from TLS client
+ *
+ * Read as much TLS handshake data as is available.
+ *
+ * \param[in] client  Client connection
+ *
+ * \retval GnuTLS error code on error
+ * \retval 0 if more data is needed
+ * \retval 1 if handshake is successfully completed
+ */
+int
+pcmk__read_handshake_data(crm_client_t *client)
+{
+    int rc = 0;
+
+    CRM_ASSERT(client && client->remote && client->remote->tls_session);
+
+    do {
+        rc = gnutls_handshake(*client->remote->tls_session);
+    } while (rc == GNUTLS_E_INTERRUPTED);
+
+    if (rc == GNUTLS_E_AGAIN) {
+        /* No more data is available at the moment. This function should be
+         * invoked again once the client sends more.
+         */
+        return 0;
+    } else if (rc != GNUTLS_E_SUCCESS) {
+        return rc;
+    }
+    return 1;
 }
 
 static int
@@ -231,7 +417,7 @@ crm_send_tls(gnutls_session_t * session, const char *buf, size_t len)
     int total_send;
 
     if (buf == NULL) {
-        return -1;
+        return -EINVAL;
     }
 
     total_send = len;
@@ -241,10 +427,13 @@ crm_send_tls(gnutls_session_t * session, const char *buf, size_t len)
         rc = gnutls_record_send(*session, unsent, len);
 
         if (rc == GNUTLS_E_INTERRUPTED || rc == GNUTLS_E_AGAIN) {
-            crm_debug("Retry");
+            crm_trace("Retrying to send %llu bytes",
+                      (unsigned long long) len);
 
         } else if (rc < 0) {
-            crm_err("Connection terminated rc = %d", rc);
+            crm_err("Connection terminated: %s " CRM_XS " rc=%d",
+                    gnutls_strerror(rc), rc);
+            rc = -ECONNABORTED;
             break;
 
         } else if (rc < len) {
@@ -270,7 +459,7 @@ crm_send_plaintext(int sock, const char *buf, size_t len)
     int total_send;
 
     if (buf == NULL) {
-        return -1;
+        return -EINVAL;
     }
     total_send = len;
 
@@ -279,6 +468,7 @@ crm_send_plaintext(int sock, const char *buf, size_t len)
   retry:
     rc = write(sock, unsent, len);
     if (rc < 0) {
+        rc = -errno;
         switch (errno) {
             case EINTR:
             case EAGAIN:
@@ -331,7 +521,7 @@ crm_remote_sendv(crm_remote_t * remote, struct iovec * iov, int iovs)
 int
 crm_remote_send(crm_remote_t * remote, xmlNode * msg)
 {
-    int rc = -1;
+    int rc = pcmk_ok;
     static uint64_t id = 0;
     char *xml_text = dump_xml_unformatted(msg);
 
@@ -339,8 +529,8 @@ crm_remote_send(crm_remote_t * remote, xmlNode * msg)
     struct crm_remote_header_v0 *header;
 
     if (xml_text == NULL) {
-        crm_err("Invalid XML, can not send msg");
-        return -1;
+        crm_err("Could not send remote message: no message provided");
+        return -EINVAL;
     }
 
     header = calloc(1, sizeof(struct crm_remote_header_v0));
@@ -362,7 +552,8 @@ crm_remote_send(crm_remote_t * remote, xmlNode * msg)
               (int)iov[0].iov_len, *(int*)(void*)xml_text);
     rc = crm_remote_sendv(remote, iov, 2);
     if (rc < 0) {
-        crm_err("Failed to send remote msg, rc = %d", rc);
+        crm_err("Could not send remote message: %s " CRM_XS " rc=%d",
+                pcmk_strerror(rc), rc);
     }
 
     free(iov[0].iov_base);
@@ -601,13 +792,16 @@ crm_remote_recv_once(crm_remote_t * remote)
 
 /*!
  * \internal
- * \brief Read data off the socket until at least one full message is present or timeout occures.
- * \retval TRUE message read
- * \retval FALSE full message not read
+ * \brief Read message(s) from a remote connection
+ *
+ * \param[in]  remote         Remote connection to read
+ * \param[in]  total_timeout  Fail if message not read in this time (ms)
+ * \param[out] disconnected   Will be set to 1 if disconnect detected
+ *
+ * \return TRUE if at least one full message read, FALSE otherwise
  */
-
 gboolean
-crm_remote_recv(crm_remote_t * remote, int total_timeout /*ms */ , int *disconnected)
+crm_remote_recv(crm_remote_t *remote, int total_timeout, int *disconnected)
 {
     int rc;
     time_t start = time(NULL);
@@ -623,28 +817,32 @@ crm_remote_recv(crm_remote_t * remote, int total_timeout /*ms */ , int *disconne
     remaining_timeout = total_timeout;
     while ((remaining_timeout > 0) && !(*disconnected)) {
 
-        /* read some more off the tls buffer if we still have time left. */
-        crm_trace("waiting to receive remote msg, starting timeout %d, remaining_timeout %d",
-                  total_timeout, remaining_timeout);
+        crm_trace("Waiting for remote data (%d of %d ms timeout remaining)",
+                  remaining_timeout, total_timeout);
         rc = crm_remote_ready(remote, remaining_timeout);
 
         if (rc == 0) {
-            crm_err("poll timed out (%d ms) while waiting to receive msg", remaining_timeout);
+            crm_err("Timed out (%d ms) while waiting for remote data",
+                    remaining_timeout);
             return FALSE;
 
-        } else if(rc < 0) {
-            crm_debug("could not poll: %s (%d)", pcmk_strerror(rc), rc);
+        } else if (rc < 0) {
+            crm_debug("Wait for remote data aborted, will try again: %s "
+                      CRM_XS " rc=%d", pcmk_strerror(rc), rc);
 
         } else {
             rc = crm_remote_recv_once(remote);
-            if(rc > 0) {
+            if (rc > 0) {
                 return TRUE;
+            } else if (rc == -EAGAIN) {
+                crm_trace("Still waiting for remote data");
             } else if (rc < 0) {
-                crm_debug("recv() failed: %s (%d)", pcmk_strerror(rc), rc);
+                crm_debug("Could not receive remote data: %s " CRM_XS " rc=%d",
+                          pcmk_strerror(rc), rc);
             }
         }
 
-        if(rc == -ENOTCONN) {
+        if (rc == -ENOTCONN) {
             *disconnected = 1;
             return FALSE;
         }
@@ -668,7 +866,7 @@ static gboolean
 check_connect_finished(gpointer userdata)
 {
     struct tcp_async_cb_data *cb_data = userdata;
-    int rc = 0;
+    int cb_arg = 0; // socket fd on success, -errno on error
     int sock = cb_data->sock;
     int error = 0;
 
@@ -685,63 +883,62 @@ check_connect_finished(gpointer userdata)
     wset = rset;
 
     crm_trace("fd %d: checking to see if connect finished", sock);
-    rc = select(sock + 1, &rset, &wset, NULL, &ts);
+    cb_arg = select(sock + 1, &rset, &wset, NULL, &ts);
 
-    if (rc < 0) {
-        rc = errno;
+    if (cb_arg < 0) {
+        cb_arg = -errno;
         if ((errno == EINPROGRESS) || (errno == EAGAIN)) {
             /* reschedule if there is still time left */
             if ((time(NULL) - cb_data->start) < (cb_data->timeout / 1000)) {
                 goto reschedule;
             } else {
-                rc = -ETIMEDOUT;
+                cb_arg = -ETIMEDOUT;
             }
         }
-        crm_trace("fd %d: select failed %d connect dispatch ", sock, rc);
+        crm_trace("fd %d: select failed %d connect dispatch ", sock, cb_arg);
         goto dispatch_done;
-    } else if (rc == 0) {
+    } else if (cb_arg == 0) {
         if ((time(NULL) - cb_data->start) < (cb_data->timeout / 1000)) {
             goto reschedule;
         }
         crm_debug("fd %d: timeout during select", sock);
-        rc = -ETIMEDOUT;
+        cb_arg = -ETIMEDOUT;
         goto dispatch_done;
     } else {
         crm_trace("fd %d: select returned success", sock);
-        rc = 0;
+        cb_arg = 0;
     }
 
     /* can we read or write to the socket now? */
     if (FD_ISSET(sock, &rset) || FD_ISSET(sock, &wset)) {
         if (getsockopt(sock, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
+            cb_arg = -errno;
             crm_trace("fd %d: call to getsockopt failed", sock);
-            rc = -1;
             goto dispatch_done;
         }
-
         if (error) {
             crm_trace("fd %d: error returned from getsockopt: %d", sock, error);
-            rc = -1;
+            cb_arg = -error;
             goto dispatch_done;
         }
     } else {
         crm_trace("neither read nor write set after select");
-        rc = -1;
+        cb_arg = -EAGAIN;
         goto dispatch_done;
     }
 
   dispatch_done:
-    if (!rc) {
+    if (!cb_arg) {
         crm_trace("fd %d: connected", sock);
         /* Success, set the return code to the sock to report to the callback */
-        rc = cb_data->sock;
+        cb_arg = cb_data->sock;
         cb_data->sock = 0;
     } else {
         close(sock);
     }
 
     if (cb_data->callback) {
-        cb_data->callback(cb_data->userdata, rc);
+        cb_data->callback(cb_data->userdata, cb_arg);
     }
     free(cb_data);
     return FALSE;
@@ -758,21 +955,21 @@ internal_tcp_connect_async(int sock,
                            int *timer_id, void *userdata, void (*callback) (void *userdata, int sock))
 {
     int rc = 0;
-    int flag = 0;
     int interval = 500;
     int timer;
     struct tcp_async_cb_data *cb_data = NULL;
 
-    if ((flag = fcntl(sock, F_GETFL)) >= 0) {
-        if (fcntl(sock, F_SETFL, flag | O_NONBLOCK) < 0) {
-            crm_err("fcntl() write failed");
-            return -1;
-        }
+    rc = crm_set_nonblocking(sock);
+    if (rc < 0) {
+        crm_warn("Could not set socket non-blocking: %s " CRM_XS " rc=%d",
+                 pcmk_strerror(rc), rc);
+        close(sock);
+        return -1;
     }
 
     rc = connect(sock, addr, addrlen);
-
     if (rc < 0 && (errno != EINPROGRESS) && (errno != EAGAIN)) {
+        crm_perror(LOG_WARNING, "connect");
         return -1;
     }
 
@@ -800,7 +997,8 @@ internal_tcp_connect_async(int sock,
      * At some point we should figure out a way to use a mainloop fd callback for this.
      * Something about the way mainloop is currently polling prevents this from working at the
      * moment though. */
-    crm_trace("fd %d: scheduling to check if connect finished in %dms second", sock, interval);
+    crm_trace("Scheduling check in %dms for whether connect to fd %d finished",
+              interval, sock);
     timer = g_timeout_add(interval, check_connect_finished, cb_data);
     if (timer_id) {
         *timer_id = timer;
@@ -812,57 +1010,68 @@ internal_tcp_connect_async(int sock,
 static int
 internal_tcp_connect(int sock, const struct sockaddr *addr, socklen_t addrlen)
 {
-    int flag = 0;
     int rc = connect(sock, addr, addrlen);
 
-    if (rc == 0) {
-        if ((flag = fcntl(sock, F_GETFL)) >= 0) {
-            if (fcntl(sock, F_SETFL, flag | O_NONBLOCK) < 0) {
-                crm_err("fcntl() write failed");
-                return -1;
-            }
-        }
+    if (rc < 0) {
+        rc = -errno;
+        crm_warn("Could not connect socket: %s " CRM_XS " rc=%d",
+                 pcmk_strerror(rc), rc);
+        return rc;
     }
 
-    return rc;
+    rc = crm_set_nonblocking(sock);
+    if (rc < 0) {
+        crm_warn("Could not set socket non-blocking: %s " CRM_XS " rc=%d",
+                 pcmk_strerror(rc), rc);
+        return rc;
+    }
+
+    return pcmk_ok;
 }
 
 /*!
  * \internal
- * \brief tcp connection to server at specified port
- * \retval negative, failed to connect.
- * \retval positive, sock fd
+ * \brief Connect to server at specified TCP port
+ *
+ * \param[in]  host      Name of server to connect to
+ * \param[in]  port      Server port to connect to
+ * \param[in]  timeout   Report error if not connected in this many milliseconds
+ * \param[out] timer_id  If non-NULL, will be set to timer ID, if asynchronous
+ * \param[in]  userdata  Data to pass to callback, if asynchronous
+ * \param[in]  callback  If non-NULL, connect asynchronously then call this
+ *
+ * \return File descriptor of connected socket on success, -ENOTCONN otherwise
  */
 int
-crm_remote_tcp_connect_async(const char *host, int port, int timeout, /*ms */
-                             int *timer_id, void *userdata, void (*callback) (void *userdata, int sock))
+crm_remote_tcp_connect_async(const char *host, int port, int timeout,
+                             int *timer_id, void *userdata,
+                             void (*callback) (void *userdata, int sock))
 {
-    char buffer[256];
+    char buffer[INET6_ADDRSTRLEN];
     struct addrinfo *res = NULL;
     struct addrinfo *rp = NULL;
     struct addrinfo hints;
     const char *server = host;
     int ret_ga;
-    int sock = -1;
+    int sock = -ENOTCONN;
 
-    /* getaddrinfo */
+    // Get host's IP address(es)
     memset(&hints, 0, sizeof(struct addrinfo));
     hints.ai_family = AF_UNSPEC;        /* Allow IPv4 or IPv6 */
     hints.ai_socktype = SOCK_STREAM;
     hints.ai_flags = AI_CANONNAME;
-
-    crm_debug("Looking up %s", server);
     ret_ga = getaddrinfo(server, NULL, &hints, &res);
     if (ret_ga) {
-        crm_err("getaddrinfo: %s", gai_strerror(ret_ga));
-        return -1;
+        crm_err("Unable to get IP address info for %s: %s",
+                server, gai_strerror(ret_ga));
+        goto async_cleanup;
     }
-
     if (!res || !res->ai_addr) {
-        crm_err("getaddrinfo failed");
+        crm_err("Unable to get IP address info for %s: no result", server);
         goto async_cleanup;
     }
 
+    // getaddrinfo() returns a list of host's addresses, try them in order
     for (rp = res; rp != NULL; rp = rp->ai_next) {
         struct sockaddr *addr = rp->ai_addr;
 
@@ -873,30 +1082,27 @@ crm_remote_tcp_connect_async(const char *host, int port, int timeout, /*ms */
         if (rp->ai_canonname) {
             server = res->ai_canonname;
         }
-        crm_debug("Got address %s for %s", server, host);
+        crm_debug("Got canonical name %s for %s", server, host);
 
-        /* create socket */
         sock = socket(rp->ai_family, SOCK_STREAM, IPPROTO_TCP);
         if (sock == -1) {
-            crm_err("Socket creation failed for remote client connection.");
+            crm_perror(LOG_WARNING, "creating socket for connection to %s",
+                       server);
+            sock = -ENOTCONN;
             continue;
         }
 
-        memset(buffer, 0, DIMOF(buffer));
+        /* Set port appropriately for address family */
+        /* (void*) casts avoid false-positive compiler alignment warnings */
         if (addr->sa_family == AF_INET6) {
-            struct sockaddr_in6 *addr_in = (struct sockaddr_in6 *)(void*)addr;
-
-            addr_in->sin6_port = htons(port);
-            inet_ntop(addr->sa_family, &addr_in->sin6_addr, buffer, DIMOF(buffer));
-
+            ((struct sockaddr_in6 *)(void*)addr)->sin6_port = htons(port);
         } else {
-            struct sockaddr_in *addr_in = (struct sockaddr_in *)(void*)addr;
-
-            addr_in->sin_port = htons(port);
-            inet_ntop(addr->sa_family, &addr_in->sin_addr, buffer, DIMOF(buffer));
+            ((struct sockaddr_in *)(void*)addr)->sin_port = htons(port);
         }
 
-        crm_info("Attempting to connect to remote server at %s:%d", buffer, port);
+        memset(buffer, 0, DIMOF(buffer));
+        crm_sockaddr2str(addr, buffer);
+        crm_info("Attempting TCP connection to %s:%d", buffer, port);
 
         if (callback) {
             if (internal_tcp_connect_async
@@ -904,14 +1110,12 @@ crm_remote_tcp_connect_async(const char *host, int port, int timeout, /*ms */
                 goto async_cleanup; /* Success for now, we'll hear back later in the callback */
             }
 
-        } else {
-            if (internal_tcp_connect(sock, rp->ai_addr, rp->ai_addrlen) == 0) {
-                break;          /* Success */
-            }
+        } else if (internal_tcp_connect(sock, rp->ai_addr, rp->ai_addrlen) == 0) {
+            break;          /* Success */
         }
 
         close(sock);
-        sock = -1;
+        sock = -ENOTCONN;
     }
 
 async_cleanup:
@@ -928,29 +1132,33 @@ crm_remote_tcp_connect(const char *host, int port)
     return crm_remote_tcp_connect_async(host, port, -1, NULL, NULL, NULL);
 }
 
-
-/* Convert a struct sockaddr address to a string, IPv4 and IPv6: */
-
-static char *
-get_ip_str(const struct sockaddr_storage * sa, char * s, size_t maxlen)
+/*!
+ * \brief Convert an IP address (IPv4 or IPv6) to a string for logging
+ *
+ * \param[in]  sa  Socket address for IP
+ * \param[out] s   Storage for at least INET6_ADDRSTRLEN bytes
+ *
+ * \note sa The socket address can be a pointer to struct sockaddr_in (IPv4),
+ *          struct sockaddr_in6 (IPv6) or struct sockaddr_storage (either),
+ *          as long as its sa_family member is set correctly.
+ */
+void
+crm_sockaddr2str(void *sa, char *s)
 {
-    switch(((struct sockaddr *)sa)->sa_family) {
+    switch (((struct sockaddr*)sa)->sa_family) {
         case AF_INET:
             inet_ntop(AF_INET, &(((struct sockaddr_in *)sa)->sin_addr),
-                      s, maxlen);
+                      s, INET6_ADDRSTRLEN);
             break;
 
         case AF_INET6:
             inet_ntop(AF_INET6, &(((struct sockaddr_in6 *)sa)->sin6_addr),
-                      s, maxlen);
+                      s, INET6_ADDRSTRLEN);
             break;
 
         default:
-            strncpy(s, "Unknown AF", maxlen);
-            return NULL;
+            strcpy(s, "<invalid>");
     }
-
-    return s;
 }
 
 int
@@ -958,7 +1166,6 @@ crm_remote_accept(int ssock)
 {
     int csock = 0;
     int rc = 0;
-    int flag = 0;
     unsigned laddr = 0;
     struct sockaddr_storage addr;
     char addr_str[INET6_ADDRSTRLEN];
@@ -971,7 +1178,7 @@ crm_remote_accept(int ssock)
     laddr = sizeof(addr);
     memset(&addr, 0, sizeof(addr));
     csock = accept(ssock, (struct sockaddr *)&addr, &laddr);
-    get_ip_str(&addr, addr_str, INET6_ADDRSTRLEN);
+    crm_sockaddr2str(&addr, addr_str);
     crm_info("New remote connection from %s", addr_str);
 
     if (csock == -1) {
@@ -979,16 +1186,12 @@ crm_remote_accept(int ssock)
         return -1;
     }
 
-    if ((flag = fcntl(csock, F_GETFL)) >= 0) {
-        if ((rc = fcntl(csock, F_SETFL, flag | O_NONBLOCK)) < 0) {
-            crm_err("fcntl() write failed");
-            close(csock);
-            return rc;
-        }
-    } else {
-        crm_err("fcntl() read failed");
+    rc = crm_set_nonblocking(csock);
+    if (rc < 0) {
+        crm_err("Could not set socket non-blocking: %s " CRM_XS " rc=%d",
+                pcmk_strerror(rc), rc);
         close(csock);
-        return flag;
+        return rc;
     }
 
 #ifdef TCP_USER_TIMEOUT
@@ -1007,3 +1210,31 @@ crm_remote_accept(int ssock)
 
     return csock;
 }
+
+/*!
+ * \brief Get the default remote connection TCP port on this host
+ *
+ * \return Remote connection TCP port number
+ */
+int
+crm_default_remote_port()
+{
+    static int port = 0;
+
+    if (port == 0) {
+        const char *env = getenv("PCMK_remote_port");
+
+        if (env) {
+            errno = 0;
+            port = strtol(env, NULL, 10);
+            if (errno || (port < 1) || (port > 65535)) {
+                crm_warn("Environment variable PCMK_remote_port has invalid value '%s', using %d instead",
+                         env, DEFAULT_REMOTE_PORT);
+                port = DEFAULT_REMOTE_PORT;
+            }
+        } else {
+            port = DEFAULT_REMOTE_PORT;
+        }
+    }
+    return port;
+}
