@@ -1,19 +1,10 @@
 /*
- * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
+ * Copyright 2004-2019 the Pacemaker project contributors
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * The version control history for this file may have further details.
  *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * This source code is licensed under the GNU Lesser General Public License
+ * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
  */
 
 #include <crm_internal.h>
@@ -717,16 +708,21 @@ crm_abort(const char *file, const char *function, int line,
 int
 crm_pid_active(long pid, const char *daemon)
 {
+    static int last_asked_pid = 0;  /* log spam prevention */
+#if SUPPORT_PROCFS
     static int have_proc_pid = 0;
+#else
+    static int have_proc_pid = -1;
+#endif
+    int rc = 0;
 
-    if(have_proc_pid == 0) {
+    if (have_proc_pid == 0) {
+        /* evaluation of /proc/PID/exe applicability via self-introspection */
         char proc_path[PATH_MAX], exe_path[PATH_MAX];
-
-        /* check to make sure pid hasn't been reused by another process */
-        snprintf(proc_path, sizeof(proc_path), "/proc/%lu/exe", (long unsigned int)getpid());
-
+        snprintf(proc_path, sizeof(proc_path), "/proc/%lu/exe",
+                 (long unsigned int) getpid());
         have_proc_pid = 1;
-        if(readlink(proc_path, exe_path, PATH_MAX - 1) < 0) {
+        if (readlink(proc_path, exe_path, sizeof(exe_path) - 1) < 0) {
             have_proc_pid = -1;
         }
     }
@@ -734,40 +730,52 @@ crm_pid_active(long pid, const char *daemon)
     if (pid <= 0) {
         return -1;
 
-    } else if (kill(pid, 0) < 0 && errno == ESRCH) {
-        return 0;
+    } else if ((rc = kill(pid, 0)) < 0 && errno == ESRCH) {
+        return 0;  /* no such PID detected */
 
-    } else if(daemon == NULL || have_proc_pid == -1) {
-        return 1;
+    } else if (rc < 0 && (daemon == NULL || have_proc_pid == -1)) {
+        if (last_asked_pid != pid) {
+            crm_info("Cannot examine PID %ld: %s", pid, strerror(errno));
+            last_asked_pid = pid;
+        }
+        return -2;  /* errno != ESRCH */
 
-    } else {
-        int rc = 0;
+    } else if (rc == 0 && (daemon == NULL || have_proc_pid == -1)) {
+        return 1;  /* kill as the only indicator, cannot double check */
+
+    } else if (daemon != NULL) {
+        /* make sure PID hasn't been reused by another process
+           XXX: might still be just a zombie, which could confuse decisions */
+        bool checked_through_kill = (rc == 0);
         char proc_path[PATH_MAX], exe_path[PATH_MAX], myexe_path[PATH_MAX];
+        snprintf(proc_path, sizeof(proc_path), "/proc/%ld/exe", pid);
 
-        /* check to make sure pid hasn't been reused by another process */
-        snprintf(proc_path, sizeof(proc_path), "/proc/%lu/exe", pid);
-
-        rc = readlink(proc_path, exe_path, PATH_MAX - 1);
-        if (rc < 0 && errno == EACCES) {
-            crm_perror(LOG_INFO, "Could not read from %s", proc_path);
-            return 1;
+        rc = readlink(proc_path, exe_path, sizeof(exe_path) - 1);
+        if ((rc < 0) && (errno == EACCES)) {
+            if (last_asked_pid != pid) {
+                crm_info("Could not read from %s: %s", proc_path,
+                         strerror(errno));
+                last_asked_pid = pid;
+            }
+            return checked_through_kill ? 1 : -2;
         } else if (rc < 0) {
-            crm_perror(LOG_ERR, "Could not read from %s", proc_path);
-            return 0;
+            if (last_asked_pid != pid) {
+                crm_err("Could not read from %s: %s (%d)", proc_path,
+                        strerror(errno), errno);
+                last_asked_pid = pid;
+            }
+            return 0;  /* most likely errno == ENOENT */
         }
-        
+        exe_path[rc] = '\0';
 
-        exe_path[rc] = 0;
-
-        if(daemon[0] != '/') {
-            rc = snprintf(myexe_path, sizeof(proc_path), CRM_DAEMON_DIR"/%s", daemon);
-            myexe_path[rc] = 0;
+        if (daemon[0] != '/') {
+            rc = snprintf(myexe_path, sizeof(myexe_path), CRM_DAEMON_DIR"/%s",
+                          daemon);
         } else {
-            rc = snprintf(myexe_path, sizeof(proc_path), "%s", daemon);
-            myexe_path[rc] = 0;
+            rc = snprintf(myexe_path, sizeof(myexe_path), "%s", daemon);
         }
-        
-        if (strcmp(exe_path, myexe_path) == 0) {
+
+        if (rc > 0 && rc < sizeof(myexe_path) && !strcmp(exe_path, myexe_path)) {
             return 1;
         }
     }
@@ -1223,7 +1231,8 @@ attrd_ipc_server_init(qb_ipcs_service_t **ipcs, struct qb_ipcs_service_handlers
 void
 stonith_ipc_server_init(qb_ipcs_service_t **ipcs, struct qb_ipcs_service_handlers *cb)
 {
-    *ipcs = mainloop_add_ipc_server("stonith-ng", QB_IPC_NATIVE, cb);
+    *ipcs = mainloop_add_ipc_server_with_prio("stonith-ng", QB_IPC_NATIVE, cb,
+                                              QB_LOOP_HIGH);
 
     if (*ipcs == NULL) {
         crm_err("Failed to create stonith-ng servers: exiting and inhibiting respawn.");
@@ -1232,119 +1241,6 @@ stonith_ipc_server_init(qb_ipcs_service_t **ipcs, struct qb_ipcs_service_handler
     }
 }
 
-bool
-pcmk_acl_required(const char *user) 
-{
-#if ENABLE_ACL
-    if(user == NULL || strlen(user) == 0) {
-        crm_trace("no user set");
-        return FALSE;
-
-    } else if (strcmp(user, CRM_DAEMON_USER) == 0) {
-        return FALSE;
-
-    } else if (strcmp(user, "root") == 0) {
-        return FALSE;
-    }
-    crm_trace("acls required for %s", user);
-    return TRUE;
-#else
-    crm_trace("acls not supported");
-    return FALSE;
-#endif
-}
-
-#if ENABLE_ACL
-char *
-uid2username(uid_t uid)
-{
-    struct passwd *pwent = getpwuid(uid);
-
-    if (pwent == NULL) {
-        crm_perror(LOG_ERR, "Cannot get password entry of uid: %d", uid);
-        return NULL;
-
-    } else {
-        return strdup(pwent->pw_name);
-    }
-}
-
-const char *
-crm_acl_get_set_user(xmlNode * request, const char *field, const char *peer_user)
-{
-    /* field is only checked for backwards compatibility */
-    static const char *effective_user = NULL;
-    const char *requested_user = NULL;
-    const char *user = NULL;
-
-    if(effective_user == NULL) {
-        effective_user = uid2username(geteuid());
-    }
-
-    requested_user = crm_element_value(request, XML_ACL_TAG_USER);
-    if(requested_user == NULL) {
-        requested_user = crm_element_value(request, field);
-    }
-
-    if (is_privileged(effective_user) == FALSE) {
-        /* We're not running as a privileged user, set or overwrite any existing value for $XML_ACL_TAG_USER */
-        user = effective_user;
-
-    } else if(peer_user == NULL && requested_user == NULL) {
-        /* No user known or requested, use 'effective_user' and make sure one is set for the request */
-        user = effective_user;
-
-    } else if(peer_user == NULL) {
-        /* No user known, trusting 'requested_user' */
-        user = requested_user;
-
-    } else if (is_privileged(peer_user) == FALSE) {
-        /* The peer is not a privileged user, set or overwrite any existing value for $XML_ACL_TAG_USER */
-        user = peer_user;
-
-    } else if (requested_user == NULL) {
-        /* Even if we're privileged, make sure there is always a value set */
-        user = peer_user;
-
-    } else {
-        /* Legal delegation to 'requested_user' */
-        user = requested_user;
-    }
-
-    // This requires pointer comparison, not string comparison
-    if(user != crm_element_value(request, XML_ACL_TAG_USER)) {
-        crm_xml_add(request, XML_ACL_TAG_USER, user);
-    }
-
-    if(field != NULL && user != crm_element_value(request, field)) {
-        crm_xml_add(request, field, user);
-    }
-
-    return requested_user;
-}
-
-void
-determine_request_user(const char *user, xmlNode * request, const char *field)
-{
-    /* Get our internal validation out of the way first */
-    CRM_CHECK(user != NULL && request != NULL && field != NULL, return);
-
-    /* If our peer is a privileged user, we might be doing something on behalf of someone else */
-    if (is_privileged(user) == FALSE) {
-        /* We're not a privileged user, set or overwrite any existing value for $field */
-        crm_xml_replace(request, field, user);
-
-    } else if (crm_element_value(request, field) == NULL) {
-        /* Even if we're privileged, make sure there is always a value set */
-        crm_xml_replace(request, field, user);
-
-/*  } else { Legal delegation */
-    }
-
-    crm_trace("Processing msg as user '%s'", crm_element_value(request, field));
-}
-#endif
-
 void *
 find_library_function(void **handle, const char *lib, const char *fn, gboolean fatal)
 {
@@ -1375,13 +1271,6 @@ find_library_function(void **handle, const char *lib, const char *fn, gboolean f
     return a_function;
 }
 
-void *
-convert_const_pointer(const void *ptr)
-{
-    /* Worst function ever */
-    return (void *)ptr;
-}
-
 #ifdef HAVE_UUID_UUID_H
 #  include <uuid/uuid.h>
 #endif
@@ -1456,95 +1345,3 @@ crm_gnutls_global_init(void)
     gnutls_global_init();
 }
 #endif
-
-char *
-crm_generate_ra_key(const char *class, const char *provider, const char *type)
-{
-    if (!class && !provider && !type) {
-        return NULL;
-    }
-
-    return crm_strdup_printf("%s%s%s:%s",
-                             (class? class : ""),
-                             (provider? ":" : ""), (provider? provider : ""),
-                             (type? type : ""));
-}
-
-/*!
- * \brief Check whether a resource standard requires a provider to be specified
- *
- * \param[in] standard  Standard name
- *
- * \return TRUE if standard requires a provider, FALSE otherwise
- */
-bool
-crm_provider_required(const char *standard)
-{
-    CRM_CHECK(standard != NULL, return FALSE);
-
-    /* @TODO
-     * - this should probably be case-sensitive, but isn't,
-     *   for backward compatibility
-     * - it might be nice to keep standards' capabilities (supports provider,
-     *   master/slave, etc.) as structured data somewhere
-     */
-    if (!strcasecmp(standard, PCMK_RESOURCE_CLASS_OCF)) {
-        return TRUE;
-    }
-    return FALSE;
-}
-
-/*!
- * \brief Parse a "standard[:provider]:type" agent specification
- *
- * \param[in]  spec      Agent specification
- * \param[out] standard  Newly allocated memory containing agent standard (or NULL)
- * \param[out] provider  Newly allocated memory containing agent provider (or NULL)
- * \param[put] type      Newly allocated memory containing agent type (or NULL)
- *
- * \return pcmk_ok if the string could be parsed, -EINVAL otherwise
- *
- * \note It is acceptable for the type to contain a ':' if the standard supports
- *       that. For example, systemd supports the form "systemd:UNIT@A:B".
- * \note It is the caller's responsibility to free the returned values.
- */
-int
-crm_parse_agent_spec(const char *spec, char **standard, char **provider,
-                     char **type)
-{
-    char *colon;
-
-    CRM_CHECK(spec && standard && provider && type, return -EINVAL);
-    *standard = NULL;
-    *provider = NULL;
-    *type = NULL;
-
-    colon = strchr(spec, ':');
-    if ((colon == NULL) || (colon == spec)) {
-        return -EINVAL;
-    }
-
-    *standard = calloc(colon - spec + 1, sizeof(char));
-    strncpy(*standard, spec, colon - spec);
-    spec = colon + 1;
-
-    if (crm_provider_required(*standard)) {
-        colon = strchr(spec, ':');
-        if ((colon == NULL) || (colon == spec)) {
-            free(*standard);
-            return -EINVAL;
-        }
-        *provider = calloc(colon - spec + 1, sizeof(char));
-        strncpy(*provider, spec, colon - spec);
-        spec = colon + 1;
-    }
-
-    if (*spec == '\0') {
-        free(*standard);
-        free(*provider);
-        return -EINVAL;
-    }
-
-    *type = strdup(spec);
-    return pcmk_ok;
-}
