@@ -1,19 +1,8 @@
 /*
- * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
+ * Copyright 2004-2018 Andrew Beekhof <andrew@beekhof.net>
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * This source code is licensed under the GNU Lesser General Public License
+ * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
  */
 
 #include <crm_internal.h>
@@ -22,43 +11,28 @@
 #include <crm/pengine/status.h>
 #include <crm/pengine/internal.h>
 #include <unpack.h>
+#include <pe_status_private.h>
 #include <crm/msg_xml.h>
 
 #define VARIANT_CLONE 1
 #include "./variant.h"
 
-void force_non_unique_clone(resource_t * rsc, const char *rid, pe_working_set_t * data_set);
-resource_t *create_child_clone(resource_t * rsc, int sub_id, pe_working_set_t * data_set);
-
-static void
-mark_as_orphan(resource_t * rsc)
-{
-    GListPtr gIter = rsc->children;
-
-    set_bit(rsc->flags, pe_rsc_orphan);
-
-    for (; gIter != NULL; gIter = gIter->next) {
-        resource_t *child = (resource_t *) gIter->data;
-
-        mark_as_orphan(child);
-    }
-}
-
 void
-force_non_unique_clone(resource_t * rsc, const char *rid, pe_working_set_t * data_set)
+pe__force_anon(const char *standard, pe_resource_t *rsc, const char *rid,
+               pe_working_set_t *data_set)
 {
     if (pe_rsc_is_clone(rsc)) {
         clone_variant_data_t *clone_data = NULL;
 
         get_clone_variant_data(clone_data, rsc);
 
-        crm_config_warn("Clones %s contains non-OCF resource %s and so "
-                        "can only be used as an anonymous clone. "
-                        "Set the " XML_RSC_ATTR_UNIQUE " meta attribute to false", rsc->id, rid);
+        pe_warn("Ignoring " XML_RSC_ATTR_UNIQUE " for %s because %s resources "
+                "such as %s can be used only as anonymous clones",
+                rsc->id, standard, rid);
 
         clone_data->clone_node_max = 1;
-        clone_data->clone_max = g_list_length(data_set->nodes);
-        clear_bit_recursive(rsc, pe_rsc_unique);
+        clone_data->clone_max = QB_MIN(clone_data->clone_max,
+                                       g_list_length(data_set->nodes));
     }
 }
 
@@ -80,8 +54,8 @@ find_clone_instance(resource_t * rsc, const char *sub_id, pe_working_set_t * dat
     return child;
 }
 
-resource_t *
-create_child_clone(resource_t * rsc, int sub_id, pe_working_set_t * data_set)
+pe_resource_t *
+pe__create_clone_child(pe_resource_t *rsc, pe_working_set_t *data_set)
 {
     gboolean as_orphan = FALSE;
     char *inc_num = NULL;
@@ -94,11 +68,13 @@ create_child_clone(resource_t * rsc, int sub_id, pe_working_set_t * data_set)
 
     CRM_CHECK(clone_data->xml_obj_child != NULL, return FALSE);
 
-    if (sub_id < 0) {
+    if (clone_data->total_clones >= clone_data->clone_max) {
+        // If we've already used all available instances, this is an orphan
         as_orphan = TRUE;
-        sub_id = clone_data->total_clones;
     }
-    inc_num = crm_itoa(sub_id);
+
+    // Allocate instance numbers in numerical order (starting at 0)
+    inc_num = crm_itoa(clone_data->total_clones);
     inc_max = crm_itoa(clone_data->clone_max);
 
     child_copy = copy_xml(clone_data->xml_obj_child);
@@ -117,12 +93,11 @@ create_child_clone(resource_t * rsc, int sub_id, pe_working_set_t * data_set)
     pe_rsc_trace(child_rsc, "Setting clone attributes for: %s", child_rsc->id);
     rsc->children = g_list_append(rsc->children, child_rsc);
     if (as_orphan) {
-        mark_as_orphan(child_rsc);
+        set_bit_recursive(child_rsc, pe_rsc_orphan);
     }
 
     add_hash_param(child_rsc->meta, XML_RSC_ATTR_INCARNATION_MAX, inc_max);
-
-    print_resource(LOG_DEBUG_3, "Added ", child_rsc, FALSE);
+    pe_rsc_trace(rsc, "Added %s instance %s", rsc->id, child_rsc->id);
 
   bail:
     free(inc_num);
@@ -196,7 +171,7 @@ clone_unpack(resource_t * rsc, pe_working_set_t * data_set)
                  is_set(rsc->flags, pe_rsc_unique) ? "true" : "false");
 
     // Clones may contain a single group or primitive
-    for (a_child = __xml_first_child(xml_obj); a_child != NULL;
+    for (a_child = __xml_first_child_element(xml_obj); a_child != NULL;
          a_child = __xml_next_element(a_child)) {
 
         if (crm_str_eq((const char *)a_child->name, XML_CIB_TAG_RESOURCE, TRUE)
@@ -221,24 +196,27 @@ clone_unpack(resource_t * rsc, pe_working_set_t * data_set)
         add_hash_param(rsc->meta, XML_RSC_ATTR_STICKINESS, "1");
     }
 
-    pe_rsc_trace(rsc, "\tClone is unique (fixed): %s",
-                 is_set(rsc->flags, pe_rsc_unique) ? "true" : "false");
     clone_data->notify_confirm = is_set(rsc->flags, pe_rsc_notify);
+    /* This ensures that the globally-unique value always exists for children to
+     * inherit when being unpacked, as well as in resource agents' environment.
+     */
     add_hash_param(rsc->meta, XML_RSC_ATTR_UNIQUE,
                    is_set(rsc->flags, pe_rsc_unique) ? XML_BOOLEAN_TRUE : XML_BOOLEAN_FALSE);
 
-    for (lpc = 0; lpc < clone_data->clone_max; lpc++) {
-        if (create_child_clone(rsc, lpc, data_set) == NULL) {
+    if (clone_data->clone_max <= 0) {
+        /* Create one child instance so that unpack_find_resource() will hook up
+         * any orphans up to the parent correctly.
+         */
+        if (pe__create_clone_child(rsc, data_set) == NULL) {
             return FALSE;
         }
-    }
 
-    if (clone_data->clone_max == 0) {
-        /* create one so that unpack_find_resource() will hook up
-         * any orphans up to the parent correctly
-         */
-        if (create_child_clone(rsc, -1, data_set) == NULL) {
-            return FALSE;
+    } else {
+        // Create a child instance for each available instance number
+        for (lpc = 0; lpc < clone_data->clone_max; lpc++) {
+            if (pe__create_clone_child(rsc, data_set) == NULL) {
+                return FALSE;
+            }
         }
     }
 
@@ -426,51 +404,57 @@ clone_print(resource_t * rsc, const char *pre_text, long options, void *print_da
     for (; gIter != NULL; gIter = gIter->next) {
         gboolean print_full = FALSE;
         resource_t *child_rsc = (resource_t *) gIter->data;
+        gboolean partially_active = child_rsc->fns->active(child_rsc, FALSE);
 
         if (options & pe_print_clone_details) {
             print_full = TRUE;
         }
 
-        if (child_rsc->fns->active(child_rsc, FALSE) == FALSE) {
-            /* Inactive clone */
-            if (is_set(child_rsc->flags, pe_rsc_orphan)) {
-                continue;
-
-            } else if (is_set(rsc->flags, pe_rsc_unique)) {
+        if (is_set(rsc->flags, pe_rsc_unique)) {
+            // Print individual instance when unique (except stopped orphans)
+            if (partially_active || is_not_set(rsc->flags, pe_rsc_orphan)) {
                 print_full = TRUE;
+            }
 
-            } else if (is_not_set(options, pe_print_clone_active)) {
+        // Everything else in this block is for anonymous clones
+
+        } else if (is_set(options, pe_print_pending)
+                   && (child_rsc->pending_task != NULL)
+                   && strcmp(child_rsc->pending_task, "probe")) {
+            // Print individual instance when non-probe action is pending
+            print_full = TRUE;
+
+        } else if (partially_active == FALSE) {
+            // List stopped instances when requested (except orphans)
+            if (is_not_set(child_rsc->flags, pe_rsc_orphan)
+                && is_not_set(options, pe_print_clone_active)) {
                 stopped_list = add_list_element(stopped_list, child_rsc->id);
             }
 
-        } else if (is_set_recursive(child_rsc, pe_rsc_unique, TRUE)
-                   || is_set_recursive(child_rsc, pe_rsc_orphan, TRUE)
+        } else if (is_set_recursive(child_rsc, pe_rsc_orphan, TRUE)
                    || is_set_recursive(child_rsc, pe_rsc_managed, FALSE) == FALSE
                    || is_set_recursive(child_rsc, pe_rsc_failed, TRUE)) {
 
-            /* Unique, unmanaged or failed clone */
-            print_full = TRUE;
-
-        } else if (is_set(options, pe_print_pending) && child_rsc->pending_task != NULL) {
-            /* In a pending state */
+            // Print individual instance when active orphaned/unmanaged/failed
             print_full = TRUE;
 
         } else if (child_rsc->fns->active(child_rsc, TRUE)) {
-            /* Fully active anonymous clone */
+            // Instance of fully active anonymous clone
+
             node_t *location = child_rsc->fns->location(child_rsc, NULL, TRUE);
 
             if (location) {
+                // Instance is active on a single node
+
                 enum rsc_role_e a_role = child_rsc->fns->state(child_rsc, TRUE);
 
                 if (location->details->online == FALSE && location->details->unclean) {
                     print_full = TRUE;
 
                 } else if (a_role > RSC_ROLE_SLAVE) {
-                    /* And active on a single node as master */
                     master_list = g_list_append(master_list, location);
 
                 } else {
-                    /* And active on a single node as started/slave */
                     started_list = g_list_append(started_list, location);
                 }
 
@@ -480,7 +464,7 @@ clone_print(resource_t * rsc, const char *pre_text, long options, void *print_da
             }
 
         } else {
-            /* Partially active anonymous clone */
+            // Instance of partially active anonymous clone
             print_full = TRUE;
         }
 
@@ -584,14 +568,13 @@ clone_print(resource_t * rsc, const char *pre_text, long options, void *print_da
 void
 clone_free(resource_t * rsc)
 {
-    GListPtr gIter = rsc->children;
     clone_variant_data_t *clone_data = NULL;
 
     get_clone_variant_data(clone_data, rsc);
 
     pe_rsc_trace(rsc, "Freeing %s", rsc->id);
 
-    for (; gIter != NULL; gIter = gIter->next) {
+    for (GListPtr gIter = rsc->children; gIter != NULL; gIter = gIter->next) {
         resource_t *child_rsc = (resource_t *) gIter->data;
 
         CRM_ASSERT(child_rsc);
