@@ -1,19 +1,8 @@
 /*
- * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
+ * Copyright 2004-2018 Andrew Beekhof <andrew@beekhof.net>
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * This source code is licensed under the GNU Lesser General Public License
+ * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
  */
 
 #include <crm_internal.h>
@@ -28,34 +17,37 @@ void populate_hash(xmlNode * nvpair_list, GHashTable * hash, const char **attrs,
 
 resource_object_functions_t resource_class_functions[] = {
     {
-     native_unpack,
-     native_find_rsc,
-     native_parameter,
-     native_print,
-     native_active,
-     native_resource_state,
-     native_location,
-     native_free
+         native_unpack,
+         native_find_rsc,
+         native_parameter,
+         native_print,
+         native_active,
+         native_resource_state,
+         native_location,
+         native_free,
+         pe__count_common,
     },
     {
-     group_unpack,
-     native_find_rsc,
-     native_parameter,
-     group_print,
-     group_active,
-     group_resource_state,
-     native_location,
-     group_free
+         group_unpack,
+         native_find_rsc,
+         native_parameter,
+         group_print,
+         group_active,
+         group_resource_state,
+         native_location,
+         group_free,
+         pe__count_common,
     },
     {
-     clone_unpack,
-     native_find_rsc,
-     native_parameter,
-     clone_print,
-     clone_active,
-     clone_resource_state,
-     native_location,
-     clone_free
+         clone_unpack,
+         native_find_rsc,
+         native_parameter,
+         clone_print,
+         clone_active,
+         clone_resource_state,
+         native_location,
+         clone_free,
+         pe__count_common,
     },
     {
      master_unpack,
@@ -65,17 +57,19 @@ resource_object_functions_t resource_class_functions[] = {
      clone_active,
      clone_resource_state,
      native_location,
-     clone_free
+         clone_free,
+         pe__count_common,
     },
     {
-     container_unpack,
-     native_find_rsc,
-     native_parameter,
-     container_print,
-     container_active,
-     container_resource_state,
-     native_location,
-     container_free
+         container_unpack,
+         native_find_rsc,
+         native_parameter,
+         container_print,
+         container_active,
+         container_resource_state,
+         native_location,
+         container_free,
+         pe__count_bundle,
     }
 };
 
@@ -293,7 +287,7 @@ unpack_template(xmlNode * xml_obj, xmlNode ** expanded_xml, pe_working_set_t * d
 
     template_ops = find_xml_node(new_xml, "operations", FALSE);
 
-    for (child_xml = __xml_first_child(xml_obj); child_xml != NULL;
+    for (child_xml = __xml_first_child_element(xml_obj); child_xml != NULL;
          child_xml = __xml_next_element(child_xml)) {
         xmlNode *new_child = NULL;
 
@@ -309,13 +303,17 @@ unpack_template(xmlNode * xml_obj, xmlNode ** expanded_xml, pe_working_set_t * d
         GHashTable *rsc_ops_hash =
             g_hash_table_new_full(crm_str_hash, g_str_equal, g_hash_destroy_str, NULL);
 
-        for (op = __xml_first_child(rsc_ops); op != NULL; op = __xml_next_element(op)) {
+        for (op = __xml_first_child_element(rsc_ops); op != NULL;
+             op = __xml_next_element(op)) {
+
             char *key = template_op_key(op);
 
             g_hash_table_insert(rsc_ops_hash, key, op);
         }
 
-        for (op = __xml_first_child(template_ops); op != NULL; op = __xml_next_element(op)) {
+        for (op = __xml_first_child_element(template_ops); op != NULL;
+             op = __xml_next_element(op)) {
+
             char *key = template_op_key(op);
 
             if (g_hash_table_lookup(rsc_ops_hash, key) == NULL) {
@@ -628,7 +626,8 @@ common_unpack(xmlNode * xml_obj, resource_t ** rsc,
             set_bit((*rsc)->flags, pe_rsc_maintenance);
         }
 
-    } else if (is_set(data_set->flags, pe_flag_maintenance_mode)) {
+    }
+    if (is_set(data_set->flags, pe_flag_maintenance_mode)) {
         clear_bit((*rsc)->flags, pe_rsc_managed);
         set_bit((*rsc)->flags, pe_rsc_maintenance);
     }
@@ -784,6 +783,14 @@ common_unpack(xmlNode * xml_obj, resource_t ** rsc,
         if(is_set((*rsc)->flags, pe_rsc_fence_device)) {
             value = "quorum";
 
+        } else if (((*rsc)->variant == pe_native)
+                   && safe_str_eq(crm_element_value((*rsc)->xml, XML_AGENT_ATTR_CLASS),
+                                  PCMK_RESOURCE_CLASS_OCF)
+                   && safe_str_eq(crm_element_value((*rsc)->xml, XML_AGENT_ATTR_PROVIDER), "pacemaker")
+                   && safe_str_eq(crm_element_value((*rsc)->xml, XML_ATTR_TYPE), "remote")
+            ) {
+            value = "quorum";
+
         } else if (is_set(data_set->flags, pe_flag_enable_unfencing)) {
             value = "unfencing";
 
@@ -973,3 +980,127 @@ common_free(resource_t * rsc)
     free(rsc->pending_task);
     free(rsc);
 }
+
+/*!
+ * \brief
+ * \internal Find a node (and optionally count all) where resource is active
+ *
+ * \param[in]  rsc          Resource to check
+ * \param[out] count_all    If not NULL, will be set to count of active nodes
+ * \param[out] count_clean  If not NULL, will be set to count of clean nodes
+ *
+ * \return An active node (or NULL if resource is not active anywhere)
+ *
+ * \note The order of preference is: an active node that is the resource's
+ *       partial migration source; if the resource's "requires" is "quorum" or
+ *       "nothing", the first active node in the list that is clean and online;
+ *       the first active node in the list.
+ */
+pe_node_t *
+pe__find_active_on(const resource_t *rsc, unsigned int *count_all,
+                   unsigned int *count_clean)
+{
+    pe_node_t *active = NULL;
+    pe_node_t *node = NULL;
+    bool keep_looking = FALSE;
+    bool is_happy = FALSE;
+
+    if (count_all) {
+        *count_all = 0;
+    }
+    if (count_clean) {
+        *count_clean = 0;
+    }
+    if (rsc == NULL) {
+        return NULL;
+    }
+
+    for (GList *node_iter = rsc->running_on; node_iter != NULL;
+         node_iter = node_iter->next) {
+
+        node = node_iter->data;
+        keep_looking = FALSE;
+
+        is_happy = node->details->online && !node->details->unclean;
+
+        if (count_all) {
+            ++*count_all;
+        }
+        if (count_clean && is_happy) {
+            ++*count_clean;
+        }
+        if (count_all || count_clean) {
+            // If we're counting, we need to go through entire list
+            keep_looking = TRUE;
+        }
+
+        if (rsc->partial_migration_source != NULL) {
+            if (node->details == rsc->partial_migration_source->details) {
+                // This is the migration source
+                active = node;
+            } else {
+                keep_looking = TRUE;
+            }
+        } else if (is_not_set(rsc->flags, pe_rsc_needs_fencing)) {
+            if (is_happy && (!active || !active->details->online
+                             || active->details->unclean)) {
+                // This is the first clean node
+                active = node;
+            } else {
+                keep_looking = TRUE;
+            }
+        }
+        if (active == NULL) {
+            // This is first node in list
+            active = node;
+        }
+
+        if (keep_looking == FALSE) {
+            // Don't waste time iterating if we don't have to
+            break;
+        }
+    }
+    return active;
+}
+
+/*!
+ * \brief
+ * \internal Find and count active nodes according to "requires"
+ *
+ * \param[in]  rsc    Resource to check
+ * \param[out] count  If not NULL, will be set to count of active nodes
+ *
+ * \return An active node (or NULL if resource is not active anywhere)
+ *
+ * \note This is a convenience wrapper for pe__find_active_on() where the count
+ *       of all active nodes or only clean active nodes is desired according to
+ *       the "requires" meta-attribute.
+ */
+pe_node_t *
+pe__find_active_requires(const resource_t *rsc, unsigned int *count)
+{
+    if (rsc && is_not_set(rsc->flags, pe_rsc_needs_fencing)) {
+        return pe__find_active_on(rsc, NULL, count);
+    }
+    return pe__find_active_on(rsc, count, NULL);
+}
+
+void
+pe__count_common(pe_resource_t *rsc)
+{
+    if (rsc->children != NULL) {
+        for (GList *item = rsc->children; item != NULL; item = item->next) {
+            ((pe_resource_t *) item->data)->fns->count(item->data);
+        }
+
+    } else if (is_not_set(rsc->flags, pe_rsc_orphan)
+               || (rsc->role > RSC_ROLE_STOPPED)) {
+        rsc->cluster->ninstances++;
+        if (pe__resource_is_disabled(rsc)) {
+            rsc->cluster->disabled_resources++;
+        }
+        if (is_set(rsc->flags, pe_rsc_block)) {
+            rsc->cluster->blocked_resources++;
+        }
+    }
+}
