@@ -1,19 +1,8 @@
 /*
- * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
+ * Copyright 2004-2018 Andrew Beekhof <andrew@beekhof.net>
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * This source code is licensed under the GNU Lesser General Public License
+ * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
  */
 
 #include <crm_internal.h>
@@ -24,10 +13,71 @@
 #include <crm/pengine/internal.h>
 #include <unpack.h>
 #include <crm/msg_xml.h>
+#include <pe_status_private.h>
 
 #define VARIANT_NATIVE 1
 #include "./variant.h"
 
+/*!
+ * \internal
+ * \brief Check whether a resource is active on multiple nodes
+ */
+static bool
+is_multiply_active(pe_resource_t *rsc)
+{
+    unsigned int count = 0;
+
+    if (rsc->variant == pe_native) {
+        pe__find_active_requires(rsc, &count);
+    }
+    return count > 1;
+}
+
+static void
+native_priority_to_node(pe_resource_t * rsc, pe_node_t * node)
+{
+    int priority = 0;
+
+    if (rsc->priority == 0) {
+        return;
+    }
+
+    if (rsc->role == RSC_ROLE_MASTER) {
+        // Promoted instance takes base priority + 1
+        priority = rsc->priority + 1;
+
+    } else {
+        priority = rsc->priority;
+    }
+
+    node->details->priority += priority;
+    pe_rsc_trace(rsc, "Node '%s' now has priority %d with %s'%s' (priority: %d%s)",
+                 node->details->uname, node->details->priority,
+                 rsc->role == RSC_ROLE_MASTER ? "promoted " : "",
+                 rsc->id, rsc->priority,
+                 rsc->role == RSC_ROLE_MASTER ? " + 1" : "");
+
+    /* Priority of a resource running on a guest node is added to the cluster
+     * node as well. */
+    if (node->details->remote_rsc
+        && node->details->remote_rsc->container) {
+        GListPtr gIter = node->details->remote_rsc->container->running_on;
+
+        for (; gIter != NULL; gIter = gIter->next) {
+            pe_node_t *a_node = gIter->data;
+
+            a_node->details->priority += priority;
+            pe_rsc_trace(rsc, "Node '%s' now has priority %d with %s'%s' (priority: %d%s) "
+                         "from guest node '%s'",
+                         a_node->details->uname, a_node->details->priority,
+                         rsc->role == RSC_ROLE_MASTER ? "promoted " : "",
+                         rsc->id, rsc->priority,
+                         rsc->role == RSC_ROLE_MASTER ? " + 1" : "",
+                         node->details->uname);
+        }
+    }
+}
+
 void
 native_add_running(resource_t * rsc, node_t * node, pe_working_set_t * data_set)
 {
@@ -49,6 +99,8 @@ native_add_running(resource_t * rsc, node_t * node, pe_working_set_t * data_set)
     rsc->running_on = g_list_append(rsc->running_on, node);
     if (rsc->variant == pe_native) {
         node->details->running_rsc = g_list_append(node->details->running_rsc, rsc);
+
+        native_priority_to_node(rsc, node);
     }
 
     if (rsc->variant == pe_native && node->details->maintenance) {
@@ -69,7 +121,7 @@ native_add_running(resource_t * rsc, node_t * node, pe_working_set_t * data_set)
         return;
     }
 
-    if (rsc->variant == pe_native && g_list_length(rsc->running_on) > 1) {
+    if (is_multiply_active(rsc)) {
         switch (rsc->recovery_type) {
             case recovery_stop_only:
                 {
@@ -77,7 +129,9 @@ native_add_running(resource_t * rsc, node_t * node, pe_working_set_t * data_set)
                     node_t *local_node = NULL;
 
                     /* make sure it doesn't come up again */
-                    g_hash_table_destroy(rsc->allowed_nodes);
+                    if (rsc->allowed_nodes != NULL) {
+                        g_hash_table_destroy(rsc->allowed_nodes);
+                    }
                     rsc->allowed_nodes = node_hash_from_list(data_set->nodes);
                     g_hash_table_iter_init(&gIter, rsc->allowed_nodes);
                     while (g_hash_table_iter_next(&gIter, NULL, (void **)&local_node)) {
@@ -108,8 +162,8 @@ native_add_running(resource_t * rsc, node_t * node, pe_working_set_t * data_set)
                 }
                 break;
         }
-        crm_debug("%s is active on %d nodes including %s: %s",
-                  rsc->id, g_list_length(rsc->running_on), node->details->uname,
+        crm_debug("%s is active on multiple nodes including %s: %s",
+                  rsc->id, node->details->uname,
                   recovery2text(rsc->recovery_type));
 
     } else {
@@ -121,40 +175,59 @@ native_add_running(resource_t * rsc, node_t * node, pe_working_set_t * data_set)
     }
 }
 
-extern void force_non_unique_clone(resource_t * rsc, const char *rid, pe_working_set_t * data_set);
+static void
+recursive_clear_unique(pe_resource_t *rsc)
+{
+    clear_bit(rsc->flags, pe_rsc_unique);
+    add_hash_param(rsc->meta, XML_RSC_ATTR_UNIQUE, XML_BOOLEAN_FALSE);
+
+    for (GList *child = rsc->children; child != NULL; child = child->next) {
+        recursive_clear_unique((pe_resource_t *) child->data);
+    }
+}
 
 gboolean
 native_unpack(resource_t * rsc, pe_working_set_t * data_set)
 {
     resource_t *parent = uber_parent(rsc);
     native_variant_data_t *native_data = NULL;
-    const char *class = crm_element_value(rsc->xml, XML_AGENT_ATTR_CLASS);
+    const char *standard = crm_element_value(rsc->xml, XML_AGENT_ATTR_CLASS);
+    uint32_t ra_caps = pcmk_get_ra_caps(standard);
 
     pe_rsc_trace(rsc, "Processing resource %s...", rsc->id);
 
     native_data = calloc(1, sizeof(native_variant_data_t));
     rsc->variant_opaque = native_data;
 
-    if (is_set(rsc->flags, pe_rsc_unique) && rsc->parent) {
+    // Only some agent standards support unique and promotable clones
+    if (is_not_set(ra_caps, pcmk_ra_cap_unique)
+        && is_set(rsc->flags, pe_rsc_unique) && pe_rsc_is_clone(parent)) {
 
-        if (safe_str_eq(class, PCMK_RESOURCE_CLASS_LSB)) {
-            resource_t *top = uber_parent(rsc);
+        /* @COMPAT We should probably reject this situation as an error (as we
+         * do for promotable below) rather than warn and convert, but that would
+         * be a backward-incompatible change that we should probably do with a
+         * transform at a schema major version bump.
+         */
+        pe__force_anon(standard, parent, rsc->id, data_set);
 
-            force_non_unique_clone(top, rsc->id, data_set);
-        }
+        /* Clear globally-unique on the parent and all its descendents unpacked
+         * so far (clearing the parent should make any future children unpacking
+         * correct). We have to clear this resource explicitly because it isn't
+         * hooked into the parent's children yet.
+         */
+        recursive_clear_unique(parent);
+        recursive_clear_unique(rsc);
     }
-
-    if (safe_str_eq(class, PCMK_RESOURCE_CLASS_OCF) == FALSE) {
+    if (is_not_set(ra_caps, pcmk_ra_cap_promotable)) {
         const char *stateful = g_hash_table_lookup(parent->meta, "stateful");
 
         if (safe_str_eq(stateful, XML_BOOLEAN_TRUE)) {
             pe_err
                 ("Resource %s is of type %s and therefore cannot be used as a master/slave resource",
-                 rsc->id, class);
+                 rsc->id, standard);
             return FALSE;
         }
     }
-
     return TRUE;
 }
 
@@ -244,7 +317,7 @@ native_parameter(resource_t * rsc, node_t * node, gboolean create, const char *n
 {
     char *value_copy = NULL;
     const char *value = NULL;
-    GHashTable *hash = rsc->parameters;
+    GHashTable *hash = NULL;
     GHashTable *local_hash = NULL;
 
     CRM_CHECK(rsc != NULL, return NULL);
@@ -264,6 +337,8 @@ native_parameter(resource_t * rsc, node_t * node, gboolean create, const char *n
         get_rsc_attributes(local_hash, rsc, node, data_set);
 
         hash = local_hash;
+    } else {
+        hash = rsc->parameters;
     }
 
     value = g_hash_table_lookup(hash, name);
@@ -449,7 +524,7 @@ native_print_xml(resource_t * rsc, const char *pre_text, long options, void *pri
     if (options & pe_print_rsconly) {
         status_print("/>\n");
         /* do nothing */
-    } else if (g_list_length(rsc->running_on) > 0) {
+    } else if (rsc->running_on != NULL) {
         GListPtr gIter = rsc->running_on;
 
         status_print(">\n");
@@ -521,7 +596,7 @@ common_print(resource_t * rsc, const char *pre_text, const char *name, node_t *n
         } else if (is_set(rsc->flags, pe_rsc_failed)) {
             status_print("<font color=\"red\">");
 
-        } else if (rsc->variant == pe_native && g_list_length(rsc->running_on) == 0) {
+        } else if (rsc->variant == pe_native && (rsc->running_on == NULL)) {
             status_print("<font color=\"red\">");
 
         } else if (g_list_length(rsc->running_on) > 1) {
@@ -540,7 +615,7 @@ common_print(resource_t * rsc, const char *pre_text, const char *name, node_t *n
     }
     offset += snprintf(buffer + offset, LINE_MAX - offset, "%s", name);
     offset += snprintf(buffer + offset, LINE_MAX - offset, "\t(%s", class);
-    if (crm_provider_required(class)) {
+    if (is_set(pcmk_get_ra_caps(class), pcmk_ra_cap_provider)) {
         const char *prov = crm_element_value(rsc->xml, XML_AGENT_ATTR_PROVIDER);
         offset += snprintf(buffer + offset, LINE_MAX - offset, "::%s", prov);
     }
@@ -585,20 +660,17 @@ common_print(resource_t * rsc, const char *pre_text, const char *name, node_t *n
         if (target_role_e == RSC_ROLE_STOPPED) {
             flagOffset += snprintf(flagBuffer + flagOffset, LINE_MAX - flagOffset,
                                    "%sdisabled", comma_if(flagOffset));
-            rsc->cluster->disabled_resources++;
 
         } else if (uber_parent(rsc)->variant == pe_master
                    && target_role_e == RSC_ROLE_SLAVE) {
             flagOffset += snprintf(flagBuffer + flagOffset, LINE_MAX - flagOffset,
                                    "%starget-role:%s", comma_if(flagOffset), target_role);
-            rsc->cluster->disabled_resources++;
         }
     }
 
     if (is_set(rsc->flags, pe_rsc_block)) {
         flagOffset += snprintf(flagBuffer + flagOffset, LINE_MAX - flagOffset,
                                "%sblocked", comma_if(flagOffset));
-        rsc->cluster->blocked_resources++;
 
     } else if (is_not_set(rsc->flags, pe_rsc_managed)) {
         flagOffset += snprintf(flagBuffer + flagOffset, LINE_MAX - flagOffset,
@@ -734,9 +806,13 @@ native_print(resource_t * rsc, const char *pre_text, long options, void *print_d
         return;
     }
 
-    if (rsc->running_on != NULL) {
-        node = rsc->running_on->data;
+    node = pe__current_node(rsc);
+
+    if (node == NULL) {
+        // This is set only if a non-probe action is pending on this node
+        node = rsc->pending_node;
     }
+
     common_print(rsc, pre_text, rsc_printable_id(rsc), node, options, print_data);
 }
 
@@ -759,8 +835,18 @@ native_resource_state(const resource_t * rsc, gboolean current)
     return role;
 }
 
-node_t *
-native_location(resource_t * rsc, GListPtr * list, gboolean current)
+/*!
+ * \internal
+ * \brief List nodes where a resource (or any of its children) is
+ *
+ * \param[in]  rsc      Resource to check
+ * \param[out] list     List to add result to
+ * \param[in]  current  0 = where known, 1 = running, 2 = running or pending
+ *
+ * \return If list contains only one node, that node
+ */
+pe_node_t *
+native_location(const pe_resource_t *rsc, GList **list, int current)
 {
     node_t *one = NULL;
     GListPtr result = NULL;
@@ -774,15 +860,22 @@ native_location(resource_t * rsc, GListPtr * list, gboolean current)
             child->fns->location(child, &result, current);
         }
 
-    } else if (current && rsc->running_on) {
-        result = g_list_copy(rsc->running_on);
+    } else if (current) {
+
+        if (rsc->running_on) {
+            result = g_list_copy(rsc->running_on);
+        }
+        if ((current == 2) && rsc->pending_node
+            && !pe_find_node_id(result, rsc->pending_node->details->id)) {
+                result = g_list_append(result, rsc->pending_node);
+        }
 
     } else if (current == FALSE && rsc->allocated_to) {
         result = g_list_append(NULL, rsc->allocated_to);
     }
 
-    if (result && g_list_length(result) == 1) {
-        one = g_list_nth_data(result, 0);
+    if (result && (result->next == NULL)) {
+        one = result->data;
     }
 
     if (list) {
@@ -823,7 +916,7 @@ get_rscs_brief(GListPtr rsc_list, GHashTable * rsc_table, GHashTable * active_ta
         }
 
         offset += snprintf(buffer + offset, LINE_MAX - offset, "%s", class);
-        if (crm_provider_required(class)) {
+        if (is_set(pcmk_get_ra_caps(class), pcmk_ra_cap_provider)) {
             const char *prov = crm_element_value(rsc->xml, XML_AGENT_ATTR_PROVIDER);
             offset += snprintf(buffer + offset, LINE_MAX - offset, "::%s", prov);
         }
