@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2016 Andrew Beekhof <andrew@beekhof.net>
+ * Copyright 2010-2019 Andrew Beekhof <andrew@beekhof.net>
  *
  * This source code is licensed under the GNU Lesser General Public License
  * version 2.1 or later (LGPLv2.1+) WITHOUT ANY WARRANTY.
@@ -55,6 +55,25 @@ services_action_create(const char *name, const char *action, int interval, int t
                                    action, interval, timeout, NULL, 0);
 }
 
+static char *
+services__lsb_agent_path(const char *agent)
+{
+    return (*agent == '/')? strdup(agent)
+        : crm_strdup_printf("%s/%s", LSB_ROOT_DIR, agent);
+}
+
+static bool
+services__lsb_agent_exists(const char *agent)
+{
+    bool rc = FALSE;
+    struct stat st;
+    char *path = services__lsb_agent_path(agent);
+
+    rc = (stat(path, &st) == 0);
+    free(path);
+    return rc;
+}
+
 /*!
  * \brief Find first service class that can provide a specified agent
  *
@@ -74,18 +93,9 @@ resources_find_service_class(const char *agent)
      * - systemd
      * - upstart
      */
-    int rc = 0;
-    struct stat st;
-    char *path = NULL;
-
-#ifdef LSB_ROOT_DIR
-    rc = asprintf(&path, "%s/%s", LSB_ROOT_DIR, agent);
-    if (rc > 0 && stat(path, &st) == 0) {
-        free(path);
+    if (services__lsb_agent_exists(agent)) {
         return PCMK_RESOURCE_CLASS_LSB;
     }
-    free(path);
-#endif
 
 #if SUPPORT_SYSTEMD
     if (systemd_unit_exists(agent)) {
@@ -167,6 +177,7 @@ resources_action_create(const char *name, const char *standard, const char *prov
                         GHashTable * params, enum svc_action_flags flags)
 {
     svc_action_t *op = NULL;
+    uint32_t ra_caps = 0;
 
     /*
      * Do some up front sanity checks before we go off and
@@ -182,9 +193,10 @@ resources_action_create(const char *name, const char *standard, const char *prov
         crm_err("Cannot create operation for %s without resource class", name);
         goto return_error;
     }
+    ra_caps = pcmk_get_ra_caps(standard);
 
-    if (crm_provider_required(standard) && crm_strlen_zero(provider)) {
-        crm_err("Cannot create OCF operation for %s without provider", name);
+    if (is_set(ra_caps, pcmk_ra_cap_provider) && crm_strlen_zero(provider)) {
+        crm_err("Cannot create operation for %s without provider", name);
         goto return_error;
     }
 
@@ -213,39 +225,34 @@ resources_action_create(const char *name, const char *standard, const char *prov
     op->flags = flags;
     op->id = generate_op_key(name, action, interval);
 
-    if (safe_str_eq(action, "monitor") && (
-#if SUPPORT_HEARTBEAT
-        safe_str_eq(op->standard, PCMK_RESOURCE_CLASS_HB) ||
-#endif
-        safe_str_eq(op->standard, PCMK_RESOURCE_CLASS_LSB))) {
-        action = "status";
+    if (is_set(ra_caps, pcmk_ra_cap_status) && safe_str_eq(action, "monitor")) {
+        op->action = strdup("status");
+    } else {
+        op->action = strdup(action);
     }
-    op->action = strdup(action);
 
-    if (crm_provider_required(op->standard)) {
+    if (is_set(ra_caps, pcmk_ra_cap_provider)) {
         op->provider = strdup(provider);
-        op->params = params;
-        params = NULL;
+    }
 
+    if (is_set(ra_caps, pcmk_ra_cap_params)) {
+        op->params = params;
+        params = NULL; // so we don't free them in this function
+    }
+
+    if (strcasecmp(op->standard, PCMK_RESOURCE_CLASS_OCF) == 0) {
         if (asprintf(&op->opaque->exec, "%s/resource.d/%s/%s", OCF_ROOT_DIR, provider, agent) == -1) {
             crm_err("Internal error: cannot create agent path");
             goto return_error;
         }
         op->opaque->args[0] = strdup(op->opaque->exec);
-        op->opaque->args[1] = strdup(action);
+        op->opaque->args[1] = strdup(op->action);
 
     } else if (strcasecmp(op->standard, PCMK_RESOURCE_CLASS_LSB) == 0) {
-        if (op->agent[0] == '/') {
-            /* if given an absolute path, use that instead
-             * of tacking on the LSB_ROOT_DIR path to the front */
-            op->opaque->exec = strdup(op->agent);
-        } else if (asprintf(&op->opaque->exec, "%s/%s", LSB_ROOT_DIR, op->agent) == -1) {
-            crm_err("Internal error: cannot create agent path");
-            goto return_error;
-        }
+        op->opaque->exec = services__lsb_agent_path(op->agent);
         op->opaque->args[0] = strdup(op->opaque->exec);
         op->opaque->args[1] = strdup(op->action);
-        op->opaque->args[2] = NULL;
+
 #if SUPPORT_HEARTBEAT
     } else if (strcasecmp(op->standard, PCMK_RESOURCE_CLASS_HB) == 0) {
         int index;
@@ -266,10 +273,10 @@ resources_action_create(const char *name, const char *standard, const char *prov
         /* The "heartbeat" agent class only has positional arguments,
          * which we keyed by their decimal position number. */
         param_num = 1;
-        if (params) {
+        if (op->params) {
             for (index = 1; index <= MAX_ARGC - 3; index++ ) {
                 snprintf(buf_tmp, sizeof(buf_tmp), "%d", index);
-                value_tmp = g_hash_table_lookup(params, buf_tmp);
+                value_tmp = g_hash_table_lookup(op->params, buf_tmp);
                 if (value_tmp == NULL) {
                     /* maybe: strdup("") ??
                      * But the old lrmd did simply continue as well. */
@@ -277,6 +284,10 @@ resources_action_create(const char *name, const char *standard, const char *prov
                 }
                 op->opaque->args[param_num++] = strdup(value_tmp);
             }
+
+            // Heartbeat actions don't need to keep the parameters
+            g_hash_table_destroy(op->params);
+            op->params = NULL;
         }
 
         /* Add operation code as the last argument, */
@@ -294,8 +305,6 @@ resources_action_create(const char *name, const char *standard, const char *prov
 #endif
 #if SUPPORT_NAGIOS
     } else if (strcasecmp(op->standard, PCMK_RESOURCE_CLASS_NAGIOS) == 0) {
-        int index = 0;
-
         if (op->agent[0] == '/') {
             /* if given an absolute path, use that instead
              * of tacking on the NAGIOS_PLUGIN_DIR path to the front */
@@ -307,20 +316,19 @@ resources_action_create(const char *name, const char *standard, const char *prov
         }
 
         op->opaque->args[0] = strdup(op->opaque->exec);
-        index = 1;
 
         if (safe_str_eq(op->action, "monitor") && op->interval == 0) {
             /* Invoke --version for a nagios probe */
-            op->opaque->args[index] = strdup("--version");
-            index++;
+            op->opaque->args[1] = strdup("--version");
 
-        } else if (params) {
+        } else if (op->params) {
             GHashTableIter iter;
             char *key = NULL;
             char *value = NULL;
+            int index = 1;
             static int args_size = sizeof(op->opaque->args) / sizeof(char *);
 
-            g_hash_table_iter_init(&iter, params);
+            g_hash_table_iter_init(&iter, op->params);
 
             while (g_hash_table_iter_next(&iter, (gpointer *) & key, (gpointer *) & value) &&
                    index <= args_size - 3) {
@@ -341,12 +349,16 @@ resources_action_create(const char *name, const char *standard, const char *prov
                 index += 2;
             }
         }
-        op->opaque->args[index] = NULL;
+
+        // Nagios actions don't need to keep the parameters
+        if (op->params != NULL) {
+            g_hash_table_destroy(op->params);
+            op->params = NULL;
+        }
 #endif
     } else {
         crm_err("Unknown resource standard: %s", op->standard);
-        services_action_free(op);
-        op = NULL;
+        goto return_error;
     }
 
     if(params) {
@@ -438,35 +450,6 @@ services_action_user(svc_action_t *op, const char *user)
     return crm_user_lookup(user, &(op->opaque->uid), &(op->opaque->gid));
 }
 
-static void
-set_alert_env(gpointer key, gpointer value, gpointer user_data)
-{
-    int rc;
-
-    if (value) {
-        rc = setenv(key, value, 1);
-    } else {
-        rc = unsetenv(key);
-    }
-
-    if (rc < 0) {
-        crm_perror(LOG_ERR, "setenv %s=%s",
-                  (char*)key, (value? (char*)value : ""));
-    } else {
-        crm_trace("setenv %s=%s", (char*)key, (value? (char*)value : ""));
-    }
-}
-
-static void
-unset_alert_env(gpointer key, gpointer value, gpointer user_data)
-{
-    if (unsetenv(key) < 0) {
-        crm_perror(LOG_ERR, "unset %s", (char*)key);
-    } else {
-        crm_trace("unset %s", (char*)key);
-    }
-}
-
 /*!
  * \brief Execute an alert agent action
  *
@@ -481,18 +464,9 @@ unset_alert_env(gpointer key, gpointer value, gpointer user_data)
 gboolean
 services_alert_async(svc_action_t *action, void (*cb)(svc_action_t *op))
 {
-    gboolean responsible;
-
     action->synchronous = false;
     action->opaque->callback = cb;
-    if (action->params) {
-        g_hash_table_foreach(action->params, set_alert_env, NULL);
-    }
-    responsible = services_os_action_execute(action);
-    if (action->params) {
-        g_hash_table_foreach(action->params, unset_alert_env, NULL);
-    }
-    return responsible;
+    return services_os_action_execute(action);
 }
 
 #if SUPPORT_DBUS
@@ -526,7 +500,7 @@ services_set_op_pending(svc_action_t *op, DBusPendingCall *pending)
 void
 services_action_cleanup(svc_action_t * op)
 {
-    if(op->opaque == NULL) {
+    if ((op == NULL) || (op->opaque == NULL)) {
         return;
     }
 
@@ -538,13 +512,16 @@ services_action_cleanup(svc_action_t * op)
     }
 
     if(op->opaque->pending) {
-        crm_trace("Cleaning up pending dbus call %p %s for %s", op->opaque->pending, op->action, op->rsc);
-        if(dbus_pending_call_get_completed(op->opaque->pending)) {
-            crm_warn("Pending dbus call %s for %s did not complete", op->action, op->rsc);
+        if (dbus_pending_call_get_completed(op->opaque->pending)) {
+            // This should never be the case
+            crm_warn("Result of %s op %s was unhandled",
+                     op->standard, op->id);
+        } else {
+            crm_debug("Will ignore any result of canceled %s op %s",
+                      op->standard, op->id);
         }
         dbus_pending_call_cancel(op->opaque->pending);
-        dbus_pending_call_unref(op->opaque->pending);
-        op->opaque->pending = NULL;
+        services_set_op_pending(op, NULL);
     }
 #endif
 
@@ -654,7 +631,7 @@ services_action_cancel(const char *name, const char *action, int interval)
     /* Tell operation_finalize() not to reschedule the operation */
     op->cancel = TRUE;
 
-    /* Stop tracking it as a recurring operation, and stop its timer */
+    /* Stop tracking it as a recurring operation, and stop its repeat timer */
     cancel_recurring_action(op);
 
     /* If the op has a PID, it's an in-flight child process, so kill it.
@@ -673,19 +650,22 @@ services_action_cancel(const char *name, const char *action, int interval)
         goto done;
     }
 
-    /* In-flight systemd and upstart ops don't have a pid. The relevant handlers
-     * will call operation_finalize() when the operation completes.
-     * @TODO: Can we request early termination, maybe using
-     * dbus_pending_call_cancel()?
-     */
+#if SUPPORT_DBUS
+    // In-flight systemd and upstart ops don't have a pid
     if (inflight_systemd_or_upstart(op)) {
-        crm_info("Will cancel %s op %s when in-flight instance completes",
-                 op->standard, op->id);
-        cancelled = FALSE;
-        goto done;
-    }
+        inflight_ops = g_list_remove(inflight_ops, op);
 
-    /* Otherwise, operation is not in-flight, just report as cancelled */
+        /* This will cause any result that comes in later to be discarded, so we
+         * don't call the callback and free the operation twice.
+         */
+        services_action_cleanup(op);
+    }
+#endif
+
+    // The rest of this is essentially equivalent to operation_finalize(),
+    // except without calling handle_blocked_ops()
+
+    // Report operation as cancelled
     op->status = PCMK_LRM_OP_CANCELLED;
     if (op->opaque->callback) {
         op->opaque->callback(op);
@@ -694,6 +674,7 @@ services_action_cancel(const char *name, const char *action, int interval)
     blocked_ops = g_list_remove(blocked_ops, op);
     services_action_free(op);
     cancelled = TRUE;
+    // @TODO Initiate handle_blocked_ops() asynchronously
 
 done:
     free(id);
@@ -824,12 +805,17 @@ services_untrack_op(svc_action_t *op)
 }
 
 gboolean
-services_action_async(svc_action_t * op, void (*action_callback) (svc_action_t *))
+services_action_async_fork_notify(svc_action_t * op,
+                                  void (*action_callback) (svc_action_t *),
+                                  void (*action_fork_callback) (svc_action_t *))
 {
     op->synchronous = false;
     if (action_callback) {
         op->opaque->callback = action_callback;
     }
+    if (action_fork_callback) {
+        op->opaque->fork_callback = action_fork_callback;
+    }
 
     if (op->interval > 0) {
         init_recurring_actions();
@@ -841,7 +827,8 @@ services_action_async(svc_action_t * op, void (*action_callback) (svc_action_t *
         g_hash_table_replace(recurring_actions, op->id, op);
     }
 
-    if (op->rsc && is_op_blocked(op->rsc)) {
+    if (is_not_set(op->flags, SVC_ACTION_NON_BLOCKED)
+        && op->rsc && is_op_blocked(op->rsc)) {
         blocked_ops = g_list_append(blocked_ops, op);
         return TRUE;
     }
@@ -849,6 +836,12 @@ services_action_async(svc_action_t * op, void (*action_callback) (svc_action_t *
     return action_exec_helper(op);
 }
 
+gboolean
+services_action_async(svc_action_t * op,
+                      void (*action_callback) (svc_action_t *))
+{
+    return services_action_async_fork_notify(op, action_callback, NULL);
+}
 
 static gboolean processing_blocked_ops = FALSE;
 
@@ -1414,7 +1407,7 @@ resources_list_standards(void)
 GList *
 resources_list_providers(const char *standard)
 {
-    if (crm_provider_required(standard)) {
+    if (is_set(pcmk_get_ra_caps(standard), pcmk_ra_cap_provider)) {
         return resources_os_list_ocf_providers();
     }
 
