@@ -1,5 +1,7 @@
 /*
- * Copyright (c) 2012 David Vossel <davidvossel@gmail.com>
+ * Copyright 2012-2020 the Pacemaker project contributors
+ *
+ * The version control history for this file may have further details.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -135,9 +137,7 @@ static const char *
 normalize_action_name(lrmd_rsc_t * rsc, const char *action)
 {
     if (safe_str_eq(action, "monitor") &&
-        (safe_str_eq(rsc->class, PCMK_RESOURCE_CLASS_LSB) ||
-         safe_str_eq(rsc->class, PCMK_RESOURCE_CLASS_SERVICE)
-         || safe_str_eq(rsc->class, PCMK_RESOURCE_CLASS_SYSTEMD))) {
+        is_set(pcmk_get_ra_caps(rsc->class), pcmk_ra_cap_status)) {
         return "status";
     }
     return action;
@@ -158,6 +158,7 @@ build_rsc_from_xml(xmlNode * msg)
     rsc->provider = crm_element_value_copy(rsc_xml, F_LRMD_PROVIDER);
     rsc->type = crm_element_value_copy(rsc_xml, F_LRMD_TYPE);
     rsc->work = mainloop_add_trigger(G_PRIORITY_HIGH, lrmd_rsc_dispatch, rsc);
+    rsc->st_probe_rc = -ENODEV; // if stonith, initialize to "not running"
     return rsc;
 }
 
@@ -354,7 +355,7 @@ schedule_lrmd_cmd(lrmd_rsc_t * rsc, lrmd_cmd_t * cmd)
     }
 
     /* crmd expects lrmd to automatically cancel recurring ops before rsc stops. */
-    if (rsc && safe_str_eq(cmd->action, "stop")) {
+    if (safe_str_eq(cmd->action, "stop")) {
         cancel_all_recurring(rsc, NULL);
     }
 
@@ -372,23 +373,15 @@ schedule_lrmd_cmd(lrmd_rsc_t * rsc, lrmd_cmd_t * cmd)
     }
 }
 
-static void
-send_reply(crm_client_t * client, int rc, uint32_t id, int call_id)
+static xmlNode *
+create_lrmd_reply(const char *origin, int rc, int call_id)
 {
-    int send_rc = 0;
-    xmlNode *reply = NULL;
+    xmlNode *reply = create_xml_node(NULL, T_LRMD_REPLY);
 
-    reply = create_xml_node(NULL, T_LRMD_REPLY);
-    crm_xml_add(reply, F_LRMD_ORIGIN, __FUNCTION__);
+    crm_xml_add(reply, F_LRMD_ORIGIN, origin);
     crm_xml_add_int(reply, F_LRMD_RC, rc);
     crm_xml_add_int(reply, F_LRMD_CALLID, call_id);
-
-    send_rc = lrmd_server_send_reply(client, id, reply);
-
-    free_xml(reply);
-    if (send_rc < 0) {
-        crm_warn("LRMD reply to %s failed: %d", client->name, send_rc);
-    }
+    return reply;
 }
 
 static void
@@ -582,8 +575,10 @@ cmd_reset(lrmd_cmd_t * cmd)
 {
     cmd->lrmd_op_status = 0;
     cmd->last_pid = 0;
+#ifdef HAVE_SYS_TIMEB_H
     memset(&cmd->t_run, 0, sizeof(cmd->t_run));
     memset(&cmd->t_queue, 0, sizeof(cmd->t_queue));
+#endif
     free(cmd->exit_reason);
     cmd->exit_reason = NULL;
     free(cmd->output);
@@ -877,7 +872,7 @@ action_complete(svc_action_t * action)
     rsc = cmd->rsc_id ? g_hash_table_lookup(rsc_list, cmd->rsc_id) : NULL;
 
     if (rsc && safe_str_eq(rsc->class, PCMK_RESOURCE_CLASS_SERVICE)) {
-        rclass = resources_find_service_class(rsc->class);
+        rclass = resources_find_service_class(rsc->type);
     } else if(rsc) {
         rclass = rsc->class;
     }
@@ -919,8 +914,15 @@ action_complete(svc_action_t * action)
                 cmd_original_times(cmd);
 #endif
 
-                if(cmd->lrmd_op_status == PCMK_LRM_OP_DONE && cmd->exec_rc == PCMK_OCF_NOT_RUNNING && safe_str_eq(cmd->real_action, "stop")) {
-                    cmd->exec_rc = PCMK_OCF_OK;
+                // Monitors may return "not running", but start/stop shouldn't
+                if ((cmd->lrmd_op_status == PCMK_LRM_OP_DONE)
+                    && (cmd->exec_rc == PCMK_OCF_NOT_RUNNING)) {
+
+                    if (safe_str_eq(cmd->real_action, "start")) {
+                        cmd->exec_rc = PCMK_OCF_UNKNOWN_ERROR;
+                    } else if (safe_str_eq(cmd->real_action, "stop")) {
+                        cmd->exec_rc = PCMK_OCF_OK;
+                    }
                 }
             }
         }
@@ -944,8 +946,8 @@ action_complete(svc_action_t * action)
      * obsolete, we should eventually prefer a clock_gettime() implementation
      * (wrapped in its own ifdef) with timeb as a fallback.
      */
-#ifdef HAVE_SYS_TIMEB_H
     if(goagain) {
+#ifdef HAVE_SYS_TIMEB_H
         int time_sum = time_diff_ms(NULL, &cmd->t_first_run);
         int timeout_left = cmd->timeout_orig - time_sum;
         int delay = cmd->timeout_orig / 10;
@@ -988,8 +990,8 @@ action_complete(svc_action_t * action)
             cmd->exec_rc = PCMK_OCF_TIMEOUT;
             cmd_original_times(cmd);
         }
-    }
 #endif
+    }
 
     if (action->stderr_data) {
         cmd->output = strdup(action->stderr_data);
@@ -1017,8 +1019,9 @@ stonith_action_complete(lrmd_cmd_t * cmd, int rc)
         /* do nothing */
 
     } else if (rc == -ENODEV && safe_str_eq(cmd->action, "monitor")) {
-        /* Not registered == inactive */
-        cmd->lrmd_op_status = PCMK_LRM_OP_DONE;
+        // The device is not registered with the fencer
+
+        cmd->lrmd_op_status = PCMK_LRM_OP_ERROR;
         cmd->exec_rc = PCMK_OCF_NOT_RUNNING;
 
     } else if (rc) {
@@ -1037,8 +1040,12 @@ stonith_action_complete(lrmd_cmd_t * cmd, int rc)
     } else {
         /* command successful */
         cmd->lrmd_op_status = PCMK_LRM_OP_DONE;
-        if (safe_str_eq(cmd->action, "start") && rsc) {
-            rsc->stonith_started = 1;
+        if (rsc) {
+            if (safe_str_eq(cmd->action, "start")) {
+                rsc->st_probe_rc = pcmk_ok; // maps to PCMK_OCF_OK
+            } else if (safe_str_eq(cmd->action, "stop")) {
+                rsc->st_probe_rc = -ENODEV; // maps to PCMK_OCF_NOT_RUNNING
+            }
         }
     }
 
@@ -1070,6 +1077,18 @@ stonith_connection_failed(void)
     g_hash_table_iter_init(&iter, rsc_list);
     while (g_hash_table_iter_next(&iter, (gpointer *) & key, (gpointer *) & rsc)) {
         if (safe_str_eq(rsc->class, PCMK_RESOURCE_CLASS_STONITH)) {
+            /* If we registered this fence device, we don't know whether the
+             * fencer still has the registration or not. Cause future probes to
+             * return PCMK_OCF_UNKNOWN_ERROR until the resource is stopped or
+             * started successfully. This is especially important if the
+             * controller also went away (possibly due to a cluster layer
+             * restart) and won't receive our client notification of any
+             * monitors finalized below.
+             */
+            if (rsc->st_probe_rc == pcmk_ok) {
+                rsc->st_probe_rc = pcmk_err_generic;
+            }
+
             if (rsc->active) {
                 cmd_list = g_list_append(cmd_list, rsc->active);
             }
@@ -1140,12 +1159,11 @@ lrmd_rsc_execute_stonith(lrmd_rsc_t * rsc, lrmd_cmd_t * cmd)
         }
     } else if (safe_str_eq(cmd->action, "stop")) {
         rc = stonith_api->cmds->remove_device(stonith_api, st_opt_sync_call, cmd->rsc_id);
-        rsc->stonith_started = 0;
     } else if (safe_str_eq(cmd->action, "monitor")) {
         if (cmd->interval) {
             do_monitor = 1;
         } else {
-            rc = rsc->stonith_started ? 0 : -ENODEV;
+            rc = rsc->st_probe_rc;
         }
     }
 
@@ -1371,32 +1389,38 @@ free_rsc(gpointer data)
 }
 
 static int
-process_lrmd_signon(crm_client_t * client, uint32_t id, xmlNode * request)
+process_lrmd_signon(crm_client_t *client, xmlNode *request, int call_id,
+                    xmlNode **reply)
 {
-    xmlNode *reply = create_xml_node(NULL, "reply");
+    int rc = pcmk_ok;
     const char *is_ipc_provider = crm_element_value(request, F_LRMD_IS_IPC_PROVIDER);
     const char *protocol_version = crm_element_value(request, F_LRMD_PROTOCOL_VERSION);
 
     if (compare_version(protocol_version, LRMD_MIN_PROTOCOL_VERSION) < 0) {
         crm_err("Cluster API version must be greater than or equal to %s, not %s",
                 LRMD_MIN_PROTOCOL_VERSION, protocol_version);
-        crm_xml_add_int(reply, F_LRMD_RC, -EPROTO);
+        rc = -EPROTO;
     }
 
-    crm_xml_add(reply, F_LRMD_OPERATION, CRM_OP_REGISTER);
-    crm_xml_add(reply, F_LRMD_CLIENTID, client->id);
-    crm_xml_add(reply, F_LRMD_PROTOCOL_VERSION, LRMD_PROTOCOL_VERSION);
-    lrmd_server_send_reply(client, id, reply);
-
     if (crm_is_true(is_ipc_provider)) {
-        /* this is a remote connection from a cluster nodes crmd */
 #ifdef SUPPORT_REMOTE
-        ipc_proxy_add_provider(client);
+        if ((client->remote != NULL) && client->remote->tls_handshake_complete) {
+            // This is a remote connection from a cluster node's controller
+            ipc_proxy_add_provider(client);
+        } else {
+            rc = -EACCES;
+        }
+#else
+        rc = -EPROTONOSUPPORT;
 #endif
     }
 
-    free_xml(reply);
-    return pcmk_ok;
+    *reply = create_lrmd_reply(__func__, rc, call_id);
+    crm_xml_add(*reply, F_LRMD_OPERATION, CRM_OP_REGISTER);
+    crm_xml_add(*reply, F_LRMD_CLIENTID, client->id);
+    crm_xml_add(*reply, F_LRMD_PROTOCOL_VERSION, LRMD_PROTOCOL_VERSION);
+
+    return rc;
 }
 
 static int
@@ -1424,52 +1448,34 @@ process_lrmd_rsc_register(crm_client_t * client, uint32_t id, xmlNode * request)
     return rc;
 }
 
-static void
-process_lrmd_get_rsc_info(crm_client_t * client, uint32_t id, xmlNode * request)
+static xmlNode *
+process_lrmd_get_rsc_info(xmlNode *request, int call_id)
 {
     int rc = pcmk_ok;
-    int send_rc = 0;
-    int call_id = 0;
     xmlNode *rsc_xml = get_xpath_object("//" F_LRMD_RSC, request, LOG_ERR);
     const char *rsc_id = crm_element_value(rsc_xml, F_LRMD_RSC_ID);
     xmlNode *reply = NULL;
     lrmd_rsc_t *rsc = NULL;
 
-    crm_element_value_int(request, F_LRMD_CALLID, &call_id);
-
-    if (!rsc_id) {
+    if (rsc_id == NULL) {
         rc = -ENODEV;
-        goto get_rsc_done;
+    } else {
+        rsc = g_hash_table_lookup(rsc_list, rsc_id);
+        if (rsc == NULL) {
+            crm_info("Resource '%s' not found (%d active resources)",
+                     rsc_id, g_hash_table_size(rsc_list));
+            rc = -ENODEV;
+        }
     }
 
-    if (!(rsc = g_hash_table_lookup(rsc_list, rsc_id))) {
-        crm_info("Resource '%s' not found (%d active resources)",
-                 rsc_id, g_hash_table_size(rsc_list));
-        rc = -ENODEV;
-        goto get_rsc_done;
-    }
-
-  get_rsc_done:
-
-    reply = create_xml_node(NULL, T_LRMD_REPLY);
-    crm_xml_add(reply, F_LRMD_ORIGIN, __FUNCTION__);
-    crm_xml_add_int(reply, F_LRMD_RC, rc);
-    crm_xml_add_int(reply, F_LRMD_CALLID, call_id);
-
+    reply = create_lrmd_reply(__FUNCTION__, rc, call_id);
     if (rsc) {
         crm_xml_add(reply, F_LRMD_RSC_ID, rsc->rsc_id);
         crm_xml_add(reply, F_LRMD_CLASS, rsc->class);
         crm_xml_add(reply, F_LRMD_PROVIDER, rsc->provider);
         crm_xml_add(reply, F_LRMD_TYPE, rsc->type);
     }
-
-    send_rc = lrmd_server_send_reply(client, id, reply);
-
-    if (send_rc < 0) {
-        crm_warn("LRMD reply to %s failed: %d", client->name, send_rc);
-    }
-
-    free_xml(reply);
+    return reply;
 }
 
 static int
@@ -1649,46 +1655,93 @@ process_lrmd_message(crm_client_t * client, uint32_t id, xmlNode * request)
     const char *op = crm_element_value(request, F_LRMD_OPERATION);
     int do_reply = 0;
     int do_notify = 0;
+    xmlNode *reply = NULL;
+
+#if ENABLE_ACL
+    /* Certain IPC commands may be done only by privileged users (i.e. root or
+     * hacluster) when ACLs are enabled, because they would otherwise provide a
+     * means of bypassing ACLs.
+     */
+    bool allowed = is_set(client->flags, crm_client_flag_ipc_privileged);
+#else
+    bool allowed = true;
+#endif
 
     crm_trace("Processing %s operation from %s", op, client->id);
     crm_element_value_int(request, F_LRMD_CALLID, &call_id);
 
     if (crm_str_eq(op, CRM_OP_IPC_FWD, TRUE)) {
 #ifdef SUPPORT_REMOTE
-        ipc_proxy_forward_client(client, request);
+        if (allowed) {
+            ipc_proxy_forward_client(client, request);
+        } else {
+            rc = -EACCES;
+        }
+#else
+        rc = -EPROTONOSUPPORT;
 #endif
         do_reply = 1;
     } else if (crm_str_eq(op, CRM_OP_REGISTER, TRUE)) {
-        rc = process_lrmd_signon(client, id, request);
+        rc = process_lrmd_signon(client, request, call_id, &reply);
+        do_reply = 1;
     } else if (crm_str_eq(op, LRMD_OP_RSC_REG, TRUE)) {
-        rc = process_lrmd_rsc_register(client, id, request);
-        do_notify = 1;
+        if (allowed) {
+            rc = process_lrmd_rsc_register(client, id, request);
+            do_notify = 1;
+        } else {
+            rc = -EACCES;
+        }
         do_reply = 1;
     } else if (crm_str_eq(op, LRMD_OP_RSC_INFO, TRUE)) {
-        process_lrmd_get_rsc_info(client, id, request);
+        if (allowed) {
+            reply = process_lrmd_get_rsc_info(request, call_id);
+        } else {
+            rc = -EACCES;
+        }
+        do_reply = 1;
     } else if (crm_str_eq(op, LRMD_OP_RSC_UNREG, TRUE)) {
-        rc = process_lrmd_rsc_unregister(client, id, request);
-        /* don't notify anyone about failed un-registers */
-        if (rc == pcmk_ok || rc == -EINPROGRESS) {
-            do_notify = 1;
+        if (allowed) {
+            rc = process_lrmd_rsc_unregister(client, id, request);
+            /* don't notify anyone about failed un-registers */
+            if (rc == pcmk_ok || rc == -EINPROGRESS) {
+                do_notify = 1;
+            }
+        } else {
+            rc = -EACCES;
         }
         do_reply = 1;
     } else if (crm_str_eq(op, LRMD_OP_RSC_EXEC, TRUE)) {
-        rc = process_lrmd_rsc_exec(client, id, request);
+        if (allowed) {
+            rc = process_lrmd_rsc_exec(client, id, request);
+        } else {
+            rc = -EACCES;
+        }
         do_reply = 1;
     } else if (crm_str_eq(op, LRMD_OP_RSC_CANCEL, TRUE)) {
-        rc = process_lrmd_rsc_cancel(client, id, request);
+        if (allowed) {
+            rc = process_lrmd_rsc_cancel(client, id, request);
+        } else {
+            rc = -EACCES;
+        }
         do_reply = 1;
     } else if (crm_str_eq(op, LRMD_OP_POKE, TRUE)) {
         do_notify = 1;
         do_reply = 1;
     } else if (crm_str_eq(op, LRMD_OP_CHECK, TRUE)) {
-        xmlNode *data = get_message_xml(request, F_LRMD_CALLDATA); 
-        const char *timeout = crm_element_value(data, F_LRMD_WATCHDOG);
-        CRM_LOG_ASSERT(data != NULL);
-        check_sbd_timeout(timeout);
+        if (allowed) {
+            xmlNode *data = get_message_xml(request, F_LRMD_CALLDATA);
+
+            CRM_LOG_ASSERT(data != NULL);
+            check_sbd_timeout(crm_element_value(data, F_LRMD_WATCHDOG));
+        } else {
+            rc = -EACCES;
+        }
     } else if (crm_str_eq(op, LRMD_OP_ALERT_EXEC, TRUE)) {
-        rc = process_lrmd_alert_exec(client, id, request);
+        if (allowed) {
+            rc = process_lrmd_alert_exec(client, id, request);
+        } else {
+            rc = -EACCES;
+        }
         do_reply = 1;
     } else {
         rc = -EOPNOTSUPP;
@@ -1697,11 +1750,26 @@ process_lrmd_message(crm_client_t * client, uint32_t id, xmlNode * request)
         crm_log_xml_warn(request, "UnknownOp");
     }
 
+    if (rc == -EACCES) {
+        crm_warn("Rejecting IPC request '%s' from unprivileged client %s",
+                 op, crm_client_name(client));
+    }
+
     crm_debug("Processed %s operation from %s: rc=%d, reply=%d, notify=%d",
               op, client->id, rc, do_reply, do_notify);
 
     if (do_reply) {
-        send_reply(client, rc, id, call_id);
+        int send_rc = pcmk_ok;
+
+        if (reply == NULL) {
+            reply = create_lrmd_reply(__FUNCTION__, rc, call_id);
+        }
+        send_rc = lrmd_server_send_reply(client, id, reply);
+        free_xml(reply);
+        if (send_rc < 0) {
+            crm_warn("Reply to client %s failed: %s " CRM_XS " %d",
+                     client->name, pcmk_strerror(send_rc), send_rc);
+        }
     }
 
     if (do_notify) {
