@@ -54,20 +54,15 @@ typedef struct lrmd_rsc_s {
      * that have been handed off from the pending ops list. */
     GList *recurring_ops;
 
-    int stonith_started;
+    int st_probe_rc; // What value should be returned for a probe if stonith
 
     crm_trigger_t *work;
 } lrmd_rsc_t;
 
 #  ifdef HAVE_GNUTLS_GNUTLS_H
-/* in remote_tls.c */
+// in remoted_tls.c
 int lrmd_init_remote_tls_server(void);
 void lrmd_tls_server_destroy(void);
-
-/* Hidden in lrmd client lib */
-extern int lrmd_tls_send_msg(crm_remote_t * session, xmlNode * msg, uint32_t id,
-                             const char *msg_type);
-extern int lrmd_tls_set_key(gnutls_datum_t * key);
 #  endif
 
 int lrmd_server_send_reply(crm_client_t * client, uint32_t id, xmlNode * reply);
@@ -109,8 +104,9 @@ void ipc_proxy_remove_provider(crm_client_t *client);
 void ipc_proxy_forward_client(crm_client_t *client, xmlNode *xml);
 crm_client_t *ipc_proxy_get_provider(void);
 int ipc_proxy_shutdown_req(crm_client_t *ipc_proxy);
+void remoted_spawn_pidone(int argc, char **argv, char **envp);
 #endif
 
 int process_lrmd_alert_exec(crm_client_t *client, uint32_t id, xmlNode *request);
-void lrmd_drain_alerts(GMainContext *ctx);
+void lrmd_drain_alerts(GMainLoop *mloop);
 #endif
