@@ -27,6 +27,7 @@
 #include <crm/crm.h>
 #include <crm/msg_xml.h>
 #include <crm/common/mainloop.h>
+#include <crm/common/remote_internal.h>
 
 #include <lrmd_private.h>
 
@@ -37,6 +38,12 @@
 #include <arpa/inet.h>
 
 #ifdef HAVE_GNUTLS_GNUTLS_H
+
+#  include <gnutls/gnutls.h>
+
+// Hidden in liblrmd
+extern int lrmd_tls_set_key(gnutls_datum_t *key);
+
 #  define LRMD_REMOTE_AUTH_TIMEOUT 10000
 gnutls_psk_server_credentials_t psk_cred_s;
 gnutls_dh_params_t dh_params;
@@ -49,6 +56,44 @@ debug_log(int level, const char *str)
     fputs(str, stderr);
 }
 
+/*!
+ * \internal
+ * \brief Read (more) TLS handshake data from client
+ */
+static int
+remoted__read_handshake_data(crm_client_t *client)
+{
+    int rc = pcmk__read_handshake_data(client);
+
+    if (rc == 0) {
+        /* No more data is available at the moment. Just return for now;
+         * we'll get invoked again once the client sends more.
+         */
+        return 0;
+    } else if (rc < 0) {
+        crm_err("TLS handshake with remote client failed: %s "
+                CRM_XS " rc=%d", gnutls_strerror(rc), rc);
+        return -1;
+    }
+
+    if (client->remote->auth_timeout) {
+        g_source_remove(client->remote->auth_timeout);
+    }
+    client->remote->auth_timeout = 0;
+
+    client->remote->tls_handshake_complete = TRUE;
+    crm_notice("Remote client connection accepted");
+
+    /* Only a client with access to the TLS key can connect, so we can treat
+     * it as privileged.
+     */
+    set_bit(client->flags, crm_client_flag_ipc_privileged);
+
+    // Alert other clients of the new connection
+    notify_of_new_client(client);
+    return 0;
+}
+
 static int
 lrmd_remote_client_msg(gpointer data)
 {
@@ -59,31 +104,7 @@ lrmd_remote_client_msg(gpointer data)
     crm_client_t *client = data;
 
     if (client->remote->tls_handshake_complete == FALSE) {
-        int rc = 0;
-
-        /* Muliple calls to handshake will be required, this callback
-         * will be invoked once the client sends more handshake data. */
-        do {
-            rc = gnutls_handshake(*client->remote->tls_session);
-
-            if (rc < 0 && rc != GNUTLS_E_AGAIN) {
-                crm_err("Remote lrmd tls handshake failed");
-                return -1;
-            }
-        } while (rc == GNUTLS_E_INTERRUPTED);
-
-        if (rc == 0) {
-            crm_debug("Remote lrmd tls handshake completed");
-            client->remote->tls_handshake_complete = TRUE;
-            if (client->remote->auth_timeout) {
-                g_source_remove(client->remote->auth_timeout);
-            }
-            client->remote->auth_timeout = 0;
-
-            /* Alert other clients of the new connection */
-            notify_of_new_client(client);
-        }
-        return 0;
+        return remoted__read_handshake_data(client);
     }
 
     rc = crm_remote_ready(client->remote, 0);
@@ -91,7 +112,7 @@ lrmd_remote_client_msg(gpointer data)
         /* no msg to read */
         return 0;
     } else if (rc < 0) {
-        crm_info("Client disconnected during remote client read");
+        crm_info("Remote client disconnected while polling it");
         return -1;
     }
 
@@ -100,7 +121,7 @@ lrmd_remote_client_msg(gpointer data)
     request = crm_remote_parse_buffer(client->remote);
     while (request) {
         crm_element_value_int(request, F_LRMD_REMOTE_MSG_ID, &id);
-        crm_trace("processing request from remote client with remote msg id %d", id);
+        crm_trace("Processing remote client request %d", id);
         if (!client->name) {
             const char *value = crm_element_value(request, F_LRMD_CLIENTNAME);
 
@@ -126,7 +147,7 @@ lrmd_remote_client_msg(gpointer data)
     }
 
     if (disconnected) {
-        crm_info("Client disconnect detected in tls msg dispatcher.");
+        crm_info("Remote client disconnected while reading from it");
         return -1;
     }
 
@@ -188,6 +209,7 @@ lrmd_auth_timeout_cb(gpointer data)
     return FALSE;
 }
 
+// Dispatch callback for remote server socket
 static int
 lrmd_remote_listen(gpointer data)
 {
@@ -195,19 +217,22 @@ lrmd_remote_listen(gpointer data)
     gnutls_session_t *session = NULL;
     crm_client_t *new_client = NULL;
 
+    // For client socket
     static struct mainloop_fd_callbacks lrmd_remote_fd_cb = {
         .dispatch = lrmd_remote_client_msg,
         .destroy = lrmd_remote_client_destroy,
     };
 
+    CRM_CHECK(ssock >= 0, return TRUE);
+
     csock = crm_remote_accept(ssock);
     if (csock < 0) {
         return TRUE;
     }
 
-    session = create_psk_tls_session(csock, GNUTLS_SERVER, psk_cred_s);
+    session = pcmk__new_tls_session(csock, GNUTLS_SERVER, GNUTLS_CRD_PSK,
+                                    psk_cred_s);
     if (session == NULL) {
-        crm_err("TLS session creation failed");
         close(csock);
         return TRUE;
     }
@@ -216,9 +241,13 @@ lrmd_remote_listen(gpointer data)
     new_client->remote = calloc(1, sizeof(crm_remote_t));
     new_client->kind = CRM_CLIENT_TLS;
     new_client->remote->tls_session = session;
-    new_client->remote->auth_timeout =
-        g_timeout_add(LRMD_REMOTE_AUTH_TIMEOUT, lrmd_auth_timeout_cb, new_client);
-    crm_notice("LRMD client connection established. %p id: %s", new_client, new_client->id);
+
+    // Require the client to authenticate within this time
+    new_client->remote->auth_timeout = g_timeout_add(LRMD_REMOTE_AUTH_TIMEOUT,
+                                                     lrmd_auth_timeout_cb,
+                                                     new_client);
+    crm_info("Remote client pending authentication "
+             CRM_XS " %p id: %s", new_client, new_client->id);
 
     new_client->remote->source =
         mainloop_add_fd("lrmd-remote-client", G_PRIORITY_DEFAULT, csock, new_client,
@@ -229,7 +258,7 @@ lrmd_remote_listen(gpointer data)
 static void
 lrmd_remote_connection_destroy(gpointer user_data)
 {
-    crm_notice("Remote tls server disconnected");
+    crm_notice("TLS server session ended");
     return;
 }
 
@@ -248,10 +277,11 @@ bind_and_listen(struct addrinfo *addr)
     char buffer[INET6_ADDRSTRLEN] = { 0, };
 
     crm_sockaddr2str(addr->ai_addr, buffer);
-    crm_trace("Attempting to bind on address %s", buffer);
+    crm_trace("Attempting to bind to address %s", buffer);
 
     fd = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
     if (fd < 0) {
+        crm_perror(LOG_ERR, "Listener socket creation failed");
         return -1;
     }
 
@@ -259,7 +289,7 @@ bind_and_listen(struct addrinfo *addr)
     optval = 1;
     rc = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));
     if (rc < 0) {
-        crm_perror(LOG_INFO, "Couldn't allow the reuse of local addresses by our remote listener, bind address %s", buffer);
+        crm_perror(LOG_ERR, "Local address reuse not allowed on %s", buffer);
         close(fd);
         return -1;
     }
@@ -268,49 +298,73 @@ bind_and_listen(struct addrinfo *addr)
         optval = 0;
         rc = setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &optval, sizeof(optval));
         if (rc < 0) {
-            crm_perror(LOG_INFO, "Couldn't disable IPV6 only on address %s", buffer);
+            crm_perror(LOG_INFO, "Couldn't disable IPV6-only on %s", buffer);
             close(fd);
             return -1;
         }
     }
 
     if (bind(fd, addr->ai_addr, addr->ai_addrlen) != 0) {
+        crm_perror(LOG_ERR, "Cannot bind to %s", buffer);
         close(fd);
         return -1;
     }
 
     if (listen(fd, 10) == -1) {
-        crm_err("Can not start listen on address %s", buffer);
+        crm_perror(LOG_ERR, "Cannot listen on %s", buffer);
         close(fd);
         return -1;
     }
-
-    crm_notice("Listening on address %s", buffer);
-
     return fd;
 }
 
+static int
+get_address_info(const char *bind_name, int port, struct addrinfo **res)
+{
+    int rc;
+    char port_str[6]; // at most "65535"
+    struct addrinfo hints;
+
+    memset(&hints, 0, sizeof(struct addrinfo));
+    hints.ai_flags = AI_PASSIVE;
+    hints.ai_family = AF_UNSPEC; // IPv6 or IPv4
+    hints.ai_socktype = SOCK_STREAM;
+    hints.ai_protocol = IPPROTO_TCP;
+
+    snprintf(port_str, sizeof(port_str), "%d", port);
+    rc = getaddrinfo(bind_name, port_str, &hints, res);
+    if (rc) {
+        crm_err("Unable to get IP address(es) for %s: %s",
+                (bind_name? bind_name : "local node"), gai_strerror(rc));
+        return -EADDRNOTAVAIL;
+    }
+    return pcmk_ok;
+}
+
 int
 lrmd_init_remote_tls_server()
 {
-    int rc;
     int filter;
     int port = crm_default_remote_port();
-    struct addrinfo hints, *res = NULL, *iter;
-    char port_str[6]; // at most "65535"
+    struct addrinfo *res = NULL, *iter;
     gnutls_datum_t psk_key = { NULL, 0 };
+    const char *bind_name = getenv("PCMK_remote_address");
 
     static struct mainloop_fd_callbacks remote_listen_fd_callbacks = {
         .dispatch = lrmd_remote_listen,
         .destroy = lrmd_remote_connection_destroy,
     };
 
-    crm_notice("Starting TLS listener on port %d", port);
+    CRM_CHECK(ssock == -1, return ssock);
+
+    crm_notice("Starting TLS listener on %s port %d",
+               (bind_name? bind_name : "all addresses on"), port);
     crm_gnutls_global_init();
     gnutls_global_set_log_function(debug_log);
 
-    gnutls_dh_params_init(&dh_params);
-    gnutls_dh_params_generate2(dh_params, 1024);
+    if (pcmk__init_tls_dh(&dh_params) != GNUTLS_E_SUCCESS) {
+        return -1;
+    }
     gnutls_psk_allocate_server_credentials(&psk_cred_s);
     gnutls_psk_set_server_credentials_function(psk_cred_s, lrmd_tls_server_key_cb);
     gnutls_psk_set_server_dh_params(psk_cred_s, dh_params);
@@ -320,31 +374,27 @@ lrmd_init_remote_tls_server()
      * read the key. We don't error out, though, because it's fine if the key is
      * going to be added later.
      */
-    rc = lrmd_tls_set_key(&psk_key);
-    if (rc != 0) {
+    if (lrmd_tls_set_key(&psk_key) != pcmk_ok) {
         crm_warn("A cluster connection will not be possible until the key is available");
     }
+    gnutls_free(psk_key.data);
 
-    memset(&hints, 0, sizeof(struct addrinfo));
-    /* Bind to the wildcard address (INADDR_ANY or IN6ADDR_ANY_INIT).
-     * @TODO allow user to specify a specific address
-     */
-    hints.ai_flags = AI_PASSIVE;
-    hints.ai_family = AF_UNSPEC; /* Return IPv6 or IPv4 */
-    hints.ai_socktype = SOCK_STREAM;
-    hints.ai_protocol = IPPROTO_TCP;
-
-    snprintf(port_str, sizeof(port_str), "%d", port);
-    rc = getaddrinfo(NULL, port_str, &hints, &res);
-    if (rc) {
-        crm_err("Unable to get IP address info for local node: %s",
-                gai_strerror(rc));
+    if (get_address_info(bind_name, port, &res) != pcmk_ok) {
         return -1;
     }
 
+    /* Currently we listen on only one address from the resulting list (the
+     * first IPv6 address we can bind to if possible, otherwise the first IPv4
+     * address we can bind to). When bind_name is NULL, this should be the
+     * respective wildcard address.
+     *
+     * @TODO If there is demand for specifying more than one address, allow
+     * bind_name to be a space-separated list, call getaddrinfo() for each,
+     * and create a socket for each result (set IPV6_V6ONLY on IPv6 sockets
+     * since IPv4 listeners will have their own sockets).
+     */
     iter = res;
     filter = AF_INET6;
-    /* Try IPv6 addresses first, then IPv4 */
     while (iter) {
         if (iter->ai_family == filter) {
             ssock = bind_and_listen(iter);
@@ -360,22 +410,14 @@ lrmd_init_remote_tls_server()
         }
     }
 
-    if (ssock < 0) {
-        crm_err("unable to bind to address");
-        goto init_remote_cleanup;
-    }
-
-    mainloop_add_fd("lrmd-remote", G_PRIORITY_DEFAULT, ssock, NULL, &remote_listen_fd_callbacks);
-
-    rc = ssock;
-  init_remote_cleanup:
-    if (rc < 0) {
-        close(ssock);
-        ssock = 0;
+    if (ssock >= 0) {
+        mainloop_add_fd("lrmd-remote", G_PRIORITY_DEFAULT, ssock,
+                        NULL, &remote_listen_fd_callbacks);
+        crm_debug("Started TLS listener on %s port %d",
+                  (bind_name? bind_name : "all addresses on"), port);
     }
     freeaddrinfo(res);
-    return rc;
-
+    return ssock;
 }
 
 void
@@ -386,9 +428,9 @@ lrmd_tls_server_destroy(void)
         psk_cred_s = 0;
     }
 
-    if (ssock > 0) {
+    if (ssock >= 0) {
         close(ssock);
-        ssock = 0;
+        ssock = -1;
     }
 }
 #endif
