@@ -36,7 +36,7 @@ struct resource_alloc_functions_s {
     void (*rsc_colocation_lh) (resource_t *, resource_t *, rsc_colocation_t *);
     void (*rsc_colocation_rh) (resource_t *, resource_t *, rsc_colocation_t *);
 
-    void (*rsc_location) (resource_t *, rsc_to_node_t *);
+    void (*rsc_location) (pe_resource_t *, pe__location_t *);
 
     enum pe_action_flags (*action_flags) (action_t *, node_t *);
     enum pe_graph_flags (*update_actions) (action_t *, action_t *, node_t *, enum pe_action_flags,
@@ -46,25 +46,19 @@ struct resource_alloc_functions_s {
     void (*append_meta) (resource_t * rsc, xmlNode * xml);
 };
 
-extern GHashTable *rsc_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes,
-                                     const char *attr, float factor, enum pe_weights flags);
-
-extern GHashTable *clone_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes,
-                                       const char *attr, float factor, enum pe_weights flags);
-
-extern GHashTable *container_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes,
-                                       const char *attr, float factor, enum pe_weights flags);
-
 extern GHashTable *master_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes,
                                         const char *attr, float factor, enum pe_weights flags);
 
-extern GHashTable *native_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes,
-                                        const char *attr, float factor, enum pe_weights flags);
+GHashTable *pcmk__native_merge_weights(pe_resource_t *rsc, const char *rhs,
+                                       GHashTable *nodes, const char *attr,
+                                       float factor, uint32_t flags);
 
-extern GHashTable *group_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes,
-                                       const char *attr, float factor, enum pe_weights flags);
+GHashTable *pcmk__group_merge_weights(pe_resource_t *rsc, const char *rhs,
+                                      GHashTable *nodes, const char *attr,
+                                      float factor, uint32_t flags);
 
-extern node_t *native_color(resource_t * rsc, node_t * preferred, pe_working_set_t * data_set);
+pe_node_t *pcmk__native_allocate(pe_resource_t *rsc, pe_node_t *preferred,
+                                 pe_working_set_t *data_set);
 extern void native_create_actions(resource_t * rsc, pe_working_set_t * data_set);
 extern void native_internal_constraints(resource_t * rsc, pe_working_set_t * data_set);
 extern void native_rsc_colocation_lh(resource_t * lh_rsc, resource_t * rh_rsc,
@@ -75,7 +69,7 @@ extern void rsc_ticket_constraint(resource_t * lh_rsc, rsc_ticket_t * rsc_ticket
                                   pe_working_set_t * data_set);
 extern enum pe_action_flags native_action_flags(action_t * action, node_t * node);
 
-extern void native_rsc_location(resource_t * rsc, rsc_to_node_t * constraint);
+void native_rsc_location(pe_resource_t *rsc, pe__location_t *constraint);
 extern void native_expand(resource_t * rsc, pe_working_set_t * data_set);
 extern void native_dump(resource_t * rsc, const char *pre_text, gboolean details);
 extern gboolean native_create_probe(resource_t * rsc, node_t * node, action_t * complete,
@@ -83,7 +77,8 @@ extern gboolean native_create_probe(resource_t * rsc, node_t * node, action_t *
 extern void native_append_meta(resource_t * rsc, xmlNode * xml);
 
 extern int group_num_allowed_nodes(resource_t * rsc);
-extern node_t *group_color(resource_t * rsc, node_t * preferred, pe_working_set_t * data_set);
+pe_node_t *pcmk__group_allocate(pe_resource_t *rsc, pe_node_t *preferred,
+                                pe_working_set_t *data_set);
 extern void group_create_actions(resource_t * rsc, pe_working_set_t * data_set);
 extern void group_internal_constraints(resource_t * rsc, pe_working_set_t * data_set);
 extern void group_rsc_colocation_lh(resource_t * lh_rsc, resource_t * rh_rsc,
@@ -91,19 +86,20 @@ extern void group_rsc_colocation_lh(resource_t * lh_rsc, resource_t * rh_rsc,
 extern void group_rsc_colocation_rh(resource_t * lh_rsc, resource_t * rh_rsc,
                                     rsc_colocation_t * constraint);
 extern enum pe_action_flags group_action_flags(action_t * action, node_t * node);
-extern void group_rsc_location(resource_t * rsc, rsc_to_node_t * constraint);
+void group_rsc_location(pe_resource_t *rsc, pe__location_t *constraint);
 extern void group_expand(resource_t * rsc, pe_working_set_t * data_set);
 extern void group_append_meta(resource_t * rsc, xmlNode * xml);
 
 extern int container_num_allowed_nodes(resource_t * rsc);
-extern node_t *container_color(resource_t * rsc, node_t * preferred, pe_working_set_t * data_set);
+pe_node_t *pcmk__bundle_allocate(pe_resource_t *rsc, pe_node_t *preferred,
+                                 pe_working_set_t *data_set);
 extern void container_create_actions(resource_t * rsc, pe_working_set_t * data_set);
 extern void container_internal_constraints(resource_t * rsc, pe_working_set_t * data_set);
 extern void container_rsc_colocation_lh(resource_t * lh_rsc, resource_t * rh_rsc,
                                     rsc_colocation_t * constraint);
 extern void container_rsc_colocation_rh(resource_t * lh_rsc, resource_t * rh_rsc,
                                     rsc_colocation_t * constraint);
-extern void container_rsc_location(resource_t * rsc, rsc_to_node_t * constraint);
+void container_rsc_location(pe_resource_t *rsc, pe__location_t *constraint);
 extern enum pe_action_flags container_action_flags(action_t * action, node_t * node);
 extern void container_expand(resource_t * rsc, pe_working_set_t * data_set);
 extern gboolean container_create_probe(resource_t * rsc, node_t * node, action_t * complete,
@@ -111,14 +107,15 @@ extern gboolean container_create_probe(resource_t * rsc, node_t * node, action_t
 extern void container_append_meta(resource_t * rsc, xmlNode * xml);
 
 extern int clone_num_allowed_nodes(resource_t * rsc);
-extern node_t *clone_color(resource_t * rsc, node_t * preferred, pe_working_set_t * data_set);
+pe_node_t *pcmk__clone_allocate(pe_resource_t *rsc, pe_node_t *preferred,
+                                pe_working_set_t *data_set);
 extern void clone_create_actions(resource_t * rsc, pe_working_set_t * data_set);
 extern void clone_internal_constraints(resource_t * rsc, pe_working_set_t * data_set);
 extern void clone_rsc_colocation_lh(resource_t * lh_rsc, resource_t * rh_rsc,
                                     rsc_colocation_t * constraint);
 extern void clone_rsc_colocation_rh(resource_t * lh_rsc, resource_t * rh_rsc,
                                     rsc_colocation_t * constraint);
-extern void clone_rsc_location(resource_t * rsc, rsc_to_node_t * constraint);
+void clone_rsc_location(pe_resource_t *rsc, pe__location_t *constraint);
 extern enum pe_action_flags clone_action_flags(action_t * action, node_t * node);
 extern void clone_expand(resource_t * rsc, pe_working_set_t * data_set);
 extern gboolean clone_create_probe(resource_t * rsc, node_t * node, action_t * complete,
@@ -126,7 +123,8 @@ extern gboolean clone_create_probe(resource_t * rsc, node_t * node, action_t * c
 extern void clone_append_meta(resource_t * rsc, xmlNode * xml);
 
 extern gboolean master_unpack(resource_t * rsc, pe_working_set_t * data_set);
-extern node_t *master_color(resource_t * rsc, node_t * preferred, pe_working_set_t * data_set);
+pe_node_t *pcmk__set_instance_roles(pe_resource_t *rsc, pe_node_t *preferred,
+                                    pe_working_set_t *data_set);
 extern void master_create_actions(resource_t * rsc, pe_working_set_t * data_set);
 extern void master_internal_constraints(resource_t * rsc, pe_working_set_t * data_set);
 extern void master_rsc_colocation_rh(resource_t * lh_rsc, resource_t * rh_rsc,
@@ -135,7 +133,7 @@ extern void master_append_meta(resource_t * rsc, xmlNode * xml);
 
 /* extern resource_object_functions_t resource_variants[]; */
 extern resource_alloc_functions_t resource_class_alloc_functions[];
-extern gboolean is_active(rsc_to_node_t * cons);
+gboolean is_active(pe__location_t *cons);
 
 extern gboolean native_constraint_violated(resource_t * rsc_lh, resource_t * rsc_rh,
                                            rsc_colocation_t * constraint);
@@ -156,8 +154,6 @@ void LogNodeActions(pe_working_set_t * data_set, gboolean terminal);
 void LogActions(resource_t * rsc, pe_working_set_t * data_set, gboolean terminal);
 void container_LogActions(resource_t * rsc, pe_working_set_t * data_set, gboolean terminal);
 
-extern void cleanup_alloc_calculations(pe_working_set_t * data_set);
-
 extern void rsc_stonith_ordering(resource_t * rsc, action_t * stonith_op,
                                  pe_working_set_t * data_set);
 
@@ -180,4 +176,5 @@ void master_promotion_constraints(resource_t * rsc, pe_working_set_t * data_set)
 void clone_create_pseudo_actions(
     resource_t * rsc, GListPtr children, notify_data_t **start_notify, notify_data_t **stop_notify,  pe_working_set_t * data_set);
 
+void libpengine_fini(void);
 #endif
