@@ -1,19 +1,8 @@
-/* 
- * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
- * 
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- * 
- * This software is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- * 
- * You should have received a copy of the GNU General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+/*
+ * Copyright 2004-2019 Andrew Beekhof <andrew@beekhof.net>
+ *
+ * This source code is licensed under the GNU General Public License version 2
+ * or later (GPLv2+) WITHOUT ANY WARRANTY.
  */
 
 #include <crm_internal.h>
@@ -80,6 +69,10 @@ sort_clone_instance(gconstpointer a, gconstpointer b, gpointer data_set)
     int rc = 0;
     node_t *node1 = NULL;
     node_t *node2 = NULL;
+    node_t *current_node1 = NULL;
+    node_t *current_node2 = NULL;
+    unsigned int nnodes1 = 0;
+    unsigned int nnodes2 = 0;
 
     gboolean can1 = TRUE;
     gboolean can2 = TRUE;
@@ -98,24 +91,22 @@ sort_clone_instance(gconstpointer a, gconstpointer b, gpointer data_set)
      *  - inactive instances
      */
 
-    if (resource1->running_on && resource2->running_on) {
-        if (g_list_length(resource1->running_on) < g_list_length(resource2->running_on)) {
+    current_node1 = pe__find_active_on(resource1, &nnodes1, NULL);
+    current_node2 = pe__find_active_on(resource2, &nnodes2, NULL);
+
+    if (nnodes1 && nnodes2) {
+        if (nnodes1 < nnodes2) {
             crm_trace("%s < %s: running_on", resource1->id, resource2->id);
             return -1;
 
-        } else if (g_list_length(resource1->running_on) > g_list_length(resource2->running_on)) {
+        } else if (nnodes1 > nnodes2) {
             crm_trace("%s > %s: running_on", resource1->id, resource2->id);
             return 1;
         }
     }
 
-    if (resource1->running_on) {
-        node1 = resource1->running_on->data;
-    }
-    if (resource2->running_on) {
-        node2 = resource2->running_on->data;
-    }
-
+    node1 = current_node1;
+    node2 = current_node2;
     if (node1) {
         node_t *match = pe_hash_table_lookup(resource1->allowed_nodes, node1->details->id);
 
@@ -227,31 +218,41 @@ sort_clone_instance(gconstpointer a, gconstpointer b, gpointer data_set)
         GHashTable *hash2 =
             g_hash_table_new_full(crm_str_hash, g_str_equal, NULL, g_hash_destroy_str);
 
-        n = node_copy(resource1->running_on->data);
+        n = node_copy(current_node1);
         g_hash_table_insert(hash1, (gpointer) n->details->id, n);
 
-        n = node_copy(resource2->running_on->data);
+        n = node_copy(current_node2);
         g_hash_table_insert(hash2, (gpointer) n->details->id, n);
 
         if(resource1->parent) {
             for (gIter = resource1->parent->rsc_cons; gIter; gIter = gIter->next) {
                 rsc_colocation_t *constraint = (rsc_colocation_t *) gIter->data;
 
+                if (constraint->score == 0) {
+                    continue;
+                }
                 crm_trace("Applying %s to %s", constraint->id, resource1->id);
 
-                hash1 = native_merge_weights(constraint->rsc_rh, resource1->id, hash1,
-                                             constraint->node_attribute,
-                                             (float)constraint->score / INFINITY, 0);
+                hash1 = pcmk__native_merge_weights(constraint->rsc_rh,
+                                                   resource1->id, hash1,
+                                                   constraint->node_attribute,
+                                                   constraint->score / (float) INFINITY,
+                                                   0);
             }
 
             for (gIter = resource1->parent->rsc_cons_lhs; gIter; gIter = gIter->next) {
                 rsc_colocation_t *constraint = (rsc_colocation_t *) gIter->data;
 
+                if (constraint->score == 0) {
+                    continue;
+                }
                 crm_trace("Applying %s to %s", constraint->id, resource1->id);
 
-                hash1 = native_merge_weights(constraint->rsc_lh, resource1->id, hash1,
-                                             constraint->node_attribute,
-                                             (float)constraint->score / INFINITY, pe_weights_positive);
+                hash1 = pcmk__native_merge_weights(constraint->rsc_lh,
+                                                   resource1->id, hash1,
+                                                   constraint->node_attribute,
+                                                   constraint->score / (float) INFINITY,
+                                                   pe_weights_positive);
             }
         }
 
@@ -261,9 +262,11 @@ sort_clone_instance(gconstpointer a, gconstpointer b, gpointer data_set)
 
                 crm_trace("Applying %s to %s", constraint->id, resource2->id);
 
-                hash2 = native_merge_weights(constraint->rsc_rh, resource2->id, hash2,
-                                             constraint->node_attribute,
-                                             (float)constraint->score / INFINITY, 0);
+                hash2 = pcmk__native_merge_weights(constraint->rsc_rh,
+                                                   resource2->id, hash2,
+                                                   constraint->node_attribute,
+                                                   constraint->score / (float) INFINITY,
+                                                   0);
             }
 
             for (gIter = resource2->parent->rsc_cons_lhs; gIter; gIter = gIter->next) {
@@ -271,18 +274,17 @@ sort_clone_instance(gconstpointer a, gconstpointer b, gpointer data_set)
 
                 crm_trace("Applying %s to %s", constraint->id, resource2->id);
 
-                hash2 = native_merge_weights(constraint->rsc_lh, resource2->id, hash2,
-                                             constraint->node_attribute,
-                                             (float)constraint->score / INFINITY, pe_weights_positive);
+                hash2 = pcmk__native_merge_weights(constraint->rsc_lh,
+                                                   resource2->id, hash2,
+                                                   constraint->node_attribute,
+                                                   constraint->score / (float) INFINITY,
+                                                   pe_weights_positive);
             }
         }
 
         /* Current location score */
-        node1 = g_list_nth_data(resource1->running_on, 0);
-        node1 = g_hash_table_lookup(hash1, node1->details->id);
-
-        node2 = g_list_nth_data(resource2->running_on, 0);
-        node2 = g_hash_table_lookup(hash2, node2->details->id);
+        node1 = g_hash_table_lookup(hash1, current_node1->details->id);
+        node2 = g_hash_table_lookup(hash2, current_node2->details->id);
 
         if (node1->weight < node2->weight) {
             if (node1->weight < 0) {
@@ -306,12 +308,8 @@ sort_clone_instance(gconstpointer a, gconstpointer b, gpointer data_set)
         list1 = g_hash_table_get_values(hash1);
         list2 = g_hash_table_get_values(hash2);
 
-        list1 =
-            g_list_sort_with_data(list1, sort_node_weight,
-                                  g_list_nth_data(resource1->running_on, 0));
-        list2 =
-            g_list_sort_with_data(list2, sort_node_weight,
-                                  g_list_nth_data(resource2->running_on, 0));
+        list1 = g_list_sort_with_data(list1, sort_node_weight, current_node1);
+        list2 = g_list_sort_with_data(list2, sort_node_weight, current_node2);
         max = g_list_length(list1);
         if (max < g_list_length(list2)) {
             max = g_list_length(list2);
@@ -409,8 +407,9 @@ can_run_instance(resource_t * rsc, node_t * node, int limit)
     return NULL;
 }
 
-static node_t *
-color_instance(resource_t * rsc, node_t * prefer, gboolean all_coloc, int limit, pe_working_set_t * data_set)
+static pe_node_t *
+allocate_instance(pe_resource_t *rsc, pe_node_t *prefer, gboolean all_coloc,
+                  int limit, pe_working_set_t *data_set)
 {
     node_t *chosen = NULL;
     GHashTable *backup = NULL;
@@ -446,10 +445,11 @@ color_instance(resource_t * rsc, node_t * prefer, gboolean all_coloc, int limit,
     can_run_instance(rsc, NULL, limit);
 
     backup = node_hash_dup(rsc->allowed_nodes);
+    pe_rsc_trace(rsc, "Allocating instance %s", rsc->id);
     chosen = rsc->cmds->allocate(rsc, prefer, data_set);
     if (chosen) {
         node_t *local_node = parent_node_instance(rsc, chosen);
-        if (prefer && chosen && chosen->details != prefer->details) {
+        if (prefer && (chosen->details != prefer->details)) {
             crm_notice("Pre-allocation failed: got %s instead of %s",
                        chosen->details->uname, prefer->details->uname);
             g_hash_table_destroy(rsc->allowed_nodes);
@@ -485,6 +485,9 @@ append_parent_colocation(resource_t * rsc, resource_t * child, gboolean all)
     for (; gIter != NULL; gIter = gIter->next) {
         rsc_colocation_t *cons = (rsc_colocation_t *) gIter->data;
 
+        if (cons->score == 0) {
+            continue;
+        }
         if (all || cons->score < 0 || cons->score == INFINITY) {
             child->rsc_cons = g_list_prepend(child->rsc_cons, cons);
         }
@@ -494,6 +497,9 @@ append_parent_colocation(resource_t * rsc, resource_t * child, gboolean all)
     for (; gIter != NULL; gIter = gIter->next) {
         rsc_colocation_t *cons = (rsc_colocation_t *) gIter->data;
 
+        if (cons->score == 0) {
+            continue;
+        }
         if (all || cons->score < 0) {
             child->rsc_cons_lhs = g_list_prepend(child->rsc_cons_lhs, cons);
         }
@@ -539,8 +545,8 @@ distribute_children(resource_t *rsc, GListPtr children, GListPtr nodes,
 
         if (child->running_on && is_set(child->flags, pe_rsc_provisional)
             && is_not_set(child->flags, pe_rsc_failed)) {
-            node_t *child_node = child->running_on->data;
-            node_t *local_node = parent_node_instance(child, child->running_on->data);
+            node_t *child_node = pe__current_node(child);
+            node_t *local_node = parent_node_instance(child, child_node);
 
             pe_rsc_trace(rsc, "Checking pre-allocation of %s to %s (%d remaining of %d)",
                          child->id, child_node->details->uname, max - allocated, max);
@@ -554,7 +560,9 @@ distribute_children(resource_t *rsc, GListPtr children, GListPtr nodes,
                              "Not pre-allocating because %s already allocated optimal instances",
                              child_node->details->uname);
 
-            } else if (color_instance(child, child_node, max < available_nodes, per_host_max, data_set)) {
+            } else if (allocate_instance(child, child_node,
+                                         max < available_nodes, per_host_max,
+                                         data_set)) {
                 pe_rsc_trace(rsc, "Pre-allocated %s to %s", child->id,
                              child_node->details->uname);
                 allocated++;
@@ -567,9 +575,9 @@ distribute_children(resource_t *rsc, GListPtr children, GListPtr nodes,
     for (GListPtr gIter = children; gIter != NULL; gIter = gIter->next) {
         resource_t *child = (resource_t *) gIter->data;
 
-        if (g_list_length(child->running_on) > 0) {
-            node_t *child_node = child->running_on->data;
-            node_t *local_node = parent_node_instance(child, child->running_on->data);
+        if (child->running_on != NULL) {
+            node_t *child_node = pe__current_node(child);
+            node_t *local_node = parent_node_instance(child, child_node);
 
             if (local_node == NULL) {
                 crm_err("%s is running on %s which isn't allowed",
@@ -580,9 +588,10 @@ distribute_children(resource_t *rsc, GListPtr children, GListPtr nodes,
         if (is_not_set(child->flags, pe_rsc_provisional)) {
         } else if (allocated >= max) {
             pe_rsc_debug(rsc, "Child %s not allocated - limit reached %d %d", child->id, allocated, max);
-            resource_location(child, NULL, -INFINITY, "clone_color:limit_reached", data_set);
+            resource_location(child, NULL, -INFINITY, "clone:limit_reached", data_set);
         } else {
-            if (color_instance(child, NULL, max < available_nodes, per_host_max, data_set)) {
+            if (allocate_instance(child, NULL, max < available_nodes,
+                                  per_host_max, data_set)) {
                 allocated++;
             }
         }
@@ -593,8 +602,9 @@ distribute_children(resource_t *rsc, GListPtr children, GListPtr nodes,
 }
 
 
-node_t *
-clone_color(resource_t * rsc, node_t * prefer, pe_working_set_t * data_set)
+pe_node_t *
+pcmk__clone_allocate(pe_resource_t *rsc, pe_node_t *prefer,
+                     pe_working_set_t *data_set)
 {
     GListPtr nodes = NULL;
 
@@ -611,7 +621,6 @@ clone_color(resource_t * rsc, node_t * prefer, pe_working_set_t * data_set)
     }
 
     set_bit(rsc->flags, pe_rsc_allocating);
-    pe_rsc_trace(rsc, "Processing %s", rsc->id);
 
     /* this information is used by sort_clone_instance() when deciding in which 
      * order to allocate clone instances
@@ -619,13 +628,20 @@ clone_color(resource_t * rsc, node_t * prefer, pe_working_set_t * data_set)
     for (GListPtr gIter = rsc->rsc_cons; gIter != NULL; gIter = gIter->next) {
         rsc_colocation_t *constraint = (rsc_colocation_t *) gIter->data;
 
-        pe_rsc_trace(rsc, "%s: Coloring %s first", rsc->id, constraint->rsc_rh->id);
+        if (constraint->score == 0) {
+            continue;
+        }
+        pe_rsc_trace(rsc, "%s: Allocating %s first",
+                     rsc->id, constraint->rsc_rh->id);
         constraint->rsc_rh->cmds->allocate(constraint->rsc_rh, prefer, data_set);
     }
 
     for (GListPtr gIter = rsc->rsc_cons_lhs; gIter != NULL; gIter = gIter->next) {
         rsc_colocation_t *constraint = (rsc_colocation_t *) gIter->data;
 
+        if (constraint->score == 0) {
+            continue;
+        }
         rsc->allowed_nodes =
             constraint->rsc_lh->cmds->merge_weights(constraint->rsc_lh, rsc->id, rsc->allowed_nodes,
                                                     constraint->node_attribute,
@@ -633,7 +649,7 @@ clone_color(resource_t * rsc, node_t * prefer, pe_working_set_t * data_set)
                                                     (pe_weights_rollback | pe_weights_positive));
     }
 
-    dump_node_scores(show_scores ? 0 : scores_log_level, rsc, __FUNCTION__, rsc->allowed_nodes);
+    pe__show_node_weights(!show_scores, rsc, __FUNCTION__, rsc->allowed_nodes);
 
     nodes = g_hash_table_get_values(rsc->allowed_nodes);
     nodes = g_list_sort_with_data(nodes, sort_node_weight, NULL);
@@ -955,6 +971,7 @@ is_child_compatible(resource_t *child_rsc, node_t * local_node, enum rsc_role_e
     node_t *node = NULL;
     enum rsc_role_e next_role = child_rsc->fns->state(child_rsc, current);
 
+    CRM_CHECK(child_rsc && local_node, return FALSE);
     if (is_set_recursive(child_rsc, pe_rsc_block, TRUE) == FALSE) {
         /* We only want instances that haven't failed */
         node = child_rsc->fns->location(child_rsc, NULL, current);
@@ -965,7 +982,7 @@ is_child_compatible(resource_t *child_rsc, node_t * local_node, enum rsc_role_e
         return FALSE;
     }
 
-    if (node && local_node && node->details == local_node->details) {
+    if (node && (node->details == local_node->details)) {
         return TRUE;
 
     } else if (node) {
@@ -1032,6 +1049,9 @@ clone_rsc_colocation_rh(resource_t * rsc_lh, resource_t * rsc_rh, rsc_colocation
     CRM_CHECK(rsc_rh != NULL, pe_err("rsc_rh was NULL for %s", constraint->id); return);
     CRM_CHECK(rsc_lh->variant == pe_native, return);
 
+    if (constraint->score == 0) {
+        return;
+    }
     pe_rsc_trace(rsc_rh, "Processing constraint %s: %s -> %s %d",
                  constraint->id, rsc_lh->id, rsc_rh->id, constraint->score);
 
@@ -1189,7 +1209,7 @@ clone_action_flags(action_t * action, node_t * node)
 }
 
 void
-clone_rsc_location(resource_t * rsc, rsc_to_node_t * constraint)
+clone_rsc_location(pe_resource_t *rsc, pe__location_t *constraint)
 {
     GListPtr gIter = rsc->children;
 
@@ -1265,85 +1285,105 @@ clone_expand(resource_t * rsc, pe_working_set_t * data_set)
     clone_data->promote_notify = NULL;
 }
 
-node_t *
-rsc_known_on(resource_t * rsc, GListPtr * list)
+// Check whether a resource or any of its children is known on node
+static bool
+rsc_known_on(const pe_resource_t *rsc, const pe_node_t *node)
 {
-    GListPtr gIter = NULL;
-    node_t *one = NULL;
-    GListPtr result = NULL;
-
     if (rsc->children) {
+        for (GList *child_iter = rsc->children; child_iter != NULL;
+             child_iter = child_iter->next) {
 
-        gIter = rsc->children;
-        for (; gIter != NULL; gIter = gIter->next) {
-            resource_t *child = (resource_t *) gIter->data;
+            resource_t *child = (resource_t *) child_iter->data;
 
-            rsc_known_on(child, &result);
+            if (rsc_known_on(child, node)) {
+                return TRUE;
+            }
         }
 
     } else if (rsc->known_on) {
-        result = g_hash_table_get_values(rsc->known_on);
-    }
+        GHashTableIter iter;
+        node_t *known_node = NULL;
 
-    if (result && g_list_length(result) == 1) {
-        one = g_list_nth_data(result, 0);
-    }
-
-    if (list) {
-        GListPtr gIter = NULL;
-
-        gIter = result;
-        for (; gIter != NULL; gIter = gIter->next) {
-            node_t *node = (node_t *) gIter->data;
-
-            if (*list == NULL || pe_find_node_id(*list, node->details->id) == NULL) {
-                *list = g_list_prepend(*list, node);
+        g_hash_table_iter_init(&iter, rsc->known_on);
+        while (g_hash_table_iter_next(&iter, NULL, (gpointer *) &known_node)) {
+            if (node->details == known_node->details) {
+                return TRUE;
             }
         }
     }
-
-    g_list_free(result);
-    return one;
+    return FALSE;
 }
 
-static resource_t *
-find_instance_on(resource_t * rsc, node_t * node)
+// Look for an instance of clone that is known on node
+static pe_resource_t *
+find_instance_on(const pe_resource_t *clone, const pe_node_t *node)
 {
-    GListPtr gIter = NULL;
-
-    gIter = rsc->children;
-    for (; gIter != NULL; gIter = gIter->next) {
-        GListPtr gIter2 = NULL;
-        GListPtr known_list = NULL;
+    for (GList *gIter = clone->children; gIter != NULL; gIter = gIter->next) {
         resource_t *child = (resource_t *) gIter->data;
 
-        rsc_known_on(child, &known_list);
+        if (rsc_known_on(child, node)) {
+            return child;
+        }
+    }
+    return NULL;
+}
 
-        gIter2 = known_list;
-        for (; gIter2 != NULL; gIter2 = gIter2->next) {
-            node_t *known = (node_t *) gIter2->data;
+// For unique clones, probe each instance separately
+static gboolean
+probe_unique_clone(pe_resource_t *rsc, pe_node_t *node, pe_action_t *complete,
+                   gboolean force, pe_working_set_t *data_set)
+{
+    gboolean any_created = FALSE;
 
-            if (node->details == known->details) {
-                g_list_free(known_list);
-                return child;
+    for (GList *child_iter = rsc->children; child_iter != NULL;
+         child_iter = child_iter->next) {
+
+        resource_t *child = (resource_t *) child_iter->data;
+
+        any_created |= child->cmds->create_probe(child, node, complete, force,
+                                                 data_set);
+    }
+    return any_created;
+}
+
+// For anonymous clones, only a single instance needs to be probed
+static gboolean
+probe_anonymous_clone(pe_resource_t *rsc, pe_node_t *node,
+                      pe_action_t *complete, gboolean force,
+                      pe_working_set_t *data_set)
+{
+    // First, check if we probed an instance on this node last time
+    pe_resource_t *child = find_instance_on(rsc, node);
+
+    // Otherwise, check if we plan to start an instance on this node
+    if (child == NULL) {
+        for (GList *child_iter = rsc->children; child_iter && !child;
+             child_iter = child_iter->next) {
+
+            node_t *local_node = NULL;
+            resource_t *child_rsc = (resource_t *) child_iter->data;
+
+            local_node = child_rsc->fns->location(child_rsc, NULL, FALSE);
+            if (local_node && (local_node->details == node->details)) {
+                child = child_rsc;
             }
         }
-        g_list_free(known_list);
     }
 
-    return NULL;
+    // Otherwise, use the first clone instance
+    if (child == NULL) {
+        child = rsc->children->data;
+    }
+    return child->cmds->create_probe(child, node, complete, force, data_set);
 }
 
 gboolean
 clone_create_probe(resource_t * rsc, node_t * node, action_t * complete,
                    gboolean force, pe_working_set_t * data_set)
 {
-    GListPtr gIter = NULL;
     gboolean any_created = FALSE;
-    clone_variant_data_t *clone_data = NULL;
 
     CRM_ASSERT(rsc);
-    get_clone_variant_data(clone_data, rsc);
 
     rsc->children = g_list_sort(rsc->children, sort_rsc_id);
     if (rsc->children == NULL) {
@@ -1368,55 +1408,12 @@ clone_create_probe(resource_t * rsc, node_t * node, action_t * complete,
         }
     }
 
-    if (is_not_set(rsc->flags, pe_rsc_unique)
-        && clone_data->clone_node_max == 1) {
-        /* only look for one copy */
-        resource_t *child = NULL;
-
-        /* Try whoever we probed last time */
-        child = find_instance_on(rsc, node);
-        if (child) {
-            return child->cmds->create_probe(child, node, complete, force, data_set);
-        }
-
-        /* Try whoever we plan on starting there */
-        gIter = rsc->children;
-        for (; gIter != NULL; gIter = gIter->next) {
-            node_t *local_node = NULL;
-            resource_t *child_rsc = (resource_t *) gIter->data;
-
-            CRM_ASSERT(child_rsc);
-            local_node = child_rsc->fns->location(child_rsc, NULL, FALSE);
-            if (local_node == NULL) {
-                continue;
-            }
-
-            if (local_node->details == node->details) {
-                return child_rsc->cmds->create_probe(child_rsc, node, complete, force, data_set);
-            }
-        }
-
-        /* Fall back to the first clone instance */
-        CRM_ASSERT(rsc->children);
-        child = rsc->children->data;
-        return child->cmds->create_probe(child, node, complete, force, data_set);
+    if (is_set(rsc->flags, pe_rsc_unique)) {
+        any_created = probe_unique_clone(rsc, node, complete, force, data_set);
+    } else {
+        any_created = probe_anonymous_clone(rsc, node, complete, force,
+                                            data_set);
     }
-
-    gIter = rsc->children;
-    for (; gIter != NULL; gIter = gIter->next) {
-        resource_t *child_rsc = (resource_t *) gIter->data;
-
-        if (child_rsc->cmds->create_probe(child_rsc, node, complete, force, data_set)) {
-            any_created = TRUE;
-        }
-
-        if (any_created && is_not_set(rsc->flags, pe_rsc_unique)
-            && clone_data->clone_node_max == 1) {
-            /* only look for one copy (clone :0) */
-            break;
-        }
-    }
-
     return any_created;
 }
 
@@ -1444,10 +1441,3 @@ clone_append_meta(resource_t * rsc, xmlNode * xml)
     crm_xml_add_int(xml, name, clone_data->clone_node_max);
     free(name);
 }
-
-GHashTable *
-clone_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes, const char *attr,
-                    float factor, enum pe_weights flags)
-{
-    return rsc_merge_weights(rsc, rhs, nodes, attr, factor, flags);
-}
