@@ -1,19 +1,8 @@
-/* 
- * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
- * 
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- * 
- * This software is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- * 
- * You should have received a copy of the GNU General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+/*
+ * Copyright 2004-2018 Andrew Beekhof <andrew@beekhof.net>
+ *
+ * This source code is licensed under the GNU General Public License version 2
+ * or later (GPLv2+) WITHOUT ANY WARRANTY.
  */
 
 #include <crm_internal.h>
@@ -27,8 +16,9 @@
 #define VARIANT_GROUP 1
 #include <lib/pengine/variant.h>
 
-node_t *
-group_color(resource_t * rsc, node_t * prefer, pe_working_set_t * data_set)
+pe_node_t *
+pcmk__group_allocate(pe_resource_t *rsc, pe_node_t *prefer,
+                     pe_working_set_t *data_set)
 {
     node_t *node = NULL;
     node_t *group_node = NULL;
@@ -40,7 +30,6 @@ group_color(resource_t * rsc, node_t * prefer, pe_working_set_t * data_set)
     if (is_not_set(rsc->flags, pe_rsc_provisional)) {
         return rsc->allocated_to;
     }
-    pe_rsc_trace(rsc, "Processing %s", rsc->id);
     if (is_set(rsc->flags, pe_rsc_allocating)) {
         pe_rsc_debug(rsc, "Dependency loop detected involving %s", rsc->id);
         return NULL;
@@ -63,13 +52,14 @@ group_color(resource_t * rsc, node_t * prefer, pe_working_set_t * data_set)
         g_list_concat(group_data->last_child->rsc_cons_lhs, rsc->rsc_cons_lhs);
     rsc->rsc_cons_lhs = NULL;
 
-    dump_node_scores(show_scores ? 0 : scores_log_level, rsc, __FUNCTION__,
-                     rsc->allowed_nodes);
+    pe__show_node_weights(!show_scores, rsc, __FUNCTION__, rsc->allowed_nodes);
 
     gIter = rsc->children;
     for (; gIter != NULL; gIter = gIter->next) {
         resource_t *child_rsc = (resource_t *) gIter->data;
 
+        pe_rsc_trace(rsc, "Allocating group %s member %s",
+                     rsc->id, child_rsc->id);
         node = child_rsc->cmds->allocate(child_rsc, prefer, data_set);
         if (group_node == NULL) {
             group_node = node;
@@ -305,6 +295,9 @@ group_rsc_colocation_lh(resource_t * rsc_lh, resource_t * rsc_rh, rsc_colocation
         pe_err("rsc_rh was NULL for %s", constraint->id);
         return;
     }
+    if (constraint->score == 0) {
+        return;
+    }
 
     gIter = rsc_lh->children;
     pe_rsc_trace(rsc_lh, "Processing constraints from %s", rsc_lh->id);
@@ -338,8 +331,11 @@ group_rsc_colocation_rh(resource_t * rsc_lh, resource_t * rsc_rh, rsc_colocation
     get_group_variant_data(group_data, rsc_rh);
     CRM_CHECK(rsc_lh->variant == pe_native, return);
 
-    pe_rsc_trace(rsc_rh, "Processing RH of constraint %s", constraint->id);
-    print_resource(LOG_DEBUG_3, "LHS", rsc_lh, TRUE);
+    if (constraint->score == 0) {
+        return;
+    }
+    pe_rsc_trace(rsc_rh, "Processing RH %s of constraint %s (LH is %s)",
+                 rsc_rh->id, constraint->id, rsc_lh->id);
 
     if (is_set(rsc_rh->flags, pe_rsc_provisional)) {
         return;
@@ -434,7 +430,7 @@ group_update_actions(action_t * first, action_t * then, node_t * node, enum pe_a
 }
 
 void
-group_rsc_location(resource_t * rsc, rsc_to_node_t * constraint)
+group_rsc_location(pe_resource_t *rsc, pe__location_t *constraint)
 {
     GListPtr gIter = rsc->children;
     GListPtr saved = constraint->node_list_rh;
@@ -465,14 +461,12 @@ group_rsc_location(resource_t * rsc, rsc_to_node_t * constraint)
 void
 group_expand(resource_t * rsc, pe_working_set_t * data_set)
 {
-    GListPtr gIter = rsc->children;
+    CRM_CHECK(rsc != NULL, return);
 
     pe_rsc_trace(rsc, "Processing actions from %s", rsc->id);
-
-    CRM_CHECK(rsc != NULL, return);
     native_expand(rsc, data_set);
 
-    for (; gIter != NULL; gIter = gIter->next) {
+    for (GListPtr gIter = rsc->children; gIter != NULL; gIter = gIter->next) {
         resource_t *child_rsc = (resource_t *) gIter->data;
 
         child_rsc->cmds->expand(child_rsc, data_set);
@@ -480,8 +474,9 @@ group_expand(resource_t * rsc, pe_working_set_t * data_set)
 }
 
 GHashTable *
-group_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes, const char *attr,
-                    float factor, enum pe_weights flags)
+pcmk__group_merge_weights(pe_resource_t *rsc, const char *rhs,
+                          GHashTable *nodes, const char *attr, float factor,
+                          uint32_t flags)
 {
     GListPtr gIter = rsc->rsc_cons_lhs;
     group_variant_data_t *group_data = NULL;
@@ -502,9 +497,13 @@ group_merge_weights(resource_t * rsc, const char *rhs, GHashTable * nodes, const
     for (; gIter != NULL; gIter = gIter->next) {
         rsc_colocation_t *constraint = (rsc_colocation_t *) gIter->data;
 
-        nodes = native_merge_weights(constraint->rsc_lh, rsc->id, nodes,
-                                     constraint->node_attribute,
-                                     (float)constraint->score / INFINITY, flags);
+        if (constraint->score == 0) {
+            continue;
+        }
+        nodes = pcmk__native_merge_weights(constraint->rsc_lh, rsc->id, nodes,
+                                           constraint->node_attribute,
+                                           constraint->score / (float) INFINITY,
+                                           flags);
     }
 
     clear_bit(rsc->flags, pe_rsc_merging);
