@@ -68,6 +68,7 @@ do_test group-fail "Ensure stop order is preserved for partially active groups"
 do_test group-unmanaged "No need to restart r115 because r114 is unmanaged"
 do_test group-unmanaged-stopped "Make sure r115 is stopped when r114 fails"
 do_test group-dependents "Account for the location preferences of things colocated with a group"
+do_test group-stop-ordering "Ensure blocked group member stop does not force other member stops"
 
 echo ""
 do_test rsc_dep1 "Must not     "
@@ -184,6 +185,8 @@ do_test anti-colocation-master "Organize order of actions for master resources i
 do_test anti-colocation-slave "Organize order of actions for slave resources in anti-colocations"
 do_test enforce-colo1 "Always enforce B with A INFINITY."
 do_test complex_enforce_colo "Always enforce B with A INFINITY. (make sure heat-engine stops)"
+do_test coloc-dependee-should-stay "Stickiness outweighs gorup colocation"
+do_test coloc-dependee-should-move "Group colocation outweighs stickiness"
 
 echo ""
 do_test rsc-sets-seq-true "Resource Sets - sequential=false"
@@ -283,6 +286,7 @@ do_test migrate-fail-6 "Failed migrate_to"
 do_test migrate-fail-7 "Failed migrate_to + stop on source"
 do_test migrate-fail-8 "Failed migrate_to + stop on target - ideally we wouldn't need to re-stop on target"
 do_test migrate-fail-9 "Failed migrate_to + stop on source and target"
+do_test migration-ping-pong "Old migrate_to failure + successful migrate_from on same node"
 
 do_test migrate-stop "Migration in a stopping stack"
 do_test migrate-start "Migration in a starting stack"
@@ -334,6 +338,8 @@ echo ""
 do_test clone-anon-probe-1 "Probe the correct (anonymous) clone instance for each node"
 do_test clone-anon-probe-2 "Avoid needless re-probing of anonymous clones"
 do_test clone-anon-failcount "Merge failcounts for anonymous clones"
+do_test force-anon-clone-max "Update clone-max properly when forcing a clone to be anonymous"
+do_test anon-instance-pending "Assign anonymous clone instance numbers properly when action pending"
 do_test inc0 "Incarnation start"
 do_test inc1 "Incarnation start order"
 do_test inc2 "Incarnation silent restart, stop, move"
@@ -393,6 +399,7 @@ echo ""
 do_test unfence-startup "Clean unfencing"
 do_test unfence-definition "Unfencing when the agent changes"
 do_test unfence-parameters "Unfencing when the agent parameters changes"
+do_test unfence-device "Unfencing when a cluster has only fence devices"
 
 echo ""
 do_test master-0 "Stopped -> Slave"
@@ -480,6 +487,8 @@ do_test notify-2 "Notify simple, confirm"
 do_test notify-3 "Notify move, confirm"
 do_test novell-239079 "Notification priority"
 #do_test notify-2 "Notify - 764"
+do_test route-remote-notify "Route remote notify actions through correct cluster node"
+do_test notify-behind-stopping-remote "Don't schedule notifications behind stopped remote"
 
 echo ""
 do_test 594 "OSDL #594 - Unrunnable actions scheduled in transition"
@@ -502,6 +511,7 @@ do_test stonith-1 "Stonith loop - 2"
 do_test stonith-2 "Stonith loop - 3"
 do_test stonith-3 "Stonith startup"
 do_test stonith-4 "Stonith node state"
+do_test dc-fence-ordering "DC needs fencing while other nodes are shutting down"
 do_test bug-1572-1 "Recovery of groups depending on master/slave"
 do_test bug-1572-2 "Recovery of groups depending on master/slave when the master is never re-promoted"
 do_test bug-1685 "Depends-on-master ordering"
@@ -548,6 +558,7 @@ do_test bug-5069-op-disabled "Test on-fail-ignore with failure when monitor is d
 do_test obsolete-lrm-resource "cl#5115 - Do not use obsolete lrm_resource sections"
 do_test expire-non-blocked-failure "Ignore failure-timeout only if the failed operation has on-fail=block"
 do_test asymmetrical-order-move "Respect asymmetrical ordering when trying to move resources"
+do_test asymmetrical-order-restart "Respect asymmetrical ordering when restarting dependent resource"
 do_test start-then-stop-with-unfence "Avoid graph loop with start-then-stop constraint plus unfencing"
 
 do_test ignore_stonith_rsc_order1 "cl#5056- Ignore order constraint between stonith and non-stonith rsc."
@@ -559,8 +570,10 @@ do_test honor_stonith_rsc_order2 "cl#5056- Honor order constraint, stonith clone
 do_test honor_stonith_rsc_order3 "cl#5056- Honor order constraint, stonith clones with nested pure stonith group."
 do_test honor_stonith_rsc_order4 "cl#5056- Honor order constraint, between two native stonith rscs."
 do_test probe-timeout "cl#5099 - Default probe timeout"
+do_test order-first-probes "cl#5301 - respect order constraints when relevant resources are being probed"
 
 do_test concurrent-fencing "Allow performing fencing operations in parallel"
+#do_test priority-fencing-delay "Delay fencing targeting the more significant node"
 
 echo ""
 do_test systemhealth1  "System Health ()               #1"
@@ -594,7 +607,7 @@ echo ""
 do_test utilization-order1 "Utilization Order - Simple"
 do_test utilization-order2 "Utilization Order - Complex"
 do_test utilization-order3 "Utilization Order - Migrate"
-do_test utilization-order4 "Utilization Order - Live Mirgration (bnc#695440)"
+do_test utilization-order4 "Utilization Order - Live Migration (bnc#695440)"
 do_test utilization-shuffle "Don't displace prmExPostgreSQLDB2 on act2, Start prmExPostgreSQLDB1 on act3"
 do_test load-stopped-loop "Avoid transition loop due to load_stopped (cl#5044)"
 do_test load-stopped-loop-2 "cl#5235 - Prevent graph loops that can be introduced by load_stopped -> migrate_to ordering"
@@ -818,6 +831,7 @@ do_test bundle-order-partial-start "Bundle startup ordering when some dependanci
 do_test bundle-order-partial-start-2 "Bundle startup ordering when some dependancies and the container are already running"
 do_test bundle-order-stop    "Bundle stop ordering"
 do_test bundle-order-partial-stop "Bundle startup ordering when some dependancies are already stopped"
+do_test bundle-order-stop-on-remote "Stop nested resource after bringing up the connection"
 
 do_test bundle-order-startup-clone "Prevent startup because bundle isn't promoted"
 do_test bundle-order-startup-clone-2 "Bundle startup with clones"
@@ -830,6 +844,8 @@ do_test bundle-probe-order-1 "order 1"
 do_test bundle-probe-order-2 "order 2"
 do_test bundle-probe-order-3 "order 3"
 do_test bundle-probe-remotes "Ensure remotes get probed too"
+do_test bundle-replicas-change "Change bundle from 1 replica to multiple"
+do_test nested-remote-recovery "Recover bundle's container hosted on remote node"
 
 echo ""
 do_test whitebox-fail1 "Fail whitebox container rsc."
@@ -848,6 +864,8 @@ do_test whitebox-migrate1 "Migrate both container and connection resource"
 do_test whitebox-imply-stop-on-fence "imply stop action on container node rsc when host node is fenced"
 do_test whitebox-nested-group "Verify guest remote-node works nested in a group"
 do_test guest-node-host-dies "Verify guest node is recovered if host goes away"
+do_test guest-node-cleanup "Order guest node connection recovery after container probe"
+do_test guest-host-not-fenceable "Actions on guest node are unrunnable if host is unclean and cannot be fenced"
 
 echo ""
 do_test remote-startup-probes  "Baremetal remote-node startup probes"
