@@ -231,7 +231,8 @@ add_unallocated_utilization(GHashTable * all_utilization, resource_t * rsc,
             }
         }
 
-        if (existing == FALSE) {
+        // rsc->children is always non-NULL but this makes static analysis happy
+        if (!existing && (rsc->children != NULL)) {
             resource_t *first_child = (resource_t *) rsc->children->data;
 
             pe_rsc_trace(orig_rsc, "%s: Adding %s as colocated utilization",
@@ -339,8 +340,7 @@ find_colocated_rscs(GListPtr colocated_rscs, resource_t * rsc, resource_t * orig
 void
 process_utilization(resource_t * rsc, node_t ** prefer, pe_working_set_t * data_set)
 {
-    int alloc_details = scores_log_level + 1;
-
+    CRM_CHECK(rsc && prefer && data_set, return);
     if (safe_str_neq(data_set->placement_strategy, "default")) {
         GHashTableIter iter;
         GListPtr colocated_rscs = NULL;
@@ -416,7 +416,7 @@ process_utilization(resource_t * rsc, node_t ** prefer, pe_working_set_t * data_
                 }
             }
         }
-        dump_node_scores(alloc_details, rsc, "Post-utilization", rsc->allowed_nodes);
+        pe__show_node_weights(true, rsc, "Post-utilization", rsc->allowed_nodes);
     }
 }
 
