@@ -126,6 +126,8 @@ main(int argc, char **argv)
     int option_index = 0;
     int is_remote_node = 0;
 
+    bool try_attrd = true;
+
     crm_log_cli_init("crm_attribute");
     crm_set_options(NULL, "<command> -n <attribute> [options]", long_options,
                     "Manage node's attributes and cluster options."
@@ -281,8 +283,15 @@ main(int argc, char **argv)
         attr_name = attr_pattern;
     }
 
-    if (((command == 'v') || (command == 'D') || (command == 'u'))
-        && safe_str_eq(type, XML_CIB_TAG_STATUS)
+    // Only go through attribute manager for transient attributes
+    try_attrd = safe_str_eq(type, XML_CIB_TAG_STATUS);
+
+    // Don't try to contact attribute manager if we're using a file as CIB
+    if (getenv("CIB_file") || getenv("CIB_shadow")) {
+        try_attrd = FALSE;
+    }
+
+    if (((command == 'v') || (command == 'D') || (command == 'u')) && try_attrd
         && pcmk_ok == attrd_update_delegate(NULL, command, dest_uname, attr_name,
                                             attr_value, type, set_name, NULL, NULL,
                                             is_remote_node?attrd_opt_remote:attrd_opt_none)) {
@@ -323,7 +332,8 @@ main(int argc, char **argv)
         crm_info("Read %s=%s %s%s",
                  attr_name, crm_str(read_value), set_name ? "in " : "", set_name ? set_name : "");
 
-        if (rc == -EINVAL) {
+        if (rc == -ENOTUNIQ) {
+            // Multiple matches (already displayed) are not error for queries
             rc = pcmk_ok;
 
         } else if (BE_QUIET == FALSE) {
@@ -339,7 +349,7 @@ main(int argc, char **argv)
         free(read_value);
     }
 
-    if (rc == -EINVAL) {
+    if (rc == -ENOTUNIQ) {
         printf("Please choose from one of the matches above and supply the 'id' with --attr-id\n");
 
     } else if (rc != pcmk_ok) {
