@@ -39,14 +39,14 @@ int cib_options = cib_sync_call;
 
 GMainLoop *mainloop = NULL;
 
-#define message_timeout_ms 60*1000
+#define MESSAGE_TIMEOUT_S 60
 
 static gboolean
 resource_ipc_timeout(gpointer data)
 {
-    fprintf(stderr, "No messages received in %d seconds.. aborting\n",
-            (int)message_timeout_ms / 1000);
-    crm_err("No messages received in %d seconds", (int)message_timeout_ms / 1000);
+    fprintf(stderr, "Aborting because no messages received in %d seconds\n",
+            MESSAGE_TIMEOUT_S);
+    crm_err("No messages received in %d seconds", MESSAGE_TIMEOUT_S);
     return crm_exit(-1);
 }
 
@@ -65,10 +65,12 @@ start_mainloop(void)
     }
 
     mainloop = g_main_new(FALSE);
-    fprintf(stderr, "Waiting for %d replies from the CRMd", crmd_replies_needed);
-    crm_debug("Waiting for %d replies from the CRMd", crmd_replies_needed);
+    fprintf(stderr, "Waiting for %d repl%s from the CRMd",
+            crmd_replies_needed, (crmd_replies_needed == 1)? "y" : "ies");
+    crm_debug("Waiting for %d repl%s from the CRMd",
+              crmd_replies_needed, (crmd_replies_needed == 1)? "y" : "ies");
 
-    g_timeout_add(message_timeout_ms, resource_ipc_timeout, NULL);
+    g_timeout_add(MESSAGE_TIMEOUT_S * 1000, resource_ipc_timeout, NULL);
     g_main_run(mainloop);
 }
 
@@ -214,15 +216,20 @@ static struct crm_option long_options[] = {
         "cleanup", no_argument, NULL, 'C',
 #if 0
         // new behavior disabled until 2.0.0
-        "\t\tDelete failed operations from a resource's history allowing its current state to be rechecked.\n"
+        "\t\tIf resource has any past failures, clear its history and fail count.\n"
         "\t\t\t\tOptionally filtered by --resource, --node, --operation, and --interval (otherwise all).\n"
+        "\t\t\t\t--operation and --interval apply to fail counts, but entire history is always cleared,\n"
+        "\t\t\t\tto allow current state to be rechecked. If the named resource is part of a group, or\n"
+        "\t\t\t\tone numbered instance of a clone or bundled resource, the clean-up applies to the\n"
+        "\t\t\t\twhole collective resource unless --force is given."
     },
     {
         "refresh", no_argument, NULL, 'R',
 #endif
         "\t\tDelete resource's history (including failures) so its current state is rechecked.\n"
-        "\t\t\t\tOptionally filtered by --resource, --node, --operation, and --interval (otherwise all).\n"
-        "\t\t\t\tUnless --force is specified, resource's group or clone (if any) will also be cleaned"
+        "\t\t\t\tOptionally filtered by --resource and --node (otherwise all). If the named resource is\n"
+        "\t\t\t\tpart of a group, or one numbered instance of a clone or bundled resource, the clean-up\n"
+        "applies to the whole collective resource unless --force is given."
     },
     {
         "set-parameter", required_argument, NULL, 'p',
@@ -263,10 +270,16 @@ static struct crm_option long_options[] = {
     {
         "clear", no_argument, NULL, 'U',
         "\t\tRemove all constraints created by the --ban and/or --move commands.\n"
-        "\t\t\t\tRequires: --resource. Optional: --node, --master.\n"
+        "\t\t\t\tRequires: --resource. Optional: --node, --master, --expired.\n"
         "\t\t\t\tIf --node is not specified, all constraints created by --ban and --move\n"
         "\t\t\t\twill be removed for the named resource. If --node and --force are specified,\n"
-        "\t\t\t\tany constraint created by --move will be cleared, even if it is not for the specified node."
+        "\t\t\t\tany constraint created by --move will be cleared, even if it is not for the specified node.\n"
+        "\t\t\t\tIf --expired is specified, only those constraints whose lifetimes have expired will\n"
+        "\t\t\t\tbe removed.\n"
+    },
+    {
+        "expired", no_argument, NULL, 'e',
+        "\t\tModifies the --clear argument to remove constraints with expired lifetimes.\n"
     },
     {
         "lifetime", required_argument, NULL, 'u',
@@ -350,11 +363,13 @@ static struct crm_option long_options[] = {
     },
     {
         "operation", required_argument, NULL, 'n',
-        "\tOperation to clear instead of all (with -C -r)"
+        "\tOperation to clear instead of all (with -C -r)",
+        pcmk_option_hidden // only used with 2.0 -C behavior
     },
     {
         "interval", required_argument, NULL, 'I',
-        "\tInterval of operation to clear (default 0) (with -C -r -n)"
+        "\tInterval of operation to clear (default 0) (with -C -r -n)",
+        pcmk_option_hidden // only used with 2.0 -C behavior
     },
     {
         "set-name", required_argument, NULL, 's',
@@ -433,7 +448,7 @@ main(int argc, char **argv)
 
     char *xml_file = NULL;
     crm_ipc_t *crmd_channel = NULL;
-    pe_working_set_t data_set = { 0, };
+    pe_working_set_t *data_set = NULL;
     cib_t *cib_conn = NULL;
     resource_t *rsc = NULL;
     bool recursive = FALSE;
@@ -442,7 +457,7 @@ main(int argc, char **argv)
     bool require_resource = TRUE; /* whether command requires that resource be specified */
     bool require_dataset = TRUE;  /* whether command requires populated dataset instance */
     bool require_crmd = FALSE;    /* whether command requires connection to CRMd */
-    bool just_errors = TRUE;      /* whether cleanup command deletes all history or just errors */
+    bool clear_expired = FALSE;
 
     int rc = pcmk_ok;
     int is_ocf_rc = 0;
@@ -626,7 +641,12 @@ main(int argc, char **argv)
             case 'T':
                 timeout_ms = crm_get_msec(optarg);
                 break;
+            case 'e':
+                clear_expired = TRUE;
+                require_resource = FALSE;
+                break;
 
+            case 'C':
             case 'R':
             case 'P':
                 crm_log_args(argc, argv);
@@ -634,19 +654,7 @@ main(int argc, char **argv)
                 if (cib_file == NULL) {
                     require_crmd = TRUE;
                 }
-                just_errors = FALSE;
-                rsc_cmd = 'C';
-                find_flags = pe_find_renamed|pe_find_anon;
-                break;
-
-            case 'C':
-                crm_log_args(argc, argv);
-                require_resource = FALSE;
-                if (cib_file == NULL) {
-                    require_crmd = TRUE;
-                }
-                just_errors = FALSE; // disable until 2.0.0
-                rsc_cmd = 'C';
+                rsc_cmd = 'R'; // disable new behavior until 2.0
                 find_flags = pe_find_renamed|pe_find_anon;
                 break;
 
@@ -752,6 +760,12 @@ main(int argc, char **argv)
         require_resource = FALSE;
     }
 
+    // --expired without --clear/-U doesn't make sense
+    if (clear_expired == TRUE && rsc_cmd != 'U') {
+        CMD_ERR("--expired requires --clear or -U");
+        argerr++;
+    }
+
     if (optind < argc
         && argv[optind] != NULL
         && rsc_cmd == 0
@@ -800,8 +814,6 @@ main(int argc, char **argv)
         cib_options |= cib_quorum_override;
     }
 
-    data_set.input = NULL; /* make clean-up easier */
-
     if (require_resource && !rsc_id) {
         CMD_ERR("Must supply a resource id with -r");
         rc = -ENXIO;
@@ -814,6 +826,10 @@ main(int argc, char **argv)
 
     /* Establish a connection to the CIB */
     cib_conn = cib_new();
+    if ((cib_conn == NULL) || (cib_conn->cmds == NULL)) {
+        CMD_ERR("Could not create CIB connection: %s", pcmk_strerror(rc));
+        goto bail;
+    }
     rc = cib_conn->cmds->signon(cib_conn, crm_system_name, cib_command);
     if (rc != pcmk_ok) {
         CMD_ERR("Error signing on to the CIB service: %s", pcmk_strerror(rc));
@@ -836,17 +852,23 @@ main(int argc, char **argv)
         }
 
         /* Populate the working set instance */
-        set_working_set_defaults(&data_set);
-        rc = update_working_set_xml(&data_set, &cib_xml_copy);
+        data_set = pe_new_working_set();
+        if (data_set == NULL) {
+            rc = -ENOMEM;
+            goto bail;
+        }
+        set_bit(data_set->flags, pe_flag_no_counts);
+        set_bit(data_set->flags, pe_flag_no_compat);
+        rc = update_working_set_xml(data_set, &cib_xml_copy);
         if (rc != pcmk_ok) {
             goto bail;
         }
-        cluster_status(&data_set);
+        cluster_status(data_set);
     }
 
     // If command requires that resource exist if specified, find it
     if (find_flags && rsc_id) {
-        rsc = pe_find_resource_with_flags(data_set.resources, rsc_id,
+        rsc = pe_find_resource_with_flags(data_set->resources, rsc_id,
                                           find_flags);
         if (rsc == NULL) {
             CMD_ERR("Resource '%s' not found", rsc_id);
@@ -876,14 +898,14 @@ main(int argc, char **argv)
     /* Handle rsc_cmd appropriately */
     if (rsc_cmd == 'L') {
         rc = pcmk_ok;
-        cli_resource_print_list(&data_set, FALSE);
+        cli_resource_print_list(data_set, FALSE);
 
     } else if (rsc_cmd == 'l') {
         int found = 0;
         GListPtr lpc = NULL;
 
         rc = pcmk_ok;
-        for (lpc = data_set.resources; lpc != NULL; lpc = lpc->next) {
+        for (lpc = data_set->resources; lpc != NULL; lpc = lpc->next) {
             rsc = (resource_t *) lpc->data;
 
             found++;
@@ -896,6 +918,10 @@ main(int argc, char **argv)
         }
 
     } else if (rsc_cmd == 0 && rsc_long_cmd && safe_str_eq(rsc_long_cmd, "restart")) {
+        /* We don't pass data_set because rsc needs to stay valid for the entire
+         * lifetime of cli_resource_restart(), but it will reset and update the
+         * working set multiple times, so it needs to use its own copy.
+         */
         rc = cli_resource_restart(rsc, host_uname, timeout_ms, cib_conn);
 
     } else if (rsc_cmd == 0 && rsc_long_cmd && safe_str_eq(rsc_long_cmd, "wait")) {
@@ -904,21 +930,22 @@ main(int argc, char **argv)
     } else if (rsc_cmd == 0 && rsc_long_cmd) {
         // validate, force-(stop|start|demote|promote|check)
         rc = cli_resource_execute(rsc, rsc_id, rsc_long_cmd, override_params,
-                                  timeout_ms, cib_conn, &data_set);
+                                  timeout_ms, cib_conn, data_set);
         if (rc >= 0) {
             is_ocf_rc = 1;
         }
 
     } else if (rsc_cmd == 'A' || rsc_cmd == 'a') {
         GListPtr lpc = NULL;
-        xmlNode *cib_constraints = get_object_root(XML_CIB_TAG_CONSTRAINTS, data_set.input);
+        xmlNode *cib_constraints = get_object_root(XML_CIB_TAG_CONSTRAINTS,
+                                                   data_set->input);
 
-        unpack_constraints(cib_constraints, &data_set);
+        unpack_constraints(cib_constraints, data_set);
 
         // Constraints apply to group/clone, not member/instance
         rsc = uber_parent(rsc);
 
-        for (lpc = data_set.resources; lpc != NULL; lpc = lpc->next) {
+        for (lpc = data_set->resources; lpc != NULL; lpc = lpc->next) {
             resource_t *r = (resource_t *) lpc->data;
 
             clear_bit(r->flags, pe_rsc_allocating);
@@ -929,7 +956,7 @@ main(int argc, char **argv)
         fprintf(stdout, "* %s\n", rsc->id);
         cli_resource_print_location(rsc, NULL);
 
-        for (lpc = data_set.resources; lpc != NULL; lpc = lpc->next) {
+        for (lpc = data_set->resources; lpc != NULL; lpc = lpc->next) {
             resource_t *r = (resource_t *) lpc->data;
 
             clear_bit(r->flags, pe_rsc_allocating);
@@ -941,71 +968,74 @@ main(int argc, char **argv)
         GListPtr lpc = NULL;
 
         rc = pcmk_ok;
-        for (lpc = data_set.resources; lpc != NULL; lpc = lpc->next) {
+        for (lpc = data_set->resources; lpc != NULL; lpc = lpc->next) {
             rsc = (resource_t *) lpc->data;
             cli_resource_print_cts(rsc);
         }
-        cli_resource_print_cts_constraints(&data_set);
+        cli_resource_print_cts_constraints(data_set);
 
     } else if (rsc_cmd == 'F') {
-        rc = cli_resource_fail(crmd_channel, host_uname, rsc_id, &data_set);
+        rc = cli_resource_fail(crmd_channel, host_uname, rsc_id, data_set);
         if (rc == pcmk_ok) {
             start_mainloop();
         }
 
     } else if (rsc_cmd == 'O') {
-        rc = cli_resource_print_operations(rsc_id, host_uname, TRUE, &data_set);
+        rc = cli_resource_print_operations(rsc_id, host_uname, TRUE, data_set);
 
     } else if (rsc_cmd == 'o') {
-        rc = cli_resource_print_operations(rsc_id, host_uname, FALSE, &data_set);
+        rc = cli_resource_print_operations(rsc_id, host_uname, FALSE, data_set);
 
     } else if (rsc_cmd == 'W') {
-        rc = cli_resource_search(rsc, rsc_id, &data_set);
+        rc = cli_resource_search(rsc, rsc_id, data_set);
         if (rc >= 0) {
             rc = pcmk_ok;
         }
 
     } else if (rsc_cmd == 'q') {
-        rc = cli_resource_print(rsc, &data_set, TRUE);
+        rc = cli_resource_print(rsc, data_set, TRUE);
 
     } else if (rsc_cmd == 'w') {
-        rc = cli_resource_print(rsc, &data_set, FALSE);
+        rc = cli_resource_print(rsc, data_set, FALSE);
 
     } else if (rsc_cmd == 'Y') {
         node_t *dest = NULL;
 
         if (host_uname) {
-            dest = pe_find_node(data_set.nodes, host_uname);
+            dest = pe_find_node(data_set->nodes, host_uname);
             if (dest == NULL) {
                 CMD_ERR("Unknown node: %s", host_uname);
                 rc = -ENXIO;
                 goto bail;
             }
         }
-        cli_resource_why(cib_conn, data_set.resources, rsc, dest);
+        cli_resource_why(cib_conn, data_set->resources, rsc, dest);
         rc = pcmk_ok;
 
     } else if (rsc_cmd == 'U') {
         node_t *dest = NULL;
 
-        if (host_uname) {
-            dest = pe_find_node(data_set.nodes, host_uname);
+        if (clear_expired == TRUE) {
+            rc = cli_resource_clear_all_expired(data_set->input, cib_conn, rsc_id, host_uname, scope_master);
+
+        } else if (host_uname) {
+            dest = pe_find_node(data_set->nodes, host_uname);
             if (dest == NULL) {
                 CMD_ERR("Unknown node: %s", host_uname);
                 rc = -ENXIO;
                 goto bail;
             }
-            rc = cli_resource_clear(rsc_id, dest->details->uname, NULL, cib_conn);
+            rc = cli_resource_clear(rsc_id, dest->details->uname, NULL, cib_conn, TRUE);
 
         } else {
-            rc = cli_resource_clear(rsc_id, NULL, data_set.nodes, cib_conn);
+            rc = cli_resource_clear(rsc_id, NULL, data_set->nodes, cib_conn, TRUE);
         }
 
     } else if (rsc_cmd == 'M' && host_uname) {
-        rc = cli_resource_move(rsc, rsc_id, host_uname, cib_conn, &data_set);
+        rc = cli_resource_move(rsc, rsc_id, host_uname, cib_conn, data_set);
 
     } else if (rsc_cmd == 'B' && host_uname) {
-        node_t *dest = pe_find_node(data_set.nodes, host_uname);
+        node_t *dest = pe_find_node(data_set->nodes, host_uname);
 
         if (dest == NULL) {
             CMD_ERR("Error performing operation: node '%s' is unknown", host_uname);
@@ -1015,23 +1045,27 @@ main(int argc, char **argv)
         rc = cli_resource_ban(rsc_id, dest->details->uname, NULL, cib_conn);
 
     } else if (rsc_cmd == 'B' || rsc_cmd == 'M') {
+        pe_node_t *current = NULL;
+        unsigned int nactive = 0;
+
         rc = -EINVAL;
-        if (g_list_length(rsc->running_on) == 1) {
-            node_t *current = rsc->running_on->data;
+        current = pe__find_active_requires(rsc, &nactive);
+
+        if (nactive == 1) {
             rc = cli_resource_ban(rsc_id, current->details->uname, NULL, cib_conn);
 
         } else if(rsc->variant == pe_master) {
             int count = 0;
             GListPtr iter = NULL;
-            node_t *current = NULL;
 
+            current = NULL;
             for(iter = rsc->children; iter; iter = iter->next) {
                 resource_t *child = (resource_t *)iter->data;
                 enum rsc_role_e child_role = child->fns->state(child, TRUE);
 
                 if(child_role == RSC_ROLE_MASTER) {
                     count++;
-                    current = child->running_on->data;
+                    current = pe__current_node(child);
                 }
             }
 
@@ -1039,19 +1073,23 @@ main(int argc, char **argv)
                 rc = cli_resource_ban(rsc_id, current->details->uname, NULL, cib_conn);
 
             } else {
-                CMD_ERR("Resource '%s' not moved: active in %d locations (promoted in %d).", rsc_id, g_list_length(rsc->running_on), count);
-                CMD_ERR("You can prevent '%s' from running on a specific location with: --ban --node <name>", rsc_id);
-                CMD_ERR("You can prevent '%s' from being promoted at a specific location with:"
-                        " --ban --master --node <name>", rsc_id);
+                CMD_ERR("Resource '%s' not moved: active in %d locations (promoted in %d).",
+                        rsc_id, nactive, count);
+                CMD_ERR("To prevent '%s' from running on a specific location, "
+                        "specify a node.", rsc_id);
+                CMD_ERR("To prevent '%s' from being promoted at a specific "
+                        "location, specify a node and the master option.",
+                        rsc_id);
             }
 
         } else {
-            CMD_ERR("Resource '%s' not moved: active in %d locations.", rsc_id, g_list_length(rsc->running_on));
-            CMD_ERR("You can prevent '%s' from running on a specific location with: --ban --node <name>", rsc_id);
+            CMD_ERR("Resource '%s' not moved: active in %d locations.", rsc_id, nactive);
+            CMD_ERR("To prevent '%s' from running on a specific location, "
+                    "specify a node.", rsc_id);
         }
 
     } else if (rsc_cmd == 'G') {
-        rc = cli_resource_print_property(rsc, prop_name, &data_set);
+        rc = cli_resource_print_property(rsc, prop_name, data_set);
 
     } else if (rsc_cmd == 'S') {
         xmlNode *msg_data = NULL;
@@ -1077,7 +1115,7 @@ main(int argc, char **argv)
         free_xml(msg_data);
 
     } else if (rsc_cmd == 'g') {
-        rc = cli_resource_print_attribute(rsc, prop_name, &data_set);
+        rc = cli_resource_print_attribute(rsc, prop_name, data_set);
 
     } else if (rsc_cmd == 'p') {
         if (prop_value == NULL || strlen(prop_value) == 0) {
@@ -1089,63 +1127,26 @@ main(int argc, char **argv)
         /* coverity[var_deref_model] False positive */
         rc = cli_resource_update_attribute(rsc, rsc_id, prop_set, prop_id,
                                            prop_name, prop_value, recursive,
-                                           cib_conn, &data_set);
+                                           cib_conn, data_set);
 
     } else if (rsc_cmd == 'd') {
         /* coverity[var_deref_model] False positive */
         rc = cli_resource_delete_attribute(rsc, rsc_id, prop_set, prop_id,
-                                           prop_name, cib_conn, &data_set);
-
-    } else if (rsc_cmd == 'C' && just_errors) {
-        crmd_replies_needed = 0;
-        for (xmlNode *xml_op = __xml_first_child(data_set.failed); xml_op != NULL;
-             xml_op = __xml_next(xml_op)) {
-
-            const char *node = crm_element_value(xml_op, XML_ATTR_UNAME);
-            const char *task = crm_element_value(xml_op, XML_LRM_ATTR_TASK);
-            const char *task_interval = crm_element_value(xml_op, XML_LRM_ATTR_INTERVAL);
-            const char *resource_name = crm_element_value(xml_op, XML_LRM_ATTR_RSCID);
-
-            if(resource_name == NULL) {
-                continue;
-            } else if(host_uname && safe_str_neq(host_uname, node)) {
-                continue;
-            } else if(rsc_id && safe_str_neq(rsc_id, resource_name)) {
-                continue;
-            } else if(operation && safe_str_neq(operation, task)) {
-                continue;
-            } else if(interval && safe_str_neq(interval, task_interval)) {
-                continue;
-            }
-
-            crm_debug("Erasing %s failure for %s (%s detected) on %s",
-                      task, rsc->id, resource_name, node);
-            rc = cli_resource_delete(crmd_channel, node, rsc, task,
-                                     task_interval, &data_set);
-        }
-
-        if(rsc && (rc == pcmk_ok) && (BE_QUIET == FALSE)) {
-            /* Now check XML_RSC_ATTR_TARGET_ROLE and XML_RSC_ATTR_MANAGED */
-            cli_resource_check(cib_conn, rsc);
-        }
-
-        if (rc == pcmk_ok) {
-            start_mainloop();
-        }
+                                           prop_name, cib_conn, data_set);
 
     } else if ((rsc_cmd == 'C') && rsc) {
-        if(do_force == FALSE) {
+        if (do_force == FALSE) {
             rsc = uber_parent(rsc);
         }
-
-        crm_debug("Re-checking the state of %s (%s requested) on %s",
-                  rsc->id, rsc_id, host_uname);
         crmd_replies_needed = 0;
-        rc = cli_resource_delete(crmd_channel, host_uname, rsc, operation,
-                                 interval, &data_set);
 
-        if(rc == pcmk_ok && BE_QUIET == FALSE) {
-            /* Now check XML_RSC_ATTR_TARGET_ROLE and XML_RSC_ATTR_MANAGED */
+        crm_debug("Erasing failures of %s (%s requested) on %s",
+                  rsc->id, rsc_id, (host_uname? host_uname: "all nodes"));
+        rc = cli_resource_delete(crmd_channel, host_uname, rsc,
+                                 operation, interval, TRUE, data_set);
+
+        if ((rc == pcmk_ok) && !BE_QUIET) {
+            // Show any reasons why resource might stay stopped
             cli_resource_check(cib_conn, rsc);
         }
 
@@ -1154,6 +1155,33 @@ main(int argc, char **argv)
         }
 
     } else if (rsc_cmd == 'C') {
+        rc = cli_cleanup_all(crmd_channel, host_uname, operation, interval,
+                             data_set);
+        if (rc == pcmk_ok) {
+            start_mainloop();
+        }
+
+     } else if ((rsc_cmd == 'R') && rsc) {
+         if (do_force == FALSE) {
+             rsc = uber_parent(rsc);
+         }
+         crmd_replies_needed = 0;
+
+         crm_debug("Re-checking the state of %s (%s requested) on %s",
+                   rsc->id, rsc_id, (host_uname? host_uname: "all nodes"));
+         rc = cli_resource_delete(crmd_channel, host_uname, rsc,
+                                  NULL, 0, FALSE, data_set);
+
+         if ((rc == pcmk_ok) && !BE_QUIET) {
+             // Show any reasons why resource might stay stopped
+             cli_resource_check(cib_conn, rsc);
+         }
+
+         if (rc == pcmk_ok) {
+             start_mainloop();
+         }
+
+    } else if (rsc_cmd == 'R') {
 #if HAVE_ATOMIC_ATTRD
         const char *router_node = host_uname;
         xmlNode *msg_data = NULL;
@@ -1161,15 +1189,15 @@ main(int argc, char **argv)
         int attr_options = attrd_opt_none;
 
         if (host_uname) {
-            node_t *node = pe_find_node(data_set.nodes, host_uname);
+            node_t *node = pe_find_node(data_set->nodes, host_uname);
 
             if (node && is_remote_node(node)) {
-                if (node->details->remote_rsc == NULL || node->details->remote_rsc->running_on == NULL) {
+                node = pe__current_node(node->details->remote_rsc);
+                if (node == NULL) {
                     CMD_ERR("No lrmd connection detected to remote node %s", host_uname);
                     rc = -ENXIO;
                     goto bail;
                 }
-                node = node->details->remote_rsc->running_on->data;
                 router_node = node->details->uname;
                 attr_options |= attrd_opt_remote;
             }
@@ -1206,10 +1234,10 @@ main(int argc, char **argv)
         GListPtr rIter = NULL;
 
         crmd_replies_needed = 0;
-        for (rIter = data_set.resources; rIter; rIter = rIter->next) {
+        for (rIter = data_set->resources; rIter; rIter = rIter->next) {
             rsc = rIter->data;
-            cli_resource_delete(crmd_channel, host_uname, rsc, NULL, NULL,
-                                &data_set);
+            cli_resource_delete(crmd_channel, host_uname, rsc, NULL, 0,
+                                FALSE, data_set);
         }
 
         start_mainloop();
@@ -1237,10 +1265,7 @@ main(int argc, char **argv)
   bail:
 
     free(our_pid);
-
-    if (data_set.input != NULL) {
-        cleanup_alloc_calculations(&data_set);
-    }
+    pe_free_working_set(data_set);
     if (cib_conn != NULL) {
         cib_conn->cmds->signoff(cib_conn);
         cib_delete(cib_conn);
@@ -1248,7 +1273,7 @@ main(int argc, char **argv)
 
     if (rc == -pcmk_err_no_quorum) {
         CMD_ERR("Error performing operation: %s", pcmk_strerror(rc));
-        CMD_ERR("Try using -f");
+        CMD_ERR("To ignore quorum, use the force option");
 
     } else if (rc != pcmk_ok && !is_ocf_rc) {
         CMD_ERR("Error performing operation: %s", pcmk_strerror(rc));
