@@ -258,7 +258,8 @@ cli_resource_update_attribute(resource_t *rsc, const char *requested_name,
             if (rc == pcmk_ok && BE_QUIET == FALSE) {
                 printf("WARNING: There is already a meta attribute for '%s' called '%s' (id=%s)\n",
                        uber_parent(rsc)->id, attr_name, local_attr_id);
-                printf("         Delete '%s' first or use --force to override\n", local_attr_id);
+                printf("         Delete '%s' first or use the force option to override\n",
+                       local_attr_id);
             }
             free(local_attr_id);
             if (rc == pcmk_ok) {
@@ -452,8 +453,9 @@ send_lrm_rsc_op(crm_ipc_t * crmd_channel, const char *op,
     int rc = -ECOMM;
     xmlNode *cmd = NULL;
     xmlNode *xml_rsc = NULL;
-    const char *value = NULL;
     const char *router_node = host_uname;
+    const char *rsc_class = NULL;
+    const char *rsc_type = NULL;
     xmlNode *params = NULL;
     xmlNode *msg_data = NULL;
     resource_t *rsc = pe_find_resource(data_set->resources, rsc_id);
@@ -465,26 +467,48 @@ send_lrm_rsc_op(crm_ipc_t * crmd_channel, const char *op,
     } else if (rsc->variant != pe_native) {
         CMD_ERR("We can only process primitive resources, not %s", rsc_id);
         return -EINVAL;
+    }
 
-    } else if (host_uname == NULL) {
-        CMD_ERR("Please supply a node name with --node");
+    rsc_class = crm_element_value(rsc->xml, XML_AGENT_ATTR_CLASS);
+    rsc_type = crm_element_value(rsc->xml, XML_ATTR_TYPE);
+    if ((rsc_class == NULL) || (rsc_type == NULL)) {
+        CMD_ERR("Resource %s does not have a class and type", rsc_id);
         return -EINVAL;
-    } else {
-        node_t *node = pe_find_node(data_set->nodes, host_uname);
+    }
 
+    if (host_uname == NULL) {
+        CMD_ERR("Please specify a node name");
+        return -EINVAL;
+
+    } else {
+        pe_node_t *node = pe_find_node(data_set->nodes, host_uname);
+
+        if (node == NULL) {
+            CMD_ERR("Node %s not found", host_uname);
+            return -pcmk_err_node_unknown;
+        }
+
+        if (!(node->details->online)) {
+            CMD_ERR("Node %s is not online", host_uname);
+            return -ENOTCONN;
+        }
         if (node && is_remote_node(node)) {
-            if (node->details->remote_rsc == NULL || node->details->remote_rsc->running_on == NULL) {
+            node = pe__current_node(node->details->remote_rsc);
+            if (node == NULL) {
                 CMD_ERR("No lrmd connection detected to remote node %s", host_uname);
-                return -ENXIO;
+                return -ENOTCONN;
             }
-            node = node->details->remote_rsc->running_on->data;
             router_node = node->details->uname;
         }
     }
 
-    key = generate_transition_key(0, getpid(), 0, "xxxxxxxx-xrsc-opxx-xcrm-resourcexxxx");
-
     msg_data = create_xml_node(NULL, XML_GRAPH_TAG_RSC_OP);
+
+    /* The controller logs the transition key from requests, so we need to have
+     * *something* for it.
+     */
+    key = generate_transition_key(0, getpid(), 0,
+                                  "xxxxxxxx-xrsc-opxx-xcrm-resourcexxxx");
     crm_xml_add(msg_data, XML_ATTR_TRANSITION_KEY, key);
     free(key);
 
@@ -502,31 +526,20 @@ send_lrm_rsc_op(crm_ipc_t * crmd_channel, const char *op,
         crm_xml_add(xml_rsc, XML_ATTR_ID, rsc->id);
     }
 
-    value = crm_copy_xml_element(rsc->xml, xml_rsc, XML_ATTR_TYPE);
-    if (value == NULL) {
-        CMD_ERR("%s has no type!  Aborting...", rsc_id);
-        return -ENXIO;
-    }
-
-    value = crm_copy_xml_element(rsc->xml, xml_rsc, XML_AGENT_ATTR_CLASS);
-    if (value == NULL) {
-        CMD_ERR("%s has no class!  Aborting...", rsc_id);
-        return -ENXIO;
-    }
-
+    crm_xml_add(xml_rsc, XML_AGENT_ATTR_CLASS, rsc_class);
     crm_copy_xml_element(rsc->xml, xml_rsc, XML_AGENT_ATTR_PROVIDER);
+    crm_xml_add(xml_rsc, XML_ATTR_TYPE, rsc_type);
 
     params = create_xml_node(msg_data, XML_TAG_ATTRS);
     crm_xml_add(params, XML_ATTR_CRM_VERSION, CRM_FEATURE_SET);
 
-    key = crm_meta_name(XML_LRM_ATTR_INTERVAL);
+    // The controller parses the timeout from the request
+    key = crm_meta_name(XML_ATTR_TIMEOUT);
     crm_xml_add(params, key, "60000");  /* 1 minute */
     free(key);
 
     our_pid = crm_getpid_s();
     cmd = create_request(op, msg_data, router_node, CRM_SYSTEM_CRMD, crm_system_name, our_pid);
-
-/* 	crm_log_xml_warn(cmd, "send_lrm_rsc_op"); */
     free_xml(msg_data);
 
     if (crm_ipc_send(crmd_channel, cmd, 0, 0, NULL) > 0) {
@@ -558,15 +571,142 @@ rsc_fail_name(resource_t *rsc)
     return is_set(rsc->flags, pe_rsc_unique)? strdup(name) : clone_strip(name);
 }
 
+static int
+clear_rsc_history(crm_ipc_t *crmd_channel, const char *host_uname,
+                  const char *rsc_id, pe_working_set_t *data_set)
+{
+    int rc = pcmk_ok;
+
+    /* Erase the resource's entire LRM history in the CIB, even if we're only
+     * clearing a single operation's fail count. If we erased only entries for a
+     * single operation, we might wind up with a wrong idea of the current
+     * resource state, and we might not re-probe the resource.
+     */
+    rc = send_lrm_rsc_op(crmd_channel, CRM_OP_LRM_DELETE, host_uname, rsc_id,
+                         TRUE, data_set);
+    if (rc != pcmk_ok) {
+        return rc;
+    }
+    crmd_replies_needed++;
+
+    crm_trace("Processing %d mainloop inputs", crmd_replies_needed);
+    while (g_main_context_iteration(NULL, FALSE)) {
+        crm_trace("Processed mainloop input, %d still remaining",
+                  crmd_replies_needed);
+    }
+
+    if (crmd_replies_needed < 0) {
+        crmd_replies_needed = 0;
+    }
+    return rc;
+}
+
+static int
+clear_rsc_failures(crm_ipc_t *crmd_channel, const char *node_name,
+                   const char *rsc_id, const char *operation,
+                   const char *interval, pe_working_set_t *data_set)
+{
+    int rc = pcmk_ok;
+    const char *failed_value = NULL;
+    const char *failed_id = NULL;
+    const char *interval_ms_str = NULL;
+    GHashTable *rscs = NULL;
+    GHashTableIter iter;
+
+    /* Create a hash table to use as a set of resources to clean. This lets us
+     * clean each resource only once (per node) regardless of how many failed
+     * operations it has.
+     */
+    rscs = g_hash_table_new_full(crm_str_hash, g_str_equal, NULL, NULL);
+
+    // Normalize interval to milliseconds for comparison to history entry
+    if (operation) {
+        interval_ms_str = crm_strdup_printf("%llu", crm_get_interval(interval));
+    }
+
+    for (xmlNode *xml_op = __xml_first_child(data_set->failed); xml_op != NULL;
+         xml_op = __xml_next(xml_op)) {
+
+        failed_id = crm_element_value(xml_op, XML_LRM_ATTR_RSCID);
+        if (failed_id == NULL) {
+            // Malformed history entry, should never happen
+            continue;
+        }
+
+        // No resource specified means all resources match
+        if (rsc_id) {
+            resource_t *fail_rsc = pe_find_resource_with_flags(data_set->resources,
+                                                               failed_id,
+                                                               pe_find_renamed|pe_find_anon);
+
+            if (!fail_rsc || safe_str_neq(rsc_id, fail_rsc->id)) {
+                continue;
+            }
+        }
+
+        // Host name should always have been provided by this point
+        failed_value = crm_element_value(xml_op, XML_ATTR_UNAME);
+        if (safe_str_neq(node_name, failed_value)) {
+            continue;
+        }
+
+        // No operation specified means all operations match
+        if (operation) {
+            failed_value = crm_element_value(xml_op, XML_LRM_ATTR_TASK);
+            if (safe_str_neq(operation, failed_value)) {
+                continue;
+            }
+
+            // Interval (if operation was specified) defaults to 0 (not all)
+            failed_value = crm_element_value(xml_op, XML_LRM_ATTR_INTERVAL);
+            if (safe_str_neq(interval_ms_str, failed_value)) {
+                continue;
+            }
+        }
+
+        /* not available until glib 2.32
+        g_hash_table_add(rscs, (gpointer) failed_id);
+        */
+        g_hash_table_insert(rscs, (gpointer) failed_id, (gpointer) failed_id);
+    }
+
+    g_hash_table_iter_init(&iter, rscs);
+    while (g_hash_table_iter_next(&iter, (gpointer *) &failed_id, NULL)) {
+        crm_debug("Erasing failures of %s on %s", failed_id, node_name);
+        rc = clear_rsc_history(crmd_channel, node_name, failed_id, data_set);
+        if (rc != pcmk_ok) {
+            return rc;
+        }
+    }
+    g_hash_table_destroy(rscs);
+    return rc;
+}
+
+static int
+clear_rsc_fail_attrs(resource_t *rsc, const char *operation,
+                     const char *interval, node_t *node)
+{
+    int rc = pcmk_ok;
+    int attr_options = attrd_opt_none;
+    char *rsc_name = rsc_fail_name(rsc);
+
+    if (is_remote_node(node)) {
+        attr_options |= attrd_opt_remote;
+    }
+    rc = attrd_clear_delegate(NULL, node->details->uname, rsc_name, operation,
+                              interval, NULL, attr_options);
+    free(rsc_name);
+    return rc;
+}
+
 int
 cli_resource_delete(crm_ipc_t *crmd_channel, const char *host_uname,
                     resource_t *rsc, const char *operation,
-                    const char *interval, pe_working_set_t *data_set)
+                    const char *interval, bool just_failures,
+                    pe_working_set_t *data_set)
 {
     int rc = pcmk_ok;
     node_t *node = NULL;
-    char *rsc_name = NULL;
-    int attr_options = attrd_opt_none;
 
     if (rsc == NULL) {
         return -ENXIO;
@@ -578,8 +718,8 @@ cli_resource_delete(crm_ipc_t *crmd_channel, const char *host_uname,
             resource_t *child = (resource_t *) lpc->data;
 
             rc = cli_resource_delete(crmd_channel, host_uname, child, operation,
-                                     interval, data_set);
-            if(rc != pcmk_ok) {
+                                     interval, just_failures, data_set);
+            if (rc != pcmk_ok) {
                 return rc;
             }
         }
@@ -611,8 +751,13 @@ cli_resource_delete(crm_ipc_t *crmd_channel, const char *host_uname,
             node = (node_t *) lpc->data;
 
             if (node->details->online) {
-                cli_resource_delete(crmd_channel, node->details->uname, rsc,
-                                    operation, interval, data_set);
+                rc = cli_resource_delete(crmd_channel, node->details->uname,
+                                         rsc, operation, interval,
+                                         just_failures, data_set);
+            }
+            if (rc != pcmk_ok) {
+                g_list_free(nodes);
+                return rc;
             }
         }
 
@@ -637,54 +782,97 @@ cli_resource_delete(crm_ipc_t *crmd_channel, const char *host_uname,
     if (crmd_channel == NULL) {
         printf("Dry run: skipping clean-up of %s on %s due to CIB_file\n",
                rsc->id, host_uname);
-        return rc;
-     }
+        return pcmk_ok;
+    }
 
-    /* Erase the resource's entire LRM history in the CIB, even if we're only
-     * clearing a single operation's fail count. If we erased only entries for a
-     * single operation, we might wind up with a wrong idea of the current
-     * resource state, and we might not re-probe the resource.
-     */
-    rc = send_lrm_rsc_op(crmd_channel, CRM_OP_LRM_DELETE, host_uname, rsc->id,
-                         TRUE, data_set);
+    rc = clear_rsc_fail_attrs(rsc, operation, interval, node);
     if (rc != pcmk_ok) {
-        printf("Unable to clean up %s history on %s: %s\n",
-               rsc->id, host_uname, pcmk_strerror(rc));
+        printf("Unable to clean up %s failures on %s: %s\n",
+                rsc->id, host_uname, pcmk_strerror(rc));
         return rc;
     }
-    crmd_replies_needed++;
 
-    crm_trace("Processing %d mainloop inputs", crmd_replies_needed);
-    while(g_main_context_iteration(NULL, FALSE)) {
-        crm_trace("Processed mainloop input, %d still remaining",
-                  crmd_replies_needed);
+    if (just_failures) {
+        rc = clear_rsc_failures(crmd_channel, host_uname, rsc->id, operation,
+                                interval, data_set);
+    } else {
+        rc = clear_rsc_history(crmd_channel, host_uname, rsc->id, data_set);
     }
-
-    if(crmd_replies_needed < 0) {
-        crmd_replies_needed = 0;
-    }
-
-    rsc_name = rsc_fail_name(rsc);
-    if (is_remote_node(node)) {
-        attr_options |= attrd_opt_remote;
-    }
-    rc = attrd_clear_delegate(NULL, host_uname, rsc_name, operation, interval,
-                              NULL, attr_options);
     if (rc != pcmk_ok) {
-        printf("Cleaned %s history on %s, but unable to clear failures: %s\n",
+        printf("Cleaned %s failures on %s, but unable to clean history: %s\n",
                rsc->id, host_uname, pcmk_strerror(rc));
     } else {
         printf("Cleaned up %s on %s\n", rsc->id, host_uname);
     }
-    free(rsc_name);
-
     return rc;
 }
 
+int
+cli_cleanup_all(crm_ipc_t *crmd_channel, const char *node_name,
+                const char *operation, const char *interval,
+                pe_working_set_t *data_set)
+{
+    int rc = pcmk_ok;
+    int attr_options = attrd_opt_none;
+    const char *display_name = node_name? node_name : "all nodes";
+
+    if (crmd_channel == NULL) {
+        printf("Dry run: skipping clean-up of %s due to CIB_file\n",
+               display_name);
+        return pcmk_ok;
+    }
+    crmd_replies_needed = 0;
+
+    if (node_name) {
+        node_t *node = pe_find_node(data_set->nodes, node_name);
+
+        if (node == NULL) {
+            CMD_ERR("Unknown node: %s", node_name);
+            return -ENXIO;
+        }
+        if (is_remote_node(node)) {
+            attr_options |= attrd_opt_remote;
+        }
+    }
+
+    rc = attrd_clear_delegate(NULL, node_name, NULL, operation, interval,
+                              NULL, attr_options);
+    if (rc != pcmk_ok) {
+        printf("Unable to clean up all failures on %s: %s\n",
+                display_name, pcmk_strerror(rc));
+        return rc;
+    }
+
+    if (node_name) {
+        rc = clear_rsc_failures(crmd_channel, node_name, NULL,
+                                operation, interval, data_set);
+        if (rc != pcmk_ok) {
+            printf("Cleaned all resource failures on %s, but unable to clean history: %s\n",
+                   node_name, pcmk_strerror(rc));
+            return rc;
+        }
+    } else {
+        for (GList *iter = data_set->nodes; iter; iter = iter->next) {
+            pe_node_t *node = (pe_node_t *) iter->data;
+
+            rc = clear_rsc_failures(crmd_channel, node->details->uname, NULL,
+                                    operation, interval, data_set);
+            if (rc != pcmk_ok) {
+                printf("Cleaned all resource failures on all nodes, but unable to clean history: %s\n",
+                       pcmk_strerror(rc));
+                return rc;
+            }
+        }
+    }
+
+    printf("Cleaned up all resources on %s\n", display_name);
+    return pcmk_ok;
+}
+
 void
 cli_resource_check(cib_t * cib_conn, resource_t *rsc)
 {
-    int need_nl = 0;
+    bool printed = false;
     char *role_s = NULL;
     char *managed = NULL;
     resource_t *parent = uber_parent(rsc);
@@ -701,21 +889,25 @@ cli_resource_check(cib_t * cib_conn, resource_t *rsc)
             // Treated as if unset
 
         } else if(role == RSC_ROLE_STOPPED) {
-            printf("\n  * The configuration specifies that '%s' should remain stopped\n", parent->id);
-            need_nl++;
+            printf("\n  * Configuration specifies '%s' should remain stopped\n",
+                   parent->id);
+            printed = true;
 
         } else if(parent->variant == pe_master && role == RSC_ROLE_SLAVE) {
-            printf("\n  * The configuration specifies that '%s' should not be promoted\n", parent->id);
-            need_nl++;
+            printf("\n  * Configuration specifies '%s' should not be promoted\n",
+                   parent->id);
+            printed = true;
         }
     }
 
-    if(managed && crm_is_true(managed) == FALSE) {
-        printf("%s  * The configuration prevents the cluster from stopping or starting '%s' (unmanaged)\n", need_nl == 0?"\n":"", parent->id);
-        need_nl++;
+    if (managed && !crm_is_true(managed)) {
+        printf("%s  * Configuration prevents cluster from stopping or starting unmanaged '%s'\n",
+               (printed? "" : "\n"), parent->id);
+        printed = true;
     }
+    free(managed);
 
-    if(need_nl) {
+    if (printed) {
         printf("\n");
     }
 }
@@ -951,7 +1143,7 @@ update_dataset(cib_t *cib, pe_working_set_t * data_set, bool simulate)
     cib_t *shadow_cib = NULL;
     int rc;
 
-    cleanup_alloc_calculations(data_set);
+    pe_reset_working_set(data_set);
     rc = update_working_set_from_cib(data_set, cib);
     if (rc != pcmk_ok) {
         return rc;
@@ -1075,7 +1267,8 @@ max_delay_in(pe_working_set_t * data_set, GList *resources)
  * \return pcmk_ok on success, -errno on failure (exits on certain failures)
  */
 int
-cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t * cib)
+cli_resource_restart(pe_resource_t *rsc, const char *host, int timeout_ms,
+                     cib_t *cib)
 {
     int rc = 0;
     int lpc = 0;
@@ -1093,7 +1286,7 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
     GList *current_active = NULL;
     GList *restart_target_active = NULL;
 
-    pe_working_set_t data_set;
+    pe_working_set_t *data_set = NULL;
 
     if(resource_is_running_on(rsc, host) == FALSE) {
         const char *id = rsc->clone_name?rsc->clone_name:rsc->id;
@@ -1131,17 +1324,22 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
       - Allow a --no-deps option (aka. --force-restart)
     */
 
-
-    set_working_set_defaults(&data_set);
-    rc = update_dataset(cib, &data_set, FALSE);
+    data_set = pe_new_working_set();
+    if (data_set == NULL) {
+        crm_perror(LOG_ERR, "Could not allocate working set");
+        rc = -ENOMEM;
+        goto done;
+    }
+    set_bit(data_set->flags, pe_flag_no_counts);
+    set_bit(data_set->flags, pe_flag_no_compat);
+    rc = update_dataset(cib, data_set, FALSE);
     if(rc != pcmk_ok) {
         fprintf(stdout, "Could not get new resource list: %s (%d)\n", pcmk_strerror(rc), rc);
-        free(rsc_id);
-        return rc;
+        goto done;
     }
 
-    restart_target_active = get_active_resources(host, data_set.resources);
-    current_active = get_active_resources(host, data_set.resources);
+    restart_target_active = get_active_resources(host, data_set->resources);
+    current_active = get_active_resources(host, data_set->resources);
 
     dump_list(current_active, "Origin");
 
@@ -1162,7 +1360,7 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
         free(lookup_id);
         rc = cli_resource_update_attribute(rsc, rsc_id, NULL, NULL,
                                            XML_RSC_ATTR_TARGET_ROLE,
-                                           RSC_STOPPED, FALSE, cib, &data_set);
+                                           RSC_STOPPED, FALSE, cib, data_set);
     }
     if(rc != pcmk_ok) {
         fprintf(stderr, "Could not set target-role for %s: %s (%d)\n", rsc_id, pcmk_strerror(rc), rc);
@@ -1172,17 +1370,16 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
         if (restart_target_active) {
             g_list_free_full(restart_target_active, free);
         }
-        free(rsc_id);
-        return crm_exit(rc);
+        goto done;
     }
 
-    rc = update_dataset(cib, &data_set, TRUE);
+    rc = update_dataset(cib, data_set, TRUE);
     if(rc != pcmk_ok) {
         fprintf(stderr, "Could not determine which resources would be stopped\n");
         goto failure;
     }
 
-    target_active = get_active_resources(host, data_set.resources);
+    target_active = get_active_resources(host, data_set->resources);
     dump_list(target_active, "Target");
 
     list_delta = subtract_lists(current_active, target_active);
@@ -1193,7 +1390,7 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
     while(g_list_length(list_delta) > 0) {
         before = g_list_length(list_delta);
         if(timeout_ms == 0) {
-            step_timeout_s = max_delay_in(&data_set, list_delta) / sleep_interval;
+            step_timeout_s = max_delay_in(data_set, list_delta) / sleep_interval;
         }
 
         /* We probably don't need the entire step timeout */
@@ -1203,7 +1400,7 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
                 timeout -= sleep_interval;
                 crm_trace("%ds remaining", timeout);
             }
-            rc = update_dataset(cib, &data_set, FALSE);
+            rc = update_dataset(cib, data_set, FALSE);
             if(rc != pcmk_ok) {
                 fprintf(stderr, "Could not determine which resources were stopped\n");
                 goto failure;
@@ -1212,7 +1409,7 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
             if (current_active) {
                 g_list_free_full(current_active, free);
             }
-            current_active = get_active_resources(host, data_set.resources);
+            current_active = get_active_resources(host, data_set->resources);
             g_list_free(list_delta);
             list_delta = subtract_lists(current_active, target_active);
             dump_list(current_active, "Current");
@@ -1231,25 +1428,24 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
     }
 
     if (stop_via_ban) {
-        rc = cli_resource_clear(rsc_id, host, NULL, cib);
+        rc = cli_resource_clear(rsc_id, host, NULL, cib, TRUE);
 
     } else if (orig_target_role) {
         rc = cli_resource_update_attribute(rsc, rsc_id, NULL, NULL,
                                            XML_RSC_ATTR_TARGET_ROLE,
                                            orig_target_role, FALSE, cib,
-                                           &data_set);
+                                           data_set);
         free(orig_target_role);
         orig_target_role = NULL;
     } else {
         rc = cli_resource_delete_attribute(rsc, rsc_id, NULL, NULL,
                                            XML_RSC_ATTR_TARGET_ROLE, cib,
-                                           &data_set);
+                                           data_set);
     }
 
     if(rc != pcmk_ok) {
         fprintf(stderr, "Could not unset target-role for %s: %s (%d)\n", rsc_id, pcmk_strerror(rc), rc);
-        free(rsc_id);
-        return crm_exit(rc);
+        goto done;
     }
 
     if (target_active) {
@@ -1267,7 +1463,7 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
     while (waiting_for_starts(list_delta, rsc, host)) {
         before = g_list_length(list_delta);
         if(timeout_ms == 0) {
-            step_timeout_s = max_delay_in(&data_set, list_delta) / sleep_interval;
+            step_timeout_s = max_delay_in(data_set, list_delta) / sleep_interval;
         }
 
         /* We probably don't need the entire step timeout */
@@ -1279,7 +1475,7 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
                 crm_trace("%ds remaining", timeout);
             }
 
-            rc = update_dataset(cib, &data_set, FALSE);
+            rc = update_dataset(cib, data_set, FALSE);
             if(rc != pcmk_ok) {
                 fprintf(stderr, "Could not determine which resources were started\n");
                 goto failure;
@@ -1292,7 +1488,7 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
             /* It's OK if dependent resources moved to a different node,
              * so we check active resources on all nodes.
              */
-            current_active = get_active_resources(NULL, data_set.resources);
+            current_active = get_active_resources(NULL, data_set->resources);
             g_list_free(list_delta);
             list_delta = subtract_lists(target_active, current_active);
             dump_list(current_active, "Current");
@@ -1314,15 +1510,15 @@ cli_resource_restart(resource_t * rsc, const char *host, int timeout_ms, cib_t *
 
   failure:
     if (stop_via_ban) {
-        cli_resource_clear(rsc_id, host, NULL, cib);
+        cli_resource_clear(rsc_id, host, NULL, cib, TRUE);
     } else if (orig_target_role) {
         cli_resource_update_attribute(rsc, rsc_id, NULL, NULL,
                                       XML_RSC_ATTR_TARGET_ROLE,
-                                      orig_target_role, FALSE, cib, &data_set);
+                                      orig_target_role, FALSE, cib, data_set);
         free(orig_target_role);
     } else {
         cli_resource_delete_attribute(rsc, rsc_id, NULL, NULL,
-                                      XML_RSC_ATTR_TARGET_ROLE, cib, &data_set);
+                                      XML_RSC_ATTR_TARGET_ROLE, cib, data_set);
     }
 
 done:
@@ -1338,15 +1534,24 @@ done:
     if (restart_target_active) {
         g_list_free_full(restart_target_active, free);
     }
-    cleanup_alloc_calculations(&data_set);
     free(rsc_id);
+    pe_free_working_set(data_set);
     return rc;
 }
 
-#define action_is_pending(action) \
-    ((is_set((action)->flags, pe_action_optional) == FALSE) \
-    && (is_set((action)->flags, pe_action_runnable) == TRUE) \
-    && (is_set((action)->flags, pe_action_pseudo) == FALSE))
+static inline int action_is_pending(action_t *action) 
+{
+    if(is_set(action->flags, pe_action_optional)) {
+        return FALSE;
+    } else if(is_set(action->flags, pe_action_runnable) == FALSE) {
+        return FALSE;
+    } else if(is_set(action->flags, pe_action_pseudo)) {
+        return FALSE;
+    } else if(safe_str_eq("notify", action->task)) {
+        return FALSE;
+    }
+    return TRUE;
+}
 
 /*!
  * \internal
@@ -1362,7 +1567,9 @@ actions_are_pending(GListPtr actions)
     GListPtr action;
 
     for (action = actions; action != NULL; action = action->next) {
-        if (action_is_pending((action_t *) action->data)) {
+        action_t *a = (action_t *)action->data;
+        if (action_is_pending(a)) {
+            crm_notice("Waiting for %s (flags=0x%.8x)", a->uuid, a->flags);
             return TRUE;
         }
     }
@@ -1419,13 +1626,20 @@ print_pending_actions(GListPtr actions)
 int
 wait_till_stable(int timeout_ms, cib_t * cib)
 {
-    pe_working_set_t data_set;
+    pe_working_set_t *data_set = NULL;
     int rc = -1;
     int timeout_s = timeout_ms? ((timeout_ms + 999) / 1000) : WAIT_DEFAULT_TIMEOUT_S;
     time_t expire_time = time(NULL) + timeout_s;
     time_t time_diff;
+    bool printed_version_warning = BE_QUIET; // i.e. don't print if quiet
+
+    data_set = pe_new_working_set();
+    if (data_set == NULL) {
+        return -ENOMEM;
+    }
+    set_bit(data_set->flags, pe_flag_no_counts);
+    set_bit(data_set->flags, pe_flag_no_compat);
 
-    set_working_set_defaults(&data_set);
     do {
 
         /* Abort if timeout is reached */
@@ -1433,8 +1647,8 @@ wait_till_stable(int timeout_ms, cib_t * cib)
         if (time_diff > 0) {
             crm_info("Waiting up to %d seconds for cluster actions to complete", time_diff);
         } else {
-            print_pending_actions(data_set.actions);
-            cleanup_alloc_calculations(&data_set);
+            print_pending_actions(data_set->actions);
+            pe_free_working_set(data_set);
             return -ETIME;
         }
         if (rc == pcmk_ok) { /* this avoids sleep on first loop iteration */
@@ -1442,16 +1656,36 @@ wait_till_stable(int timeout_ms, cib_t * cib)
         }
 
         /* Get latest transition graph */
-        cleanup_alloc_calculations(&data_set);
-        rc = update_working_set_from_cib(&data_set, cib);
+        pe_reset_working_set(data_set);
+        rc = update_working_set_from_cib(data_set, cib);
         if (rc != pcmk_ok) {
-            cleanup_alloc_calculations(&data_set);
+            pe_free_working_set(data_set);
             return rc;
         }
-        do_calculations(&data_set, data_set.input, NULL);
+        do_calculations(data_set, data_set->input, NULL);
 
-    } while (actions_are_pending(data_set.actions));
+        if (!printed_version_warning) {
+            /* If the DC has a different version than the local node, the two
+             * could come to different conclusions about what actions need to be
+             * done. Warn the user in this case.
+             *
+             * @TODO A possible long-term solution would be to reimplement the
+             * wait as a new crmd operation that would be forwarded to the DC.
+             * However, that would have potential problems of its own.
+             */
+            const char *dc_version = g_hash_table_lookup(data_set->config_hash,
+                                                         "dc-version");
 
+            if (safe_str_neq(dc_version, PACEMAKER_VERSION "-" BUILD_VERSION)) {
+                printf("warning: wait option may not work properly in "
+                       "mixed-version cluster\n");
+                printed_version_warning = TRUE;
+            }
+        }
+
+    } while (actions_are_pending(data_set->actions));
+
+    pe_free_working_set(data_set);
     return pcmk_ok;
 }
 
@@ -1488,7 +1722,8 @@ cli_resource_execute(resource_t *rsc, const char *requested_name,
             if(rc > 0 && do_force == FALSE) {
                 CMD_ERR("It is not safe to %s %s here: the cluster claims it is already active",
                         action, rsc->id);
-                CMD_ERR("Try setting target-role=stopped first or specifying --force");
+                CMD_ERR("Try setting target-role=Stopped first or specifying "
+                        "the force option");
                 crm_exit(EPERM);
             }
         }
@@ -1500,7 +1735,8 @@ cli_resource_execute(resource_t *rsc, const char *requested_name,
     }
 
     if(rsc->variant == pe_group) {
-        CMD_ERR("Sorry, --%s doesn't support group resources", rsc_action);
+        CMD_ERR("Sorry, the %s option doesn't support group resources",
+                rsc_action);
         crm_exit(EOPNOTSUPP);
     }
 
@@ -1509,7 +1745,8 @@ cli_resource_execute(resource_t *rsc, const char *requested_name,
     rtype = crm_element_value(rsc->xml, XML_ATTR_TYPE);
 
     if (safe_str_eq(rclass, PCMK_RESOURCE_CLASS_STONITH)) {
-        CMD_ERR("Sorry, --%s doesn't support %s resources yet", rsc_action, rclass);
+        CMD_ERR("Sorry, the %s option doesn't support %s resources yet",
+                rsc_action, rclass);
         crm_exit(EOPNOTSUPP);
     }
 
@@ -1547,6 +1784,12 @@ cli_resource_execute(resource_t *rsc, const char *requested_name,
         setenv("OCF_TRACE_RA", "1", 1);
     }
 
+    /* A resource agent using the standard ocf-shellfuncs library will not print
+     * messages to stderr if it doesn't have a controlling terminal (e.g. if
+     * crm_resource is called via script or ssh). This forces it to do so.
+     */
+    setenv("OCF_TRACE_FILE", "/dev/stderr", 0);
+
     if (override_hash) {
         GHashTableIter iter;
         char *name = NULL;
@@ -1618,27 +1861,36 @@ cli_resource_move(resource_t *rsc, const char *rsc_id, const char *host_name,
                   cib_t *cib, pe_working_set_t *data_set)
 {
     int rc = -EINVAL;
-    int count = 0;
+    unsigned int count = 0;
     node_t *current = NULL;
     node_t *dest = pe_find_node(data_set->nodes, host_name);
     bool cur_is_dest = FALSE;
 
+    if (dest == NULL) {
+        CMD_ERR("Error performing operation: node '%s' is unknown", host_name);
+        return -ENXIO;
+    }
+
     if (scope_master && rsc->variant != pe_master) {
         resource_t *p = uber_parent(rsc);
         if(p->variant == pe_master) {
-            CMD_ERR("Using parent '%s' for --move command instead of '%s'.", rsc->id, rsc_id);
+            CMD_ERR("Using parent '%s' for move instead of '%s'.", rsc->id, rsc_id);
             rsc_id = p->id;
             rsc = p;
 
         } else {
-            CMD_ERR("Ignoring '--master' option: not valid for %s resources.",
+            CMD_ERR("Ignoring master option: not valid for %s resources.",
                     get_resource_typename(rsc->variant));
             scope_master = FALSE;
         }
     }
 
+    current = pe__find_active_requires(rsc, &count);
+
     if(rsc->variant == pe_master) {
         GListPtr iter = NULL;
+        unsigned int master_count = 0;
+        pe_node_t *master_node = NULL;
 
         for(iter = rsc->children; iter; iter = iter->next) {
             resource_t *child = (resource_t *)iter->data;
@@ -1646,37 +1898,27 @@ cli_resource_move(resource_t *rsc, const char *rsc_id, const char *host_name,
 
             if(child_role == RSC_ROLE_MASTER) {
                 rsc = child;
-                count++;
+                master_node = pe__current_node(child);
+                master_count++;
             }
         }
-
-        if(scope_master == FALSE && count == 0) {
-            count = g_list_length(rsc->running_on);
+        if (scope_master || master_count) {
+            count = master_count;
+            current = master_node;
         }
 
-    } else if (pe_rsc_is_clone(rsc)) {
-        count = g_list_length(rsc->running_on);
-
-    } else if (g_list_length(rsc->running_on) > 1) {
-        CMD_ERR("Resource '%s' not moved: active on multiple nodes", rsc_id);
-        return rc;
     }
 
-    if(dest == NULL) {
-        CMD_ERR("Error performing operation: node '%s' is unknown", host_name);
-        return -ENXIO;
+    if (count > 1) {
+        if (pe_rsc_is_clone(rsc)) {
+            current = NULL;
+        } else {
+            CMD_ERR("Resource '%s' not moved: active on multiple nodes", rsc_id);
+            return rc;
+        }
     }
 
-    if(g_list_length(rsc->running_on) == 1) {
-        current = rsc->running_on->data;
-    }
-
-    if(current == NULL) {
-        /* Nothing to check */
-
-    } else if(scope_master && rsc->fns->state(rsc, TRUE) != RSC_ROLE_MASTER) {
-        crm_trace("%s is already active on %s but not in correct state", rsc_id, dest->details->uname);
-    } else if (safe_str_eq(current->details->uname, dest->details->uname)) {
+    if (current && (current->details == dest->details)) {
         cur_is_dest = TRUE;
         if (do_force) {
             crm_info("%s is already %s on %s, reinforcing placement with location constraint.",
@@ -1688,8 +1930,11 @@ cli_resource_move(resource_t *rsc, const char *rsc_id, const char *host_name,
         }
     }
 
-    /* Clear any previous constraints for 'dest' */
-    cli_resource_clear(rsc_id, dest->details->uname, data_set->nodes, cib);
+    /* Clear any previous prefer constraints across all nodes. */
+    cli_resource_clear(rsc_id, NULL, data_set->nodes, cib, FALSE);
+
+    /* Clear any previous ban constraints on 'dest'. */
+    cli_resource_clear(rsc_id, dest->details->uname, data_set->nodes, cib, TRUE);
 
     /* Record an explicit preference for 'dest' */
     rc = cli_resource_prefer(rsc_id, dest->details->uname, cib);
@@ -1706,10 +1951,13 @@ cli_resource_move(resource_t *rsc, const char *rsc_id, const char *host_name,
             (void)cli_resource_ban(rsc_id, current->details->uname, NULL, cib);
 
         } else if(count > 1) {
-            CMD_ERR("Resource '%s' is currently %s in %d locations.  One may now move one to %s",
-                    rsc_id, scope_master?"promoted":"active", count, dest->details->uname);
-            CMD_ERR("You can prevent '%s' from being %s at a specific location with:"
-                    " --ban %s--host <name>", rsc_id, scope_master?"promoted":"active", scope_master?"--master ":"");
+            CMD_ERR("Resource '%s' is currently %s in %d locations. "
+                    "One may now move to %s",
+                    rsc_id, (scope_master? "promoted" : "active"),
+                    count, dest->details->uname);
+            CMD_ERR("To prevent '%s' from being %s at a specific location, "
+                    "specify a node.",
+                    rsc_id, (scope_master? "promoted" : "active"));
 
         } else {
             crm_trace("Not banning %s from its current location: not active", rsc_id);
