@@ -55,11 +55,9 @@ char ticket_cmd = 'S';
 char *xml_file = NULL;
 int cib_options = cib_sync_call;
 
-extern void cleanup_alloc_calculations(pe_working_set_t * data_set);
-
-#define CMD_ERR(fmt, args...) do {		\
-	crm_warn(fmt, ##args);			\
-	fprintf(stderr, fmt, ##args);		\
+#define CMD_ERR(fmt, args...) do {          \
+        crm_warn(fmt, ##args);              \
+        fprintf(stderr, fmt "\n", ##args);  \
     } while(0)
 
 static ticket_t *
@@ -658,7 +656,7 @@ static struct crm_option long_options[] = {
     {"attr-value", 1, 0, 'v', "\tAttribute value to use with -S"},
     {"default",    1, 0, 'd', "\t(Advanced) The default attribute value to display if none is found. For use with -G"},
     {"force",      0, 0, 'f', "\t\t(Advanced) Force the action to be performed"},
-    {"xml-file",   1, 0, 'x', NULL, 1},\
+    {"xml-file",   1, 0, 'x', NULL, 1},
 
     /* legacy options */
     {"set-name",   1, 0, 'n', "\t(Advanced) ID of the instance_attributes object to change"},
@@ -702,7 +700,7 @@ static struct crm_option long_options[] = {
 int
 main(int argc, char **argv)
 {
-    pe_working_set_t data_set;
+    pe_working_set_t *data_set = NULL;
     xmlNode *cib_xml_copy = NULL;
     xmlNode *cib_constraints = NULL;
 
@@ -836,13 +834,20 @@ main(int argc, char **argv)
         crm_help('?', EX_USAGE);
     }
 
-    set_working_set_defaults(&data_set);
+    data_set = pe_new_working_set();
+    if (data_set == NULL) {
+        crm_perror(LOG_CRIT, "Could not allocate working set");
+        rc = -ENOMEM;
+        goto bail;
+    }
+    set_bit(data_set->flags, pe_flag_no_counts);
+    set_bit(data_set->flags, pe_flag_no_compat);
 
     cib_conn = cib_new();
     if (cib_conn == NULL) {
         rc = -ENOTCONN;
         CMD_ERR("Error initiating the connection to the CIB service: %s\n", pcmk_strerror(rc));
-        return rc;
+        goto bail;
     }
 
     rc = cib_conn->cmds->signon(cib_conn, crm_system_name, cib_command);
@@ -866,15 +871,15 @@ main(int argc, char **argv)
         goto bail;
     }
 
-    data_set.input = cib_xml_copy;
-    data_set.now = crm_time_new(NULL);
+    data_set->input = cib_xml_copy;
+    data_set->now = crm_time_new(NULL);
 
-    cluster_status(&data_set);
+    cluster_status(data_set);
 
     /* For recording the tickets that are referenced in rsc_ticket constraints
      * but have never been granted yet. */
-    cib_constraints = get_object_root(XML_CIB_TAG_CONSTRAINTS, data_set.input);
-    unpack_constraints(cib_constraints, &data_set);
+    cib_constraints = get_object_root(XML_CIB_TAG_CONSTRAINTS, data_set->input);
+    unpack_constraints(cib_constraints, data_set);
 
     if (ticket_cmd == 'l' || ticket_cmd == 'L' || ticket_cmd == 'w') {
         gboolean raw = FALSE;
@@ -887,7 +892,7 @@ main(int argc, char **argv)
         }
 
         if (ticket_id) {
-            ticket_t *ticket = find_ticket(ticket_id, &data_set);
+            ticket_t *ticket = find_ticket(ticket_id, data_set);
 
             if (ticket == NULL) {
                 rc = -ENXIO;
@@ -896,7 +901,7 @@ main(int argc, char **argv)
             rc = print_ticket(ticket, raw, details);
 
         } else {
-            rc = print_ticket_list(&data_set, raw, details);
+            rc = print_ticket_list(data_set, raw, details);
         }
 
     } else if (ticket_cmd == 'q') {
@@ -914,7 +919,7 @@ main(int argc, char **argv)
             goto bail;
         }
 
-        rc = get_ticket_state_attr(ticket_id, get_attr_name, &value, &data_set);
+        rc = get_ticket_state_attr(ticket_id, get_attr_name, &value, data_set);
         if (rc == pcmk_ok) {
             fprintf(stdout, "%s\n", value);
         } else if (rc == -ENXIO && attr_default) {
@@ -932,7 +937,7 @@ main(int argc, char **argv)
         if (do_force == FALSE) {
             ticket_t *ticket = NULL;
 
-            ticket = find_ticket(ticket_id, &data_set);
+            ticket = find_ticket(ticket_id, data_set);
             if (ticket == NULL) {
                 rc = -ENXIO;
                 goto bail;
@@ -973,7 +978,7 @@ main(int argc, char **argv)
             goto bail;
         }
 
-        rc = modify_ticket_state(ticket_id, attr_delete, attr_set, cib_conn, &data_set);
+        rc = modify_ticket_state(ticket_id, attr_delete, attr_set, cib_conn, data_set);
         
         if (rc != pcmk_ok) {
             goto bail;
@@ -1013,8 +1018,10 @@ main(int argc, char **argv)
     }
     attr_delete = NULL;
 
+    pe_free_working_set(data_set);
+    data_set = NULL;
+
     if (cib_conn != NULL) {
-        cleanup_alloc_calculations(&data_set);
         cib_conn->cmds->signoff(cib_conn);
         cib_delete(cib_conn);
     }
