@@ -473,7 +473,7 @@ admin_msg_callback(const char *buffer, ssize_t length, gpointer userdata)
 
     } else if (DO_HEALTH) {
         xmlNode *data = get_message_xml(xml, F_CRM_DATA);
-        const char *state = crm_element_value(data, "crmd_state");
+        const char *state = crm_element_value(data, XML_PING_ATTR_CRMDSTATE);
 
         printf("Status of %s@%s: %s (%s)\n",
                crm_element_value(data, XML_PING_ATTR_SYSFROM),
@@ -487,9 +487,11 @@ admin_msg_callback(const char *buffer, ssize_t length, gpointer userdata)
     } else if (DO_WHOIS_DC) {
         const char *dc = crm_element_value(xml, F_CRM_HOST_FROM);
 
-        printf("Designated Controller is: %s\n", dc);
-        if (BE_SILENT && dc != NULL) {
-            fprintf(stderr, "%s\n", dc);
+        if (dc != NULL) {
+            if (BE_SILENT == FALSE) {
+                printf("Designated Controller is: ");
+            }
+            printf("%s\n", dc);
         }
         crm_exit(pcmk_ok);
     }
@@ -546,7 +548,9 @@ do_find_node_list(xmlNode * xml_node)
     xmlNode *node = NULL;
     xmlNode *nodes = get_object_root(XML_CIB_TAG_NODES, xml_node);
 
-    for (node = __xml_first_child(nodes); node != NULL; node = __xml_next(node)) {
+    for (node = __xml_first_child_element(nodes); node != NULL;
+         node = __xml_next_element(node)) {
+
         if (crm_str_eq((const char *)node->name, XML_CIB_TAG_NODE, TRUE)) {
 
             if (BASH_EXPORT) {
