@@ -147,12 +147,7 @@ create_op(xmlNode * cib_resource, const char *task, int interval, int outcome)
     lrmd_event_data_t *op = NULL;
     xmlNode *xop = NULL;
 
-    op = calloc(1, sizeof(lrmd_event_data_t));
-
-    op->rsc_id = strdup(ID(cib_resource));
-    op->interval = interval;
-    op->op_type = strdup(task);
-
+    op = lrmd_new_event(ID(cib_resource), task, interval);
     op->rc = outcome;
     op->op_status = 0;
     op->params = NULL;          /* TODO: Fill me in */
@@ -160,7 +155,9 @@ create_op(xmlNode * cib_resource, const char *task, int interval, int outcome)
     op->t_rcchange = op->t_run;
 
     op->call_id = 0;
-    for (xop = __xml_first_child(cib_resource); xop != NULL; xop = __xml_next(xop)) {
+    for (xop = __xml_first_child_element(cib_resource); xop != NULL;
+         xop = __xml_next_element(xop)) {
+
         int tmp = 0;
 
         crm_element_value_int(xop, XML_LRM_ATTR_CALLID, &tmp);
@@ -270,8 +267,8 @@ find_resource_xml(xmlNode * cib_node, const char *resource)
 
 
 static xmlNode *
-inject_resource(xmlNode * cib_node, const char *resource, const char *rclass, const char *rtype,
-                const char *rprovider)
+inject_resource(xmlNode * cib_node, const char *resource, const char *lrm_name,
+                const char *rclass, const char *rtype, const char *rprovider)
 {
     xmlNode *lrm = NULL;
     xmlNode *container = NULL;
@@ -280,9 +277,20 @@ inject_resource(xmlNode * cib_node, const char *resource, const char *rclass, co
 
     cib_resource = find_resource_xml(cib_node, resource);
     if (cib_resource != NULL) {
+        /* If an existing LRM history entry uses the resource name,
+         * continue using it, even if lrm_name is different.
+         */
         return cib_resource;
     }
 
+    // Check for history entry under preferred name
+    if (strcmp(resource, lrm_name)) {
+        cib_resource = find_resource_xml(cib_node, lrm_name);
+        if (cib_resource != NULL) {
+            return cib_resource;
+        }
+    }
+
     /* One day, add query for class, provider, type */
 
     if (rclass == NULL || rtype == NULL) {
@@ -300,13 +308,14 @@ inject_resource(xmlNode * cib_node, const char *resource, const char *rclass, co
         fprintf(stderr, "Invalid class for %s: %s\n", resource, rclass);
         return NULL;
 
-    } else if (crm_provider_required(rclass) && (rprovider == NULL)) {
+    } else if (is_set(pcmk_get_ra_caps(rclass), pcmk_ra_cap_provider)
+                && (rprovider == NULL)) {
         fprintf(stderr, "Please specify the provider for resource %s\n", resource);
         return NULL;
     }
 
     xpath = (char *)xmlGetNodePath(cib_node);
-    crm_info("Injecting new resource %s into %s '%s'", resource, xpath, ID(cib_node));
+    crm_info("Injecting new resource %s into %s '%s'", lrm_name, xpath, ID(cib_node));
     free(xpath);
 
     lrm = first_named_child(cib_node, XML_CIB_TAG_LRM);
@@ -323,7 +332,9 @@ inject_resource(xmlNode * cib_node, const char *resource, const char *rclass, co
     }
 
     cib_resource = create_xml_node(container, XML_LRM_TAG_RESOURCE);
-    crm_xml_add(cib_resource, XML_ATTR_ID, resource);
+
+    // If we're creating a new entry, use the preferred name
+    crm_xml_add(cib_resource, XML_ATTR_ID, lrm_name);
 
     crm_xml_add(cib_resource, XML_AGENT_ATTR_CLASS, rclass);
     crm_xml_add(cib_resource, XML_AGENT_ATTR_PROVIDER, rprovider);
@@ -584,7 +595,8 @@ modify_configuration(pe_working_set_t * data_set, cib_t *cib,
 
             update_failcounts(cib_node, resource, task, interval, outcome);
 
-            cib_resource = inject_resource(cib_node, resource, rclass, rtype, rprovider);
+            cib_resource = inject_resource(cib_node, resource, resource,
+                                           rclass, rtype, rprovider);
             CRM_ASSERT(cib_resource != NULL);
 
             op = create_op(cib_resource, task, interval, outcome);
@@ -630,6 +642,7 @@ exec_rsc_action(crm_graph_t * graph, crm_action_t * action)
     const char *rclass = NULL;
     const char *resource = NULL;
     const char *rprovider = NULL;
+    const char *lrm_name = NULL;
     const char *operation = crm_element_value(action->xml, "operation");
     const char *target_rc_s = crm_meta_value(action->params, XML_ATTR_TE_TARGET_RC);
 
@@ -658,6 +671,7 @@ exec_rsc_action(crm_graph_t * graph, crm_action_t * action)
      * If not found use the preferred name anyway
      */
     resource = crm_element_value(action_rsc, XML_ATTR_ID);
+    lrm_name = resource; // Preferred name when writing history
     if (pe_find_resource(fake_resource_list, resource) == NULL) {
         const char *longname = crm_element_value(action_rsc, XML_ATTR_ID_LONG);
 
@@ -689,7 +703,8 @@ exec_rsc_action(crm_graph_t * graph, crm_action_t * action)
     cib_node = inject_node_state(fake_cib, node, uname_is_uuid ? node : uuid);
     CRM_ASSERT(cib_node != NULL);
 
-    cib_resource = inject_resource(cib_node, resource, rclass, rtype, rprovider);
+    cib_resource = inject_resource(cib_node, resource, lrm_name,
+                                   rclass, rtype, rprovider);
     CRM_ASSERT(cib_resource != NULL);
 
     op = convert_graph_action(cib_resource, action, 0, target_outcome);
@@ -703,29 +718,43 @@ exec_rsc_action(crm_graph_t * graph, crm_action_t * action)
     for (gIter = fake_op_fail_list; gIter != NULL; gIter = gIter->next) {
         char *spec = (char *)gIter->data;
         char *key = NULL;
+        const char *match_name = NULL;
 
-        key = calloc(1, 1 + strlen(spec));
-        snprintf(key, strlen(spec), "%s_%s_%d@%s=", resource, op->op_type, op->interval, node);
-
+        // Allow user to specify anonymous clone with or without instance number
+        key = crm_strdup_printf("%s_%s_%d@%s=", resource, op->op_type,
+                                op->interval, node);
         if (strncasecmp(key, spec, strlen(key)) == 0) {
+            match_name = resource;
+        }
+        free(key);
+
+        if ((match_name == NULL) && strcmp(resource, lrm_name)) {
+            key = crm_strdup_printf("%s_%s_%d@%s=", lrm_name, op->op_type,
+                                    op->interval, node);
+            if (strncasecmp(key, spec, strlen(key)) == 0) {
+                match_name = lrm_name;
+            }
+            free(key);
+        }
+
+        if (match_name != NULL) {
+
             rc = sscanf(spec, "%*[^=]=%d", (int *) &op->rc);
-            // ${resource}_${task}_${interval}@${node}=${rc}
+            // ${match_name}_${task}_${interval_in_ms}@${node}=${rc}
 
             if (rc != 1) {
                 fprintf(stderr,
                         "Invalid failed operation spec: %s. Result code must be integer\n",
                         spec);
-                free(key);
                 continue;
             }
             action->failed = TRUE;
             graph->abort_priority = INFINITY;
             printf("\tPretending action %d failed with rc=%d\n", action->id, op->rc);
-            update_failcounts(cib_node, resource, op->op_type, op->interval, op->rc);
-            free(key);
+            update_failcounts(cib_node, match_name, op->op_type,
+                              op->interval, op->rc);
             break;
         }
-        free(key);
     }
 
     inject_op(cib_resource, op, target_outcome);
@@ -842,7 +871,7 @@ run_simulation(pe_working_set_t * data_set, cib_t *cib, GListPtr op_fail_list, b
         int rc = fake_cib->cmds->query(fake_cib, NULL, &cib_object, cib_sync_call | cib_scope_local);
 
         CRM_ASSERT(rc == pcmk_ok);
-        cleanup_alloc_calculations(data_set);
+        pe_reset_working_set(data_set);
         data_set->input = cib_object;
     }
 
